// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Net;
#if NET_4_0
using System.Threading;
using System.Threading.Tasks;
#endif

using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  public class ImapFetchMessageBodyStream : Stream {
    public static readonly int DefaultFetchBlockSize = 10 * 1024;

    public override bool CanSeek {
      get { return /*!IsClosed &&*/ false; }
    }

    public override bool CanRead {
      get { return !IsClosed /*&& true*/; }
    }

    public override bool CanWrite {
      get { return /*!IsClosed &&*/ false; }
    }

    public override bool CanTimeout {
      get { return true; }
    }

    private bool IsClosed {
      get { return session == null; }
    }

    public override long Position {
      get { RejectDisposed(); return position; }
      set { throw ExceptionUtils.CreateNotSupportedSeekingStream(); }
    }

    public override long Length {
      get { RejectDisposed(); return length; }
    }

    public override int ReadTimeout {
      get { RejectDisposed(); return session.ReceiveTimeout; }
      set { RejectDisposed(); session.ReceiveTimeout = value; }
    }

    public override int WriteTimeout {
      get { RejectDisposed(); return session.SendTimeout; }
      set { RejectDisposed(); session.SendTimeout = value; }
    }

    protected ImapSession Session {
      get { return session; }
    }

#if NET_4_0
    protected CancellationToken CancellationToken {
      get { return cancellationToken; }
    }
#endif

    internal protected long FetchBlockSize {
      get { return fetchBlockSize; }
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  int fetchBlockSize)
    {
      if (session == null)
        throw new ArgumentNullException("session");
      if (fetchMessageUid == null)
        throw new ArgumentNullException("fetchMessageUid");
      if (!fetchMessageUid.IsUidSet)
        throw new ArgumentException("must be uid", "fetchMessageUid");
      if (!fetchMessageUid.IsSingle)
        throw new ArgumentException("must be single number", "fetchMessageUid");
      if (fetchBlockSize <= 0)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("fetchBlockSize", fetchBlockSize);

      this.session = session;
      this.fetchPeek = fetchPeek;
      this.fetchMessageUid = fetchMessageUid;
      this.fetchBlockSize = (long)fetchBlockSize;
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  string fetchSectionOrHeaderFields,
                                                  ImapPartialRange? fetchRange,
                                                  int fetchBlockSize)
#if NET_4_0
      : this(session,
             fetchPeek,
             fetchMessageUid,
             fetchSectionOrHeaderFields,
             fetchRange,
             fetchBlockSize,
             CancellationToken.None)
    {
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  string fetchSectionOrHeaderFields,
                                                  ImapPartialRange? fetchRange,
                                                  int fetchBlockSize,
                                                  CancellationToken cancellationToken)
#endif
      : this(session, fetchPeek, fetchMessageUid, fetchBlockSize)
    {
      this.fetchSectionOrHeaderFields = fetchSectionOrHeaderFields;
      this.fetchRange = fetchRange;

      this.partialFetch = (!string.IsNullOrEmpty(fetchSectionOrHeaderFields) || fetchRange != null);

      if (fetchRange.HasValue && fetchRange.Value.IsLengthSpecified)
        this.length = fetchRange.Value.Length.Value;
      else
        this.length = 0L;

#if NET_4_0
      this.cancellationToken = cancellationToken;
#endif
    }

    internal protected ImapFetchMessageBodyStream(ImapSession session,
                                                  bool fetchPeek,
                                                  ImapSequenceSet fetchMessageUid,
                                                  IEnumerable<string> fetchHeaderFields,
                                                  int fetchBlockSize)
      : this(session,
             fetchPeek,
             fetchMessageUid,
             ToHeaderFieldSpecifier(fetchHeaderFields),
             null,
#if NET_4_0
             fetchBlockSize,
             CancellationToken.None)
#else
             fetchBlockSize)
#endif
    {
    }

    private static string ToHeaderFieldSpecifier(IEnumerable<string> fetchHeaderFields)
    {
      if (fetchHeaderFields == null) {
        return "HEADER";
      }
      else {
        var headerFields = fetchHeaderFields.ToArray();

        if (headerFields.Length == 0)
          return "HEADER";
        else
          return string.Concat("HEADER.FIELDS (", string.Join(" ", headerFields), ")");
      }
    }

    public override void Close()
    {
      Dispose(true);
    }

    protected void DetachFromSession()
    {
      session = null;
    }

    protected override void Dispose(bool disposing)
    {
      base.Dispose(disposing);

      buffer = null;
      session = null;
    }

    public override void SetLength(long value)
    {
      throw ExceptionUtils.CreateNotSupportedSettingStreamLength();
    }

    public override long Seek(long offset, SeekOrigin origin)
    {
      throw ExceptionUtils.CreateNotSupportedSeekingStream();
    }

    public override void Flush()
    {
      throw ExceptionUtils.CreateNotSupportedWritingStream();
    }

    public override void Write(byte[] buffer, int offset, int count)
    {
      throw ExceptionUtils.CreateNotSupportedWritingStream();
    }

    public override int Read(byte[] dest, int offset, int count)
    {
      RejectDisposed();

#if NET_4_0
      cancellationToken.ThrowIfCancellationRequested();
#endif

      if (dest == null)
        throw new ArgumentNullException("dest");
      if (offset < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive("offset", offset);
      if (count < 0)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive("count", count);
      if (dest.Length - count < offset)
        throw ExceptionUtils.CreateArgumentAttemptToAccessBeyondEndOfArray("offset", dest, offset, count);

      var read = 0;

      for (;;) {
        if (bufRemain == 0) {
          if (endOfStream)
            return read;
          else
            FillBuffer();
        }

        if (count <= bufRemain) {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, count);

          bufOffset += count;
          bufRemain -= count;
          position  += count;

          read += count;

          return read;
        }
        else {
          Buffer.BlockCopy(buffer, bufOffset, dest, offset, bufRemain);

          offset    += bufRemain;
          count     -= bufRemain;
          position  += bufRemain;

          read += bufRemain;

          bufRemain = 0;
        }
      }
    }

    public void ReadToEnd(Stream destination)
    {
      RejectDisposed();

      if (destination == null)
        throw new ArgumentNullException("destination");
      if (!destination.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("destination");

      // copy buffered data
      if (0 < bufRemain) {
        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }

      // fetch and copy
      for (;;) {
        if (endOfStream)
          return;

        FillBuffer();

        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }
    }

    public void ReadToEnd(BinaryWriter destination)
    {
      RejectDisposed();

      if (destination == null)
        throw new ArgumentNullException("destination");

      // copy buffered data
      if (0 < bufRemain) {
        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }

      // fetch and copy
      for (;;) {
        if (endOfStream)
          return;

        FillBuffer();

        destination.Write(buffer, bufOffset, bufRemain);

        position += bufRemain;
        //bufRemain = 0;
      }
    }

    [CLSCompliant(false)]
    internal protected virtual ImapCommandResult Prepare(ImapFetchDataItem fetchDataItem,
                                                         out IImapMessageAttribute messageAttr)
    {
      ImapMessage fetched;

      if (!partialFetch) {
        if (fetchDataItem == null)
          fetchDataItem = ImapFetchDataItem.Rfc822Size;
        else if (!fetchDataItem.ContainsOneOf(ImapFetchDataItem.Rfc822Size))
          fetchDataItem = ImapFetchDataItem.Rfc822Size + fetchDataItem;
      }

      var result = FetchAndFillBuffer(fetchDataItem, out fetched);

      if (result.Succeeded) {
        messageAttr = fetched;

        if (!partialFetch) {
          length = fetched.Rfc822Size;

          ReportFetchedOctetCount(position + bufRemain, length);
        }
      }
      else {
        messageAttr = null;
      }

      return result;
    }

    private void FillBuffer()
    {
      ImapMessageBody discard; // used as type specifier

      var result = FetchAndFillBuffer(null, out discard);

      if (result == null) {
        return; // nothing to fetch
      }
      else if (result.Succeeded) {
        if (!partialFetch)
          ReportFetchedOctetCount(position + bufRemain, length);
      }
      else {
        throw GetFetchFailureException(result);
      }
    }

    internal protected virtual void ReportFetchedOctetCount(long fetchedOctetCount, long totalOctetCount)
    {
      // do nothing in this class
    }

    private ImapCommandResult FetchAndFillBuffer<TMessageBody>(ImapFetchDataItem extraFetchDataItem,
                                                               out TMessageBody fetched)
      where TMessageBody : ImapMessageBody
    {
      fetched = null;

#if NET_4_0
      cancellationToken.ThrowIfCancellationRequested();
#endif

      ImapCommandResult result;
      TMessageBody[] fetchedMessages;

      try {
        var fetchDataItem = GetNextRangeFetchDataItem(extraFetchDataItem);

        if (fetchDataItem == null) {
          bufOffset = 0;
          bufRemain = 0;
          endOfStream = true;

          return null;
        }

        result = session.Fetch(fetchMessageUid,
                               fetchDataItem,
                               out fetchedMessages);
      }
      catch (TimeoutException ex) {
        throw GetTimeoutException(ex);
      }
      catch (ImapException ex) {
        throw GetUnexpectedException(ex);
      }

      if (result.Succeeded) {
        if (fetchedMessages.Length <= 0) {
          throw GetNoSuchMessageException(result);
        }
        else {
          fetched = fetchedMessages[0];

          var len = fetched.GetFirstBody(ref buffer);

          bufOffset = 0;
          bufRemain = len;

          endOfStream = (len < fetchBlockSize);
        }
      }
      else {
        fetched = null;
      }

      return result;
    }

    private ImapFetchDataItem GetNextRangeFetchDataItem(ImapFetchDataItem extraFetchDataItem)
    {
      ImapFetchDataItem bodyTextFetchDataItem;

      if (partialFetch) {
        var fetchStart = fetchRange.HasValue ? fetchRange.Value.Start : 0L;
        var fetchLength = (fetchRange.HasValue && fetchRange.Value.IsLengthSpecified)
          ? Math.Min(fetchRange.Value.Length.Value - position, fetchBlockSize)
          : fetchBlockSize;

        if (fetchLength == 0L)
          return null; // nothing to fetch

        bodyTextFetchDataItem = ImapFetchDataItem.BodyText(fetchPeek,
                                                           fetchSectionOrHeaderFields,
                                                           position + fetchStart,
                                                           fetchLength);
      }
      else {
        bodyTextFetchDataItem = ImapFetchDataItem.BodyText(fetchPeek,
                                                           string.Empty,
                                                           position,
                                                           fetchBlockSize);
      }

      if (extraFetchDataItem == null)
        return bodyTextFetchDataItem;
      else
        return extraFetchDataItem + bodyTextFetchDataItem;
    }

    protected virtual Exception GetNoSuchMessageException(ImapCommandResult result)
    {
      return new ImapProtocolViolationException("no such message or expunged");
    }

    protected virtual Exception GetFetchFailureException(ImapCommandResult result)
    {
      return new ImapProtocolViolationException(result.ResultText);
    }

    protected virtual Exception GetTimeoutException(TimeoutException ex)
    {
      return ex;
    }

    protected virtual Exception GetUnexpectedException(ImapException ex)
    {
      return ex;
    }

    private void RejectDisposed()
    {
      if (IsClosed)
        throw new ObjectDisposedException(GetType().FullName);
    }

    private ImapSession session;
    private readonly bool partialFetch;
    private readonly bool fetchPeek;
    private readonly ImapSequenceSet fetchMessageUid;
    private readonly string fetchSectionOrHeaderFields;
    private readonly ImapPartialRange? fetchRange;
    private readonly long fetchBlockSize;
    private long length = -1L;
    private long position = 0L;
    private byte[] buffer;
    private int bufOffset = 0;
    private int bufRemain = 0;
    private bool endOfStream = false;
#if NET_4_0
    private readonly CancellationToken cancellationToken;
#endif
  }
}
