//
// Author:
//       smdn <smdn@smdn.jp>
//
// Copyright (c) 2008-2014 smdn
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;

using Smdn.IO;
using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMailboxInfo {
    /*
     * APPEND
     */
    public void AppendMessage(Stream stream)
    {
      AppendMessage(new ImapAppendMessage(stream));
    }

    public void AppendMessage(Stream stream,
                              DateTimeOffset internalDate)
    {
      AppendMessage(new ImapAppendMessage(stream,
                                          internalDate));
    }

    public void AppendMessage(Stream stream,
                              IImapMessageFlagSet flags)
    {
      AppendMessage(new ImapAppendMessage(stream,
                                          flags));
    }

    public void AppendMessage(Stream stream,
                              DateTimeOffset internalDate,
                              IImapMessageFlagSet flags)
    {
      AppendMessage(new ImapAppendMessage(stream,
                                          internalDate,
                                          flags));
    }

    public void AppendMessage(IImapAppendMessage message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      ThrowIfNonExistentOrUnselectable();

      ProcessResult(Client.Session.Append(message, Mailbox),
                    ThrowIfAppendErrorResponse);
    }

    public void AppendMessages(IEnumerable<IImapAppendMessage> messages)
    {
      if (messages == null)
        throw new ArgumentNullException("messages");

      ThrowIfNonExistentOrUnselectable();

      if (Client.IsCapabilityAvailable(ImapCapability.MultiAppend)) {
        ProcessResult(Client.Session.AppendMultiple(messages, Mailbox),
                      ThrowIfAppendErrorResponse);
      }
      else {
        foreach (var message in messages) {
          ProcessResult(Client.Session.Append(message, Mailbox),
                        ThrowIfAppendErrorResponse);
        }
      }
    }

    public void AppendMessage(Action<Stream> writeMessage)
    {
      AppendMessageCore(null,
                        null,
                        null,
                        writeMessage);
    }

    public void AppendMessage(long length,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(length,
                        null,
                        null,
                        writeMessage);
    }

    public void AppendMessage(DateTimeOffset internalDate,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(null,
                        internalDate,
                        null,
                        writeMessage);
    }

    public void AppendMessage(IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(null,
                        null,
                        flags,
                        writeMessage);
    }

    public void AppendMessage(DateTimeOffset internalDate,
                              IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(null,
                        internalDate,
                        flags,
                        writeMessage);
    }

    public void AppendMessage(long length,
                              DateTimeOffset internalDate,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(length,
                        internalDate,
                        null,
                        writeMessage);
    }

    public void AppendMessage(long length,
                              IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(length,
                        null,
                        flags,
                        writeMessage);
    }

    public void AppendMessage(long length,
                              DateTimeOffset internalDate,
                              IImapMessageFlagSet flags,
                              Action<Stream> writeMessage)
    {
      AppendMessageCore(length,
                        internalDate,
                        flags,
                        writeMessage);
    }

    public void AppendMessage(ImapMessageInfo message)
    {
      if (message == null)
        throw new ArgumentNullException("message");

      if (message.Client == this.client)
        message.CopyTo(this);
      else
        AppendMessageInternal(message);
    }

    internal void AppendMessageInternal(ImapMessageInfo message)
    {
      var ret = message.OpenRead(string.Empty,
                                 null,
                                 null,
                                 ImapMessageFetchBodyOptions.Default,
                                 ImapFetchDataItem.Rfc822Size + ImapFetchDataItem.InternalDate + ImapFetchDataItem.Flags);

      using (var sourceStream = ret.Item1) {
        AppendMessageCore(ret.Item2.Rfc822Size,
                          ret.Item2.InternalDate,
                          ret.Item2.Flags,
                          delegate(Stream destStream) {
          sourceStream.CopyTo(destStream,
                              ImapFetchMessageBodyStream.DefaultFetchBlockSize);
        });
      }
    }

    private void AppendMessageCore(long? length,
                                   DateTimeOffset? internalDate,
                                   IImapMessageFlagSet flags,
                                   Action<Stream> writeMessage)
    {
      if (writeMessage == null)
        throw new ArgumentNullException("writeMessage");
      if (length.HasValue && length.Value < 0L)
        throw ExceptionUtils.CreateArgumentMustBeZeroOrPositive("length", length.Value);

      ThrowIfNonExistentOrUnselectable();

      var appendContext = client.Session.PrepareAppend(length,
                                                       internalDate,
                                                       flags,
                                                       mailbox);

      writeMessage(appendContext.WriteStream);

      ImapAppendedUidSet appendedUidSet;

      ProcessResult(appendContext.GetResult(out appendedUidSet),
                    ThrowIfAppendErrorResponse);
    }

    private bool ThrowIfAppendErrorResponse(ImapTaggedStatusResponse status)
    {
      if (status.ResponseText.Code == ImapResponseCode.TryCreate)
        throw new ImapMailboxNotFoundException(FullName);
      else
        return true;
    }
  }
}

