// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

using Smdn.Formats;
using Smdn.Formats.Mime;
using Smdn.IO;

using Smdn.Net.Imap4.Client.Session;

namespace Smdn.Net.Imap4.Client {
  partial class ImapMessageInfo {
    private static readonly string EntireMessageSectionPartSpecififer = string.Empty;

    public IImapBodyStructure GetStructureOf(params int[] subSections)
    {
      if (subSections == null || subSections.Length == 0)
        return BodyStructure;
      else
        return ImapBodyStructureUtils.FindSection(BodyStructure, subSections);
    }

    public TBodyStructure GetStructureOf<TBodyStructure>(params int[] subSections)
      where TBodyStructure : IImapBodyStructure
    {
      return (TBodyStructure)GetStructureOf(subSections);
    }

    public IImapBodyStructure GetStructureOf(string section)
    {
      section = ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section);

      if (string.IsNullOrEmpty(section))
        return BodyStructure;
      else
        return ImapBodyStructureUtils.FindSection(BodyStructure, section);
    }

    public TBodyStructure GetStructureOf<TBodyStructure>(string section)
      where TBodyStructure : IImapBodyStructure
    {
      return (TBodyStructure)GetStructureOf(section);
    }

    /*
     * utility methods for Open*(), Read*()
     */
    private string GetSection(IImapBodyStructure structure)
    {
      if (staticAttribute == null /* not fetched yed */ ||
          !object.ReferenceEquals(staticAttribute.BodyStructure, structure.GetRootStructure()))
        throw new ArgumentException("invalid structure", "structure");

      return structure.Section;
    }

    private static readonly Encoding iso8859_1 = Encoding.GetEncoding("ISO-8859-1");

    private static string GetEncoding(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      Encoding discard;

      return GetEncoding(structure, options, out discard);
    }

    private static string GetEncoding(IImapBodyStructure structure, ImapMessageFetchBodyOptions options, out Encoding charset)
    {
      var s = structure as ImapSinglePartBodyStructure;

      if (s != null && (int)(options & ImapMessageFetchBodyOptions.DecodeContent) != 0) {
        string charsetString;

        if (s.Parameters.TryGetValue("charset", out charsetString))
          charset = EncodingUtils.GetEncodingThrowException(charsetString);
        else
          charset = iso8859_1;

        return s.Encoding;
      }
      else {
        charset = null;

        return null;
      }
    }

    /*
     * OpenRead()
     */
    public Stream OpenRead(IImapBodyStructure section)
    {
      return OpenRead(GetSection(section),
                      null,
                      null,
                      ImapMessageFetchBodyOptions.Default,
                      null).Item1;
    }

    public Stream OpenRead(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return OpenRead(GetSection(section),
                      null,
                      GetEncoding(section, options),
                      options,
                      null).Item1;
    }

    public Stream OpenRead()
    {
      return OpenRead(EntireMessageSectionPartSpecififer,
                      null,
                      null,
                      ImapMessageFetchBodyOptions.Default,
                      null).Item1;
    }

    public Stream OpenRead(ImapMessageFetchBodyOptions options)
    {
      return OpenRead(EntireMessageSectionPartSpecififer,
                      null,
                      null,
                      options,
                      null).Item1;
    }

    public Stream OpenRead(string section)
    {
      return OpenRead(ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section),
                      null,
                      null,
                      ImapMessageFetchBodyOptions.Default,
                      null).Item1;
    }

    public Stream OpenRead(string section, ImapPartialRange? range)
    {
      return OpenRead(ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section),
                      range,
                      null,
                      ImapMessageFetchBodyOptions.Default,
                      null).Item1;
    }

    public Stream OpenRead(string section, ImapMessageFetchBodyOptions options)
    {
      return OpenRead(ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section),
                      null,
                      null,
                      options,
                      null).Item1;
    }

    public Stream OpenRead(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      return OpenRead(ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section),
                      range,
                      null,
                      options,
                      null).Item1;
    }

    /*
     * ReadAs<TResult>(Converter<Stream>)
     */
    public TResult ReadAs<TResult>(IImapBodyStructure section, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TResult ReadAs<TResult>(IImapBodyStructure section, ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(section, options), converter);
    }

    public TResult ReadAs<TResult>(Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(), converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(options), converter);
    }

    public TResult ReadAs<TResult>(string section, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(section), converter);
    }

    public TResult ReadAs<TResult>(string section, ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(section, options), converter);
    }

    public TResult ReadAs<TResult>(string section, ImapPartialRange? range, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(section, range), converter);
    }

    public TResult ReadAs<TResult>(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options, Converter<Stream, TResult> converter)
    {
      return ReadAsCore(OpenRead(section, range, options), converter);
    }

    private static TResult ReadAsCore<TResult>(Stream stream, Converter<Stream, TResult> converter)
    {
      using (stream) {
        if (converter == null)
          throw new ArgumentNullException("converter");

        return converter(stream);
      }
    }

    /*
     * ReadAs<TResult>(Func<Stream, ...>)
     */
    public TResult ReadAs<T, TResult>
      (Func<Stream, T, TResult> read, T arg)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg);
    }

    public TResult ReadAs<T1, T2, TResult>
      (Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg1, arg2);
    }

    public TResult ReadAs<T1, T2, T3, TResult>
      (Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg1, arg2, arg3);
    }

    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      return ReadAs(ImapMessageFetchBodyOptions.Default, read, arg1, arg2, arg3, arg4);
    }

    public TResult ReadAs<T, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T, TResult> read, T arg)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg);
      }
    }

    public TResult ReadAs<T1, T2, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T1, T2, TResult> read, T1 arg1, T2 arg2)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg1, arg2);
      }
    }

    public TResult ReadAs<T1, T2, T3, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T1, T2, T3, TResult> read, T1 arg1, T2 arg2, T3 arg3)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg1, arg2, arg3);
      }
    }

    public TResult ReadAs<T1, T2, T3, T4, TResult>
      (ImapMessageFetchBodyOptions options,
       Func<Stream, T1, T2, T3, T4, TResult> read, T1 arg1, T2 arg2, T3 arg3, T4 arg4)
    {
      if (read == null)
        throw new ArgumentNullException("read");

      using (var stream = OpenRead(options)) {
        return read(stream, arg1, arg2, arg3, arg4);
      }
    }

    /*
     * ReadAllBytes()
     */
    public byte[] ReadAllBytes(IImapBodyStructure section)
    {
      return ReadAs<byte[]>(section, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(section, options, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes()
    {
      return ReadAs<byte[]>(ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(options, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section)
    {
      return ReadAs<byte[]>(section, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(section, options, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section, ImapPartialRange? range)
    {
      return ReadAs<byte[]>(section, range, ReadAllBytesProc);
    }

    public byte[] ReadAllBytes(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<byte[]>(section, range, options, ReadAllBytesProc);
    }

    private static byte[] ReadAllBytesProc(Stream stream)
    {
      return stream.ReadToEnd();
    }

    /*
     * WriteTo(Stream)
     */
    public void WriteTo(Stream stream, IImapBodyStructure section)
    {
      WriteToCore(stream, delegate { return OpenRead(section); });
    }

    public void WriteTo(Stream stream, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(stream, delegate { return OpenRead(section, options); });
    }

    public void WriteTo(Stream stream)
    {
      WriteToCore(stream, delegate { return OpenRead(); });
    }

    public void WriteTo(Stream stream, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(stream, delegate { return OpenRead(options); });
    }

    public void WriteTo(Stream stream, string section)
    {
      WriteToCore(stream, delegate { return OpenRead(section); });
    }

    public void WriteTo(Stream stream, string section, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(stream, delegate { return OpenRead(section, options); });
    }

    public void WriteTo(Stream stream, string section, ImapPartialRange? range)
    {
      WriteToCore(stream, delegate { return OpenRead(section, range); });
    }

    public void WriteTo(Stream stream, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(stream, delegate { return OpenRead(section, range, options); });
    }

    private static void WriteToCore(Stream stream, Func<Stream> openRead)
    {
      if (stream == null)
        throw new ArgumentNullException("stream");
      if (!stream.CanWrite)
        throw ExceptionUtils.CreateArgumentMustBeWritableStream("stream");

      using (var messageStream = openRead()) {
        CopyFetchMessageBodyStream(messageStream, stream);
      }
    }

    private static void CopyFetchMessageBodyStream(Stream messageStream, Stream destination)
    {
      var fetchStream = messageStream as ImapFetchMessageBodyStream;

      if (fetchStream == null)
        messageStream.CopyTo(destination);
      else
        fetchStream.ReadToEnd(destination);
    }

    private static void CopyFetchMessageBodyStream(Stream messageStream, BinaryWriter destination)
    {
      var fetchStream = messageStream as ImapFetchMessageBodyStream;

      if (fetchStream == null)
        messageStream.CopyTo(destination);
      else
        fetchStream.ReadToEnd(destination);
    }

    /*
     * WriteTo(BinaryWriter)
     */
    public void WriteTo(BinaryWriter writer, IImapBodyStructure section)
    {
      WriteToCore(writer, delegate { return OpenRead(section); });
    }

    public void WriteTo(BinaryWriter writer, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(writer, delegate { return OpenRead(section, options); });
    }

    public void WriteTo(BinaryWriter writer)
    {
      WriteToCore(writer, delegate { return OpenRead(); });
    }

    public void WriteTo(BinaryWriter writer, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(writer, delegate { return OpenRead(options); });
    }

    public void WriteTo(BinaryWriter writer, string section)
    {
      WriteToCore(writer, delegate { return OpenRead(section); });
    }

    public void WriteTo(BinaryWriter writer, string section, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(writer, delegate { return OpenRead(section, options); });
    }

    public void WriteTo(BinaryWriter writer, string section, ImapPartialRange? range)
    {
      WriteToCore(writer, delegate { return OpenRead(section, range); });
    }

    public void WriteTo(BinaryWriter writer, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      WriteToCore(writer, delegate { return OpenRead(section, range, options); });
    }

    private static void WriteToCore(BinaryWriter writer, Func<Stream> openRead)
    {
      if (writer == null)
        throw new ArgumentNullException("writer");

      using (var stream = openRead()) {
        CopyFetchMessageBodyStream(stream, writer);
      }
    }

    /*
     * Save()
     */
    public void Save(string path, IImapBodyStructure section)
    {
      SaveCore(path, delegate { return OpenRead(section); });
    }

    public void Save(string path, IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(section, options); });
    }

    public void Save(string path)
    {
      SaveCore(path, delegate { return OpenRead(); });
    }

    public void Save(string path, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(options); });
    }

    public void Save(string path, string section)
    {
      SaveCore(path, delegate { return OpenRead(section); });
    }

    public void Save(string path, string section, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(section, options); });
    }

    public void Save(string path, string section, ImapPartialRange? range)
    {
      SaveCore(path, delegate { return OpenRead(section, range); });
    }

    public void Save(string path, string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      SaveCore(path, delegate { return OpenRead(section, range, options); });
    }

    private static void SaveCore(string path, Func<Stream> openRead)
    {
      if (path == null)
        throw new ArgumentNullException("path");
      if (path.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("path");

      using (var stream = openRead()) {
        using (var fileStream = new FileStream(path, FileMode.Create, FileAccess.Write)) {
          CopyFetchMessageBodyStream(stream, fileStream);
        }
      }
    }

    /*
     * Download()
     */
    public Stream Download(IImapBodyStructure section)
    {
      return DownloadCore(OpenRead(section));
    }

    public Stream Download(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      return DownloadCore(OpenRead(section, options));
    }

    public Stream Download()
    {
      return DownloadCore(OpenRead());
    }

    public Stream Download(ImapMessageFetchBodyOptions options)
    {
      return DownloadCore(OpenRead(options));
    }

    public Stream Download(string section)
    {
      return DownloadCore(OpenRead(section));
    }

    public Stream Download(string section, ImapMessageFetchBodyOptions options)
    {
      return DownloadCore(OpenRead(section, options));
    }

    public Stream Download(string section, ImapPartialRange? range)
    {
      return DownloadCore(OpenRead(section, range));
    }

    public Stream Download(string section, ImapPartialRange? range, ImapMessageFetchBodyOptions options)
    {
      return DownloadCore(OpenRead(section, range, options));
    }

    private static Stream DownloadCore(Stream messageStream)
    {
      using (messageStream) {
        var ret = new ChunkedMemoryStream();

        CopyFetchMessageBodyStream(messageStream, ret);

        ret.Position = 0L;

        return ret;
      }
    }

    /*
     * OpenText
     */
    public StreamReader OpenText(IImapBodyStructure section)
    {
      return OpenText(section, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(IImapBodyStructure section, ImapMessageFetchBodyOptions options)
    {
      Encoding charset;
      var stream = OpenRead(GetSection(section),
                            null,
                            GetEncoding(section, options, out charset),
                            options,
                            null).Item1;

      return new StreamReader(stream, charset ?? iso8859_1);
    }

    public StreamReader OpenText()
    {
      return OpenText(EntireMessageSectionPartSpecififer, iso8859_1, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(ImapMessageFetchBodyOptions options)
    {
      return OpenText(EntireMessageSectionPartSpecififer, iso8859_1, options);
    }

    public StreamReader OpenText(string section)
    {
      return OpenText(section, iso8859_1, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(string section, ImapMessageFetchBodyOptions options)
    {
      return OpenText(section, iso8859_1, options);
    }

    public StreamReader OpenText(Encoding charset)
    {
      return OpenText(EntireMessageSectionPartSpecififer, charset, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return OpenText(EntireMessageSectionPartSpecififer, charset, options);
    }

    public StreamReader OpenText(string section, Encoding charset)
    {
      return OpenText(section, charset, ImapMessageFetchBodyOptions.Default);
    }

    public StreamReader OpenText(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      if (charset == null)
        throw new ArgumentNullException("charset");

      return new StreamReader(OpenRead(ImapBodyStructureUtils.GetValidatedSectionSpecOfSectionPart(section),
                                       null,
                                       null,
                                       options,
                                       null).Item1,
                              charset);
    }

    /*
     * ReadAs<TResult>(Converter<StreamReader>)
     */
    public TResult ReadAs<TResult>(Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(), converter);
    }

    public TResult ReadAs<TResult>(ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(options), converter);
    }

    public TResult ReadAs<TResult>(IImapBodyStructure section, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(section), converter);
    }

    public TResult ReadAs<TResult>(IImapBodyStructure section, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(section, options), converter);
    }

    public TResult ReadAs<TResult>(string section, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(section), converter);
    }

    public TResult ReadAs<TResult>(string section, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(section, options), converter);
    }

    public TResult ReadAs<TResult>(Encoding charset, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(charset), converter);
    }

    public TResult ReadAs<TResult>(Encoding charset, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(charset, options), converter);
    }

    public TResult ReadAs<TResult>(string section, Encoding charset, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(section, charset), converter);
    }

    public TResult ReadAs<TResult>(string section, Encoding charset, ImapMessageFetchBodyOptions options, Converter<StreamReader, TResult> converter)
    {
      return ReadAsCore(OpenText(section, charset, options), converter);
    }

    private static TResult ReadAsCore<TResult>(StreamReader reader, Converter<StreamReader, TResult> converter)
    {
      using (reader) {
        if (converter == null)
          throw new ArgumentNullException("converter");

        return converter(reader);
      }
    }

    /*
     * ReadLines()
     */
    public IEnumerable<string> ReadLines(IImapBodyStructure structure)
    {
      return ReadLinesCore(OpenText(structure));
    }

    public IEnumerable<string> ReadLines(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadLinesCore(OpenText(structure, options));
    }

    public IEnumerable<string> ReadLines()
    {
      return ReadLinesCore(OpenText());
    }

    public IEnumerable<string> ReadLines(ImapMessageFetchBodyOptions options)
    {
      return ReadLinesCore(OpenText(options));
    }

    public IEnumerable<string> ReadLines(string section)
    {
      return ReadLinesCore(OpenText(section));
    }

    public IEnumerable<string> ReadLines(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadLinesCore(OpenText(section, options));
    }

    public IEnumerable<string> ReadLines(Encoding charset)
    {
      return ReadLinesCore(OpenText(charset));
    }

    public IEnumerable<string> ReadLines(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadLinesCore(OpenText(charset, options));
    }

    public IEnumerable<string> ReadLines(string section, Encoding charset)
    {
      return ReadLinesCore(OpenText(section, charset));
    }

    public IEnumerable<string> ReadLines(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadLinesCore(OpenText(section, charset, options));
    }

    private static IEnumerable<string> ReadLinesCore(StreamReader reader)
    {
      using (reader) {
        for (;;) {
          var line = reader.ReadLine();

          if (line == null)
            break;
          else
            yield return line;
        }
      }
    }

    /*
     * ReadAllLines()
     */
    public string[] ReadAllLines(IImapBodyStructure structure)
    {
      return ReadLines(structure).ToArray();
    }

    public string[] ReadAllLines(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(structure, options).ToArray();
    }

    public string[] ReadAllLines()
    {
      return ReadLines().ToArray();
    }

    public string[] ReadAllLines(ImapMessageFetchBodyOptions options)
    {
      return ReadLines(options).ToArray();
    }

    public string[] ReadAllLines(string section)
    {
      return ReadLines(section).ToArray();
    }

    public string[] ReadAllLines(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(section, options).ToArray();
    }

    public string[] ReadAllLines(Encoding charset)
    {
      return ReadLines(charset).ToArray();
    }

    public string[] ReadAllLines(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(charset, options).ToArray();
    }

    public string[] ReadAllLines(string section, Encoding charset)
    {
      return ReadLines(section, charset).ToArray();
    }

    public string[] ReadAllLines(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadLines(section, charset, options).ToArray();
    }

    /*
     * ReadAllText()
     */
    public string ReadAllText(IImapBodyStructure structure)
    {
      return ReadAs<string>(structure, ReadAllTextProc);
    }

    public string ReadAllText(IImapBodyStructure structure, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(structure, options, ReadAllTextProc);
    }

    public string ReadAllText()
    {
      return ReadAs<string>(ReadAllTextProc);
    }

    public string ReadAllText(ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(options, ReadAllTextProc);
    }

    public string ReadAllText(string section)
    {
      return ReadAs<string>(section, ReadAllTextProc);
    }

    public string ReadAllText(string section, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(section, options, ReadAllTextProc);
    }

    public string ReadAllText(Encoding charset)
    {
      return ReadAs<string>(charset, ReadAllTextProc);
    }

    public string ReadAllText(Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(charset, options, ReadAllTextProc);
    }

    public string ReadAllText(string section, Encoding charset)
    {
      return ReadAs<string>(section, charset, ReadAllTextProc);
    }

    public string ReadAllText(string section, Encoding charset, ImapMessageFetchBodyOptions options)
    {
      return ReadAs<string>(section, charset, options, ReadAllTextProc);
    }

    private static string ReadAllTextProc(StreamReader reader)
    {
      return reader.ReadToEnd();
    }

    /*
     * GetRawHeader(), GetHeader()
     */
    public string GetRawHeader()
    {
      return GetRawHeader((IEnumerable<string>)EmptyArray<string>.Instance);
    }

    public string GetRawHeader(params string[] headerFields)
    {
      return GetRawHeader((IEnumerable<string>)headerFields);
    }

    public string GetRawHeader(IEnumerable<string> headerFields)
    {
      using (var stream = OpenReadHeaderFields(headerFields, ImapMessageFetchBodyOptions.Default)) {
        using (var reader = new StreamReader(stream, iso8859_1, false)) {
          return reader.ReadToEnd();
        }
      }
    }

    public KeyValuePair<string, string>[] GetHeader()
    {
      return GetHeader(false, (IEnumerable<string>)EmptyArray<string>.Instance);
    }

    public KeyValuePair<string, string>[] GetHeader(params string[] headerFields)
    {
      return GetHeader(false, (IEnumerable<string>)headerFields);
    }

    public KeyValuePair<string, string>[] GetHeader(bool decode, params string[] headerFields)
    {
      return GetHeader(decode, (IEnumerable<string>)headerFields);
    }

    public KeyValuePair<string, string>[] GetHeader(IEnumerable<string> headerFields)
    {
      return GetHeader(false, headerFields);
    }

    public KeyValuePair<string, string>[] GetHeader(bool decode, IEnumerable<string> headerFields)
    {
      using (var stream = OpenReadHeaderFields(headerFields, ImapMessageFetchBodyOptions.Default)) {
        var headers = Smdn.Formats.Mime.MimeUtils.ParseHeader(stream, true).ToArray();

        for (var i = 0; i < headers.Length; i++) {
          var value = headers[i].Value.Trim();

          if (decode)
            value = MimeEncoding.Decode(value);

          headers[i] = new KeyValuePair<string, string>(headers[i].Key, value);
        }

        return headers;
      }
    }
  }
}

