// 
// Author:
//       smdn <smdn@smdn.jp>
// 
// Copyright (c) 2008-2014 smdn
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights
// to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
// copies of the Software, and to permit persons to whom the Software is
// furnished to do so, subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
// THE SOFTWARE.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client {
  partial class ImapOpenedMailboxInfo {
    /*
     * GetMessages by FETCHing
     */
    public ImapMessageInfoList GetMessages()
    {
      return GetMessages(ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetMessages(ImapMessageFetchAttributeOptions options)
    {
      return new ImapMessageInfoList(this, options, new AllMessageQuery());
    }

    private class AllMessageQuery : IImapMessageQuery {
      public ImapSequenceSet GetSequenceOrUidSet(ImapOpenedMailboxInfo mailbox)
      {
        mailbox.Refresh();

        if (mailbox.ExistMessageCount <= 0L)
          return ImapSequenceSet.CreateSet(EmptyArray<long>.Instance);
        else
          return ImapSequenceSet.CreateRangeSet(1L, mailbox.ExistMessageCount);
      }
    }

    /*
     * GetMessage by FETCHing with uid or sequence
     */
    public ImapMessageInfo GetMessageByUid(long uid)
    {
      return GetMessageByUid(uid,
                             ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfo GetMessageByUid(long uid,
                                           ImapMessageFetchAttributeOptions options)
    {
      if (uid <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("uid", uid);

      /*
       * 6.4.8. UID Command
       *       A non-existent unique identifier is ignored without any error
       *       message generated.  Thus, it is possible for a UID FETCH command
       *       to return an OK without any data or a UID COPY or UID STORE to
       *       return an OK without performing any operations.
       */
      return GetMessage(ImapSequenceSet.CreateUidSet(uid),
                        options);
    }

    public ImapMessageInfo GetMessageBySequence(long sequence)
    {
      return GetMessageBySequence(sequence,
                                  ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfo GetMessageBySequence(long sequence,
                                                ImapMessageFetchAttributeOptions options)
    {
      if (sequence <= 0L)
        throw ExceptionUtils.CreateArgumentMustBeNonZeroPositive("sequence", sequence);

      // Refresh();

      if (ExistMessageCount < sequence)
        throw new ArgumentOutOfRangeException("sequence",
                                              sequence,
                                              string.Format("specified sequence number is greater than exist message count. (exist message count = {0})",
                                                             ExistMessageCount));

      return GetMessage(ImapSequenceSet.CreateSet(sequence),
                        options);
    }

    private ImapMessageInfo GetMessage(ImapSequenceSet sequenceOrUidSet,
                                       ImapMessageFetchAttributeOptions options)
    {
      var message = (new ImapMessageInfoList(this, options, sequenceOrUidSet)).FirstOrDefault();

      if (message == null)
        throw new ImapMessageNotFoundException(sequenceOrUidSet);

      return message;
    }

    public ImapMessageInfoList GetMessages(long uid, params long[] uids)
    {
      return GetMessages(ImapMessageFetchAttributeOptions.Default,
                         uid,
                         uids);
    }

    public ImapMessageInfoList GetMessages(ImapMessageFetchAttributeOptions options,
                                           long uid,
                                           params long[] uids)
    {
      return new ImapMessageInfoList(this,
                                     options,
                                     ImapSequenceSet.CreateUidSet(uid, uids));
    }

    public ImapMessageInfoList GetMessages(IEnumerable<long> uids)
    {
      return GetMessages(uids,
                         ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetMessages(IEnumerable<long> uids,
                                           ImapMessageFetchAttributeOptions options)
    {
      if (uids == null)
        throw new ArgumentNullException("uids");

      return new ImapMessageInfoList(this,
                                     options,
                                     ImapSequenceSet.CreateUidSet(uids));
    }

    /*
     * GetMessages by SEARCHing
     */
    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria)
    {
      return GetMessages(searchCriteria,
                         null,
                         ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria,
                                           ImapMessageFetchAttributeOptions options)
    {
      return GetMessages(searchCriteria,
                         null,
                         options);
    }

    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria,
                                           Encoding encoding)
    {
      return GetMessages(searchCriteria,
                         encoding,
                         ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetMessages(ImapSearchCriteria searchCriteria,
                                           Encoding encoding,
                                           ImapMessageFetchAttributeOptions options)
    {
      if (searchCriteria == null)
        throw new ArgumentNullException("searchCriteria");

      return new ImapMessageInfoList(this,
                                     options,
                                     new SearchMessageQuery(searchCriteria, encoding));
    }

    private abstract class SearchMessageQueryBase : IImapMessageQuery {
      public Encoding Encoding {
        get; private set;
      }

      protected SearchMessageQueryBase(Encoding encoding)
      {
        this.Encoding = encoding;
      }

      public abstract ImapSequenceSet GetSequenceOrUidSet(ImapOpenedMailboxInfo mailbox);

      protected bool ThrowIfBadCharset(ImapTaggedStatusResponse status)
      {
        // 6.4.4. SEARCH Command
        //       If the server does not support the specified [CHARSET], it MUST
        //       return a tagged NO response (not a BAD).  This response SHOULD
        //       contain the BADCHARSET response code, which MAY list the
        //       [CHARSET]s supported by the server.
        if (status.ResponseText.Code == ImapResponseCode.BadCharset) {
          var message = string.Format("charset '{0}' is not supported by server", Encoding.WebName);
          var supportedCharsets = ImapResponseTextConverter.FromBadCharset(status.ResponseText);

          if (0 < supportedCharsets.Length) {
            message += "; following charsets are supported: '" +
              string.Join("', '", supportedCharsets) +
                "'";
          }

          throw new Smdn.Formats.EncodingNotSupportedException(Encoding.WebName, message);
        }

        return true;
      }
    }

    private class SearchMessageQuery : SearchMessageQueryBase, IImapMessageQuery {
      public SearchMessageQuery(ImapSearchCriteria searchCriteria, Encoding encoding)
        : base(encoding)
      {
        this.searchCriteria = searchCriteria;
      }

      public override ImapSequenceSet GetSequenceOrUidSet(ImapOpenedMailboxInfo mailbox)
      {
        ImapMatchedSequenceSet matchedSequenceNumbers = null;

        if (mailbox.Client.IsCapabilityAvailable(ImapCapability.Searchres)) {
          mailbox.ProcessResult(mailbox.Client.Session.ESearch(searchCriteria,
                                                               Encoding,
                                                               ImapSearchResultOptions.Save,
                                                               out matchedSequenceNumbers),
                                delegate(ImapTaggedStatusResponse status) {
            if (status.ResponseText.Code == ImapResponseCode.NotSaved)
              return false; // throw no exception
            else
              return ThrowIfBadCharset(status);
          });

          if (matchedSequenceNumbers != null)
            return matchedSequenceNumbers;
        }

        mailbox.ProcessResult(mailbox.Client.Session.Search(searchCriteria,
                                                            Encoding,
                                                            out matchedSequenceNumbers),
                              ThrowIfBadCharset);

        return matchedSequenceNumbers;
      }

      private ImapSearchCriteria searchCriteria;
    }

    /*
     * GetMessages by SORTing
     */
    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder)
    {
      return GetSortedMessages(sortOrder,
                               ImapSearchCriteria.All,
                               null,
                               ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapMessageFetchAttributeOptions options)
    {
      return GetSortedMessages(sortOrder,
                               ImapSearchCriteria.All,
                               null,
                               options);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria)
    {
      return GetSortedMessages(sortOrder,
                               searchCriteria,
                               null,
                               ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria,
                                                 ImapMessageFetchAttributeOptions options)
    {
      return GetSortedMessages(sortOrder,
                               searchCriteria,
                               null,
                               options);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria,
                                                 Encoding encoding)
    {
      return GetSortedMessages(sortOrder,
                               searchCriteria,
                               encoding,
                               ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetSortedMessages(ImapSortCriteria sortOrder,
                                                 ImapSearchCriteria searchCriteria,
                                                 Encoding encoding,
                                                 ImapMessageFetchAttributeOptions options)
    {
      if (sortOrder == null)
        throw new ArgumentNullException("sortOrder");
      if (searchCriteria == null)
        throw new ArgumentNullException("searchCriteria");

      Client.ThrowIfCapabilityUnavailable(sortOrder);

      return new ImapMessageInfoList(this, options, true, new SortMessageQuery(sortOrder, searchCriteria, encoding));
    }

    private class SortMessageQuery : SearchMessageQueryBase, IImapMessageQuery {
      public SortMessageQuery(ImapSortCriteria sortOrder, ImapSearchCriteria searchCriteria, Encoding encoding)
        : base(encoding)
      {
        this.sortOrder = sortOrder;
        this.searchCriteria = searchCriteria;
      }

      public override ImapSequenceSet GetSequenceOrUidSet(ImapOpenedMailboxInfo mailbox)
      {
        ImapMatchedSequenceSet matchedSequenceNumbers;

        mailbox.ProcessResult(mailbox.Client.Session.Sort(sortOrder,
                                                          searchCriteria,
                                                          Encoding,
                                                          out matchedSequenceNumbers),
                              ThrowIfBadCharset);

        return matchedSequenceNumbers;
      }

      private ImapSortCriteria sortOrder;
      private ImapSearchCriteria searchCriteria;
    }

    /*
     * GetMessages by THREADing
     */
    public ImapMessageInfoThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 ImapSearchCriteria.All,
                                 null,
                                 ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                     ImapMessageFetchAttributeOptions options)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 ImapSearchCriteria.All,
                                 null,
                                 options);
    }

    public ImapMessageInfoThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                     ImapSearchCriteria searchCriteria)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 searchCriteria,
                                 null,
                                 ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                     ImapSearchCriteria searchCriteria,
                                                     Encoding encoding)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 searchCriteria,
                                 encoding,
                                 ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                     ImapSearchCriteria searchCriteria,
                                                     ImapMessageFetchAttributeOptions options)
    {
      return GetThreadedMessages(threadingAlgorithm,
                                 searchCriteria,
                                 null,
                                 options);
    }

    public ImapMessageInfoThread GetThreadedMessages(ImapThreadingAlgorithm threadingAlgorithm,
                                                     ImapSearchCriteria searchCriteria,
                                                     Encoding encoding,
                                                     ImapMessageFetchAttributeOptions options)
    {
      if (threadingAlgorithm == null)
        throw new ArgumentNullException("threadingAlgorithm");
      if (searchCriteria == null)
        throw new ArgumentNullException("searchCriteria");

      Client.ThrowIfCapabilityUnavailable(threadingAlgorithm);

      var query = new ThreadMessageQuery(threadingAlgorithm, searchCriteria, encoding);
      var list = new ImapMessageInfoList(this, options, query);
      var messages = list.ToDictionary(m => m.Sequence, m => m);

      return new ImapMessageInfoThread(query.Result, messages);
    }

    private class ThreadMessageQuery : SearchMessageQueryBase, IImapMessageQuery {
      public ImapThreadList Result {
        get { return result; }
      }

      public ThreadMessageQuery(ImapThreadingAlgorithm threadingAlgorithm, ImapSearchCriteria searchCriteria, Encoding encoding)
        : base(encoding)
      {
        this.threadingAlgorithm = threadingAlgorithm;
        this.searchCriteria = searchCriteria;
      }

      public override ImapSequenceSet GetSequenceOrUidSet(ImapOpenedMailboxInfo mailbox)
      {
        mailbox.ProcessResult(mailbox.Client.Session.Thread(threadingAlgorithm,
                                                            searchCriteria,
                                                            Encoding,
                                                            out result),
                              ThrowIfBadCharset);

        return result.ToSequenceSet();
      }

      private ImapThreadingAlgorithm threadingAlgorithm;
      private ImapSearchCriteria searchCriteria;
      private ImapThreadList result = null;
    }

    /*
     * GetMessages by SEARCHing with X-GM-RAW
     * https://developers.google.com/gmail/imap_extensions
     * https://support.google.com/mail/answer/7190
     */
    public ImapMessageInfoList GetMessagesGmailSearch(string gmailSearchQuery)
    {
      return GetMessagesGmailSearch(gmailSearchQuery,
                                    ImapMessageFetchAttributeOptions.Default);
    }

    public ImapMessageInfoList GetMessagesGmailSearch(string gmailSearchQuery,
                                                      ImapMessageFetchAttributeOptions options)
    {
      if (gmailSearchQuery == null)
        throw new ArgumentNullException("gmailSearchQuery");
      if (gmailSearchQuery.Length == 0)
        throw ExceptionUtils.CreateArgumentMustBeNonEmptyString("gmailSearchQuery");

      Client.ThrowIfCapabilityUnavailable(ImapCapability.GimapGmExt1);

      return GetMessages(ImapSearchCriteria.XGMRaw(gmailSearchQuery), Encoding.UTF8, options);
    }
  }
}

