using System;
using System.IO;
#if NET_4_0
using System.Threading;
#endif
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapFetchMessageBodyStreamTests : ImapSessionTestsBase {
    private class FetchStream : ImapFetchMessageBodyStream {
      public long BlockSize {
        get { return FetchBlockSize; }
      }

      public FetchStream(ImapSession session,
                         bool fetchPeek,
                         ImapSequenceSet fetchMessageUid,
                         int fetchBlockSize)
        : base(session, fetchPeek, fetchMessageUid, fetchBlockSize)
      {
      }

#if NET_4_0
      public FetchStream(ImapSession session,
                         bool fetchPeek,
                         ImapSequenceSet fetchMessageUid,
                         int fetchBlockSize,
                         CancellationToken cancellationToken)
        : base(session, fetchPeek, fetchMessageUid, null, null, fetchBlockSize, cancellationToken)
      {
      }
#endif

      public FetchStream(ImapSession session,
                         bool fetchPeek,
                         ImapSequenceSet fetchMessageUid,
                         string section,
                         int fetchBlockSize)
        : base(session, fetchPeek, fetchMessageUid, section, null, fetchBlockSize)
      {
      }

      public FetchStream(ImapSession session,
                         bool fetchPeek,
                         ImapSequenceSet fetchMessageUid,
                         ImapPartialRange range,
                         int fetchBlockSize)
        : base(session, fetchPeek, fetchMessageUid, null, range, fetchBlockSize)
      {
      }

      public FetchStream(ImapSession session,
                         bool fetchPeek,
                         ImapSequenceSet fetchMessageUid,
                         string section,
                         ImapPartialRange range,
                         int fetchBlockSize)
        : base(session, fetchPeek, fetchMessageUid, section, range, fetchBlockSize)
      {
      }

      public FetchStream(ImapSession session,
                         bool fetchPeek,
                         ImapSequenceSet fetchMessageUid,
                         string[] fetchHeaderFields,
                         int fetchBlockSize)
        : base(session, fetchPeek, fetchMessageUid, fetchHeaderFields, fetchBlockSize)
      {
      }

      public new ImapCommandResult Prepare(ImapFetchDataItem fetchDataItem, out IImapMessageAttribute messageAttr)
      {
        return base.Prepare(fetchDataItem, out messageAttr);
      }

      protected override Exception GetNoSuchMessageException(ImapCommandResult result)
      {
        return new NoSuchMessageException(result);
      }
    }

    private class NoSuchMessageException : Exception {
      public ImapCommandResult CausedResult {
        get;
        private set;
      }

      public NoSuchMessageException(ImapCommandResult causedResult)
      {
        this.CausedResult = causedResult;
      }
    }

    [Test]
    public void TestConstruct()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          Assert.AreEqual(16, stream.BlockSize, "BlockSize");

          Assert.IsTrue(stream.CanRead, "CanRead");
          Assert.IsFalse(stream.CanWrite, "CanWrite");
          Assert.IsFalse(stream.CanSeek, "CanSeek");
          Assert.IsTrue(stream.CanTimeout, "CanTimeout");

          Assert.AreEqual(-1L, stream.Length, "Length");

          Assert.AreEqual(session.ReceiveTimeout, stream.ReadTimeout, "ReadTimeout #1");
          Assert.AreEqual(session.SendTimeout, stream.WriteTimeout, "WriteTimeout #1");

          stream.ReadTimeout = stream.ReadTimeout + 1;
          stream.WriteTimeout = stream.WriteTimeout + 1;

          Assert.AreEqual(session.ReceiveTimeout, stream.ReadTimeout, "ReadTimeout #2");
          Assert.AreEqual(session.SendTimeout, stream.WriteTimeout, "WriteTimeout #2");
        }

        return 0;
      });
    }

    [Test]
    public void TestConstruct_SessionNull()
    {
      Assert.Throws<ArgumentNullException>(() => new FetchStream(null, false, ImapSequenceSet.CreateUidSet(1L), 16));
    }

    [Test]
    public void TestConstruct_FetchMessageUid_Null()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentNullException>(() => new FetchStream(session, false, (ImapSequenceSet)null, 16));

        return 0;
      });
    }

    [Test]
    public void TestConstruct_FetchMessageUid_NotSingle()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(() => new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L, 2L, 3L), 16));

        return 0;
      });
    }

    [Test]
    public void TestConstruct_FetchMessageUid_NotUid()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(() => new FetchStream(session, false, ImapSequenceSet.CreateSet(1L), 16));

        return 0;
      });
    }

    [Test]
    public void TestConstruct_FetchBlockSizeOutOfRange()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentOutOfRangeException>(() => new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), -1), "fetchBlockSize < 0");
        Assert.Throws<ArgumentOutOfRangeException>(() => new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 0), "fetchBlockSize = 0");

        return 0;
      });
    }

    [Test]
    public void TestWriteAndSeekOperations()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          Assert.Throws<NotSupportedException>(() => stream.WriteByte(0), "WriteByte()");

          var buffer = new byte[16];

          Assert.Throws<NotSupportedException>(() => stream.Write(buffer, 0, buffer.Length), "Write()");
          Assert.Throws<NotSupportedException>(() => stream.Flush(), "Flush()");

          Assert.Throws<NotSupportedException>(() => stream.SetLength(32L), "SetLength()");

          Assert.Throws<NotSupportedException>(() => stream.Position = 16L, "Position");
          Assert.Throws<NotSupportedException>(() => stream.Seek(0L, SeekOrigin.Begin), "Seek()");
        }

        return 0;
      });
    }

    [Test]
    public void TestClose()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16);

        Assert.DoesNotThrow(() => stream.Close(), "Close()");

        Assert.IsFalse(stream.CanRead, "CanRead");
        Assert.Throws<ObjectDisposedException>(() => Assert.AreNotEqual(0L, stream.Length), "Length");

        Assert.Throws<ObjectDisposedException>(() => stream.ReadByte(), "ReadByte()");

        var buffer = new byte[16];

        Assert.Throws<ObjectDisposedException>(() => stream.Read(buffer, 0, buffer.Length), "Read()");
        Assert.Throws<ObjectDisposedException>(() => stream.ReadToEnd(Stream.Null), "ReadToEnd(Stream)");
        Assert.Throws<ObjectDisposedException>(() => stream.ReadToEnd(BinaryWriter.Null), "ReadToEnd(BinaryWriter)");

        Assert.Throws<ObjectDisposedException>(() => Assert.AreEqual(0L, stream.Position), "Position");
        Assert.Throws<ObjectDisposedException>(() => Assert.AreEqual(0L, stream.Length), "Length");
        Assert.Throws<ObjectDisposedException>(() => stream.ReadTimeout = 0, "ReadTimeout");
        Assert.Throws<ObjectDisposedException>(() => stream.WriteTimeout = 0, "WriteTimeout");

        Assert.DoesNotThrow(() => stream.Close(), "Close() again");

        return 0;
      });
    }

    [Test]
    public void TestDispose()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16);

        using (stream) {
          Assert.IsTrue(stream.CanRead, "CanRead #1");
        }

        Assert.IsFalse(stream.CanRead, "CanRead #2");
        Assert.Throws<ObjectDisposedException>(() => Assert.AreNotEqual(0L, stream.Length), "Length");

        Assert.DoesNotThrow(() => stream.Close(), "Close()");

        Assert.Throws<ObjectDisposedException>(() => stream.ReadByte(), "ReadByte()");

        var buffer = new byte[16];

        Assert.Throws<ObjectDisposedException>(() => stream.Read(buffer, 0, buffer.Length), "Read()");
        Assert.Throws<ObjectDisposedException>(() => stream.ReadToEnd(Stream.Null), "ReadToEnd(Stream)");
        Assert.Throws<ObjectDisposedException>(() => stream.ReadToEnd(BinaryWriter.Null), "ReadToEnd(BinaryWriter)");

        return 0;
      });
    }

    [Test]
    public void TestRead()
    {
      Read(false);
    }

    [Test]
    public void TestRead_Peek()
    {
      Read(true);
    }

    private void Read(bool peek)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), 4)) {
          Assert.AreEqual(0L, stream.Position, "initial Position");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {4}\r\n0123)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[4];

          Assert.AreEqual(1, stream.Read(buffer, 0, 1), "Read #0");
          Assert.AreEqual(1L, stream.Position, "Position #0");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (BODY.PEEK[]<0.4>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.4>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(1, stream.Read(buffer, 1, 1), "Read #1");
          Assert.AreEqual(2L, stream.Position, "Position #1");
          Assert.AreEqual(1, stream.Read(buffer, 2, 1), "Read #2");
          Assert.AreEqual(3L, stream.Position, "Position #2");
          Assert.AreEqual(1, stream.Read(buffer, 3, 1), "Read #3");
          Assert.AreEqual(4L, stream.Position, "Position #3");

          Assert.AreEqual("0123",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer), "read data #0");

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<4> {3}\r\n456)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          Assert.AreEqual(1, stream.Read(buffer, 0, 1), "Read #4");
          Assert.AreEqual(5L, stream.Position, "Position #4");

          if (peek)
            Assert.AreEqual("0005 UID FETCH 1 (BODY.PEEK[]<4.4>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<4.4>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(1, stream.Read(buffer, 1, 1), "Read #5");
          Assert.AreEqual(6L, stream.Position, "Position #5");
          Assert.AreEqual(1, stream.Read(buffer, 2, 1), "Read #6");
          Assert.AreEqual(7L, stream.Position, "Position #6");
          Assert.AreEqual(0, stream.Read(buffer, 3, 1), "Read #7");
          Assert.AreEqual(7L, stream.Position, "Position #7");
          Assert.AreEqual(0, stream.Read(buffer, 3, 1), "Read #8");
          Assert.AreEqual(7L, stream.Position, "Position #8");

          Assert.AreEqual("4563",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer), "read data #1");
        }

        return 2;
      });
    }

    [Test]
    public void TestRead_SingleBlock_LessThanBlockSize()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {8}\r\n01234567)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[16];
          var len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(8, len, "Read() #1");
          Assert.AreEqual(8L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("01234567",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer, 0, len));

          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #2");
          Assert.AreEqual(8L, stream.Position, "Position #2");
          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #3");
          Assert.AreEqual(8L, stream.Position, "Position #3");
        }

        return 1;
      });
    }

    [Test]
    public void TestRead_SingleBlock_ExactBlockSize()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[16];
          var len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(16, len, "Read() #1");
          Assert.AreEqual(16L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("0123456789abcdef",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer, 0, len));

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {0}\r\n)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(0, len, "Read() #2");
          Assert.AreEqual(16L, stream.Position, "Position #2");

          Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                          server.DequeueRequest());

          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #3");
          Assert.AreEqual(16L, stream.Position, "Position #3");
          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #4");
          Assert.AreEqual(16L, stream.Position, "Position #4");
        }

        return 2;
      });
    }

    [Test]
    public void TestRead_MultipleBlock_EndWithBlockLessThanBlockSize()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[16];
          var len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(16, len, "Read() #1");
          Assert.AreEqual(16L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("0123456789abcdef",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer, 0, len));

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {1}\r\nX)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(1, len, "Read() #2");
          Assert.AreEqual(17L, stream.Position, "Position #2");

          Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("X",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer, 0, len));

          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #3");
          Assert.AreEqual(17L, stream.Position, "Position #3");
          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #4");
          Assert.AreEqual(17L, stream.Position, "Position #4");
        }

        return 2;
      });
    }

    [Test]
    public void TestRead_MultipleBlock_EndWithBlockExactBlockSize()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[16];
          var len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(16, len, "Read() #1");
          Assert.AreEqual(16L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("0123456789abcdef",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer, 0, len));

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {16}\r\n0123456789ABCDEF)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(16, len, "Read() #2");
          Assert.AreEqual(32L, stream.Position, "Position #2");

          Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                          server.DequeueRequest());
          Assert.AreEqual("0123456789ABCDEF",
                          NetworkTransferEncoding.Transfer7Bit.GetString(buffer, 0, len));

          // FETCH #3
          server.EnqueueResponse("* 1 FETCH (BODY[]<32> {0}\r\n)\r\n" +
                                 "0006 OK FETCH completed\r\n");

          len = stream.Read(buffer, 0, buffer.Length);

          Assert.AreEqual(0, len, "Read() #3");
          Assert.AreEqual(32L, stream.Position, "Position #3");

          Assert.AreEqual("0006 UID FETCH 1 (BODY[]<32.16>)\r\n",
                          server.DequeueRequest());

          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #4");
          Assert.AreEqual(32L, stream.Position, "Position #4");
          Assert.AreEqual(0, stream.Read(buffer, 0, buffer.Length), "Read() #5");
          Assert.AreEqual(32L, stream.Position, "Position #5");
        }

        return 3;
      });
    }

    [Test]
    public void TestRead_NoSuchMessage()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("0004 OK FETCH completed\r\n");

          var buffer = new byte[16];

          var ex = Assert.Throws<NoSuchMessageException>(() => stream.Read(buffer, 0, buffer.Length));

          Assert.IsTrue(ex.CausedResult.Succeeded);

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());
        }

        return 1;
      });
    }

#if NET_4_0
    [Test]
    public void TestRead_CancellationRequested()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var cts = new CancellationTokenSource();

        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 4, cts.Token)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {4}\r\n0123)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[4];

          Assert.AreEqual(1, stream.Read(buffer, 0, 1), "Read #0");
          Assert.AreEqual(1L, stream.Position, "Position #0");
          Assert.IsTrue(stream.CanRead, "CanRead #0");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.4>)\r\n",
                          server.DequeueRequest());

          cts.Cancel();

          Assert.Throws<OperationCanceledException>(() => stream.Read(buffer, 1, 1), "Read #1");

          Assert.AreEqual(1L, stream.Position, "Position #1");
          Assert.IsTrue(stream.CanRead, "CanRead #1");
        }

        return 1;
      });
    }
#endif

    [Test]
    public void TestPrepare()
    {
      Prepare(false);
    }

    [Test]
    public void TestPrepare_Peek()
    {
      Prepare(true);
    }

    private void Prepare(bool peek)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), 8)) {
          Assert.AreEqual(-1L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (RFC822.SIZE 12 BODY[]<0> {8}\r\n01234567)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(null, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(12L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (RFC822.SIZE BODY.PEEK[]<0.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (RFC822.SIZE BODY[]<0.8>)\r\n",
                            server.DequeueRequest());

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #0");
          Assert.AreEqual(8L, stream.Position, "Position #0");
          Assert.AreEqual(12L, stream.Length, "Length #0");

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<8> {4}\r\n89ab)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          Assert.AreEqual(4, stream.Read(buffer, 0, 8), "Read #1");
          Assert.AreEqual(12L, stream.Position, "Position #1");
          Assert.AreEqual(12L, stream.Length, "Length #1");

          if (peek)
            Assert.AreEqual("0005 UID FETCH 1 (BODY.PEEK[]<8.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<8.8>)\r\n",
                            server.DequeueRequest());
        }

        return 2;
      });
    }

    [Test]
    public void TestPrepare_WithExtraFetchDataItem()
    {
      Prepare_WithExtraFetchDataItem(ImapFetchDataItem.Rfc822Size + ImapFetchDataItem.InternalDate + ImapFetchDataItem.Uid);
    }

    [Test]
    public void TestPrepare_WithExtraFetchDataItem_ContainsNoRfc822Size()
    {
      Prepare_WithExtraFetchDataItem(ImapFetchDataItem.InternalDate + ImapFetchDataItem.Uid);
    }

    private void Prepare_WithExtraFetchDataItem(ImapFetchDataItem extraFetchDataItem)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 8)) {
          Assert.AreEqual(-1L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (RFC822.SIZE 12 " + 
                                 "INTERNALDATE \"19-Feb-2008 08:15:48 +0900\" " +
                                 "UID 1 " +
                                 "BODY[]<0> {8}\r\n01234567)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(extraFetchDataItem, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(12L, stream.Length, "Length after prepare");

          Assert.AreEqual("0004 UID FETCH 1 (RFC822.SIZE INTERNALDATE UID BODY[]<0.8>)\r\n",
                          server.DequeueRequest());

          Assert.AreEqual(1, messageAttr.Uid, "Uid");
          Assert.AreEqual(new DateTimeOffset(2008, 2, 19, 8, 15, 48, new TimeSpan(+9, 0, 0)),
                          messageAttr.InternalDate, "InternalDate");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #0");
          Assert.AreEqual(8L, stream.Position, "Position #0");
          Assert.AreEqual(12L, stream.Length, "Length #0");

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<8> {4}\r\n89ab)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          Assert.AreEqual(4, stream.Read(buffer, 0, 8), "Read #1");
          Assert.AreEqual(12L, stream.Position, "Position #1");
          Assert.AreEqual(12L, stream.Length, "Length #1");

          Assert.AreEqual("0005 UID FETCH 1 (BODY[]<8.8>)\r\n",
                          server.DequeueRequest());
        }

        return 2;
      });
    }

#if NET_4_0
    [Test]
    public void TestPrepare_CancellationRequested()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var cts = new CancellationTokenSource();

        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 8, cts.Token)) {
          cts.Cancel();

          IImapMessageAttribute messageAttr;

          Assert.Throws<OperationCanceledException>(() => stream.Prepare(null, out messageAttr));
        }

        return 0;
      });
    }
#endif

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithSection()
    {
      PrepareAndRead_PartialFetch_WithSection(false);
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithSection_Peek()
    {
      PrepareAndRead_PartialFetch_WithSection(true);
    }

    private void PrepareAndRead_PartialFetch_WithSection(bool peek)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), "1.1", 8)) {
          Assert.AreEqual(0L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (" +
                                 "UID 1 " +
                                 "BODY[1.1]<0> {8}\r\n01234567)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(ImapFetchDataItem.Uid, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(0L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY.PEEK[1.1]<0.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY[1.1]<0.8>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(1, messageAttr.Uid, "Uid");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #0");
          Assert.AreEqual(8L, stream.Position, "Position #0");

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[1.1]<8> {4}\r\n89ab)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          Assert.AreEqual(4, stream.Read(buffer, 0, 8), "Read #1");
          Assert.AreEqual(12L, stream.Position, "Position #1");

          if (peek)
            Assert.AreEqual("0005 UID FETCH 1 (BODY.PEEK[1.1]<8.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0005 UID FETCH 1 (BODY[1.1]<8.8>)\r\n",
                            server.DequeueRequest());
        }

        return 2;
      });
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithRangeLengthSpecified()
    {
      PrepareAndRead_PartialFetch_WithRangeLengthSpecified(false);
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithRangeLengthSpecified_Peek()
    {
      PrepareAndRead_PartialFetch_WithRangeLengthSpecified(true);
    }

    private void PrepareAndRead_PartialFetch_WithRangeLengthSpecified(bool peek)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var range = new ImapPartialRange(4L, 16L); // length > fetchBlockSize

        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), range, 8)) {
          Assert.AreEqual(16L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (" +
                                 "UID 1 " +
                                 "BODY[]<4> {8}\r\n01234567)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(ImapFetchDataItem.Uid, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(16L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY.PEEK[]<4.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY[]<4.8>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(1, messageAttr.Uid, "Uid");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #0");
          Assert.AreEqual(8L, stream.Position, "Position #0");

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<12> {8}\r\n89abcdef)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #1");
          Assert.AreEqual(16L, stream.Position, "Position #1");

          if (peek)
            Assert.AreEqual("0005 UID FETCH 1 (BODY.PEEK[]<12.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<12.8>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(0, stream.Read(buffer, 0, 8), "Read #3");
        }

        return 2;
      });
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithRangeLengthNotSpecified()
    {
      PrepareAndRead_PartialFetch_WithRangeLengthNotSpecified(false);
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithRangeLengthNotSpecified_Peek()
    {
      PrepareAndRead_PartialFetch_WithRangeLengthNotSpecified(true);
    }

    private void PrepareAndRead_PartialFetch_WithRangeLengthNotSpecified(bool peek)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var range = new ImapPartialRange(4L);

        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), range, 8)) {
          Assert.AreEqual(0L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (" +
                                 "UID 1 " +
                                 "BODY[]<4> {8}\r\n01234567)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(ImapFetchDataItem.Uid, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(0L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY.PEEK[]<4.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY[]<4.8>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(1, messageAttr.Uid, "Uid");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #0");
          Assert.AreEqual(8L, stream.Position, "Position #0");

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<12> {8}\r\n89abcdef)\r\n" +
                                 "0005 OK FETCH completed\r\n");

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read #1");
          Assert.AreEqual(16L, stream.Position, "Position #1");

          if (peek)
            Assert.AreEqual("0005 UID FETCH 1 (BODY.PEEK[]<12.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<12.8>)\r\n",
                            server.DequeueRequest());

          // FETCH #3
          server.EnqueueResponse("* 1 FETCH (BODY[]<20> {0}\r\n)\r\n" +
                                 "0006 OK FETCH completed\r\n");

          Assert.AreEqual(0, stream.Read(buffer, 0, 8), "Read #2");

          if (peek)
            Assert.AreEqual("0006 UID FETCH 1 (BODY.PEEK[]<20.8>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0006 UID FETCH 1 (BODY[]<20.8>)\r\n",
                            server.DequeueRequest());
        }

        return 3;
      });
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithSectionAndRange()
    {
      PrepareAndRead_PartialFetch_WithSectionAndRange(false);
    }

    [Test]
    public void TestPrepareAndRead_PartialFetch_WithSectionAndRange_Peek()
    {
      PrepareAndRead_PartialFetch_WithSectionAndRange(true);
    }

    private void PrepareAndRead_PartialFetch_WithSectionAndRange(bool peek)
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        var range = new ImapPartialRange(4L, 4L); // length < fetchBlockSize

        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), "1.2.3", range, 16)) {
          Assert.AreEqual(4L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (" +
                                 "UID 1 " +
                                 "BODY[1.2.3]<4> {4}\r\n0123)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(ImapFetchDataItem.Uid, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(4L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY.PEEK[1.2.3]<4.4>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (UID BODY[1.2.3]<4.4>)\r\n",
                            server.DequeueRequest());

          Assert.AreEqual(1, messageAttr.Uid, "Uid");

          var buffer = new byte[4];

          Assert.AreEqual(4, stream.Read(buffer, 0, 4), "Read #0");
          Assert.AreEqual(4L, stream.Position, "Position #0");

          Assert.AreEqual(0, stream.Read(buffer, 0, 4), "Read #2");
        }

        return 1;
      });
    }

    [Test]
    public void TestPrepareAndRead_FetchHeader_FetchHeaderFieldsEmpty()
    {
      PrepareAndRead_FetchHeader(false, EmptyArray<string>.Instance);
    }

    [Test]
    public void TestPrepareAndRead_FetchHeader_FetchHeaderFieldsEmpty_Peek()
    {
      PrepareAndRead_FetchHeader(true, EmptyArray<string>.Instance);
    }

    [Test]
    public void TestPrepareAndRead_FetchHeader_FetchHeaderFieldsNull()
    {
      PrepareAndRead_FetchHeader(false, null);
    }

    [Test]
    public void TestPrepareAndRead_FetchHeader_FetchHeaderFieldsNull_Peek()
    {
      PrepareAndRead_FetchHeader(true, null);
    }

    private void PrepareAndRead_FetchHeader(bool peek, string[] fetchHeaderFields)
    {
      const string messageHeader = @"Return-Path: <test@example.net>
X-Original-To: test@example.net
Delivered-To: test@example.net
Message-ID: <4BAA561E.5000004@test@example.net>
Date: Thu, 25 Mar 2010 03:12:46 +0900
From: test <test@example.net>
Reply-To: test@example.net
User-Agent: Thunderbird 2.0.0.24 (X11/20100317)
MIME-Version: 1.0
To: hoge@example.net
Subject: =?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=
Content-Type: text/plain; charset=ISO-2022-JP
Content-Transfer-Encoding: 7bit

";

      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        const int fetchBlockSize = 128;
        var fetchOffset = 0;
        var fetchCommandCount = 0;

        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), fetchHeaderFields, fetchBlockSize)) {
          Assert.AreEqual(0L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (" +
                                 string.Format("BODY[HEADER]<0> {{{0}}}\r\n{1})\r\n",
                                               fetchBlockSize,
                                               messageHeader.Substring(fetchOffset, fetchBlockSize)) +
                                 "0004 OK FETCH completed\r\n");

          fetchOffset += fetchBlockSize;
          fetchCommandCount++;

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(null, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(0L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual(string.Format("0004 UID FETCH 1 (BODY.PEEK[HEADER]<0.{0}>)\r\n", fetchBlockSize),
                            server.DequeueRequest());
          else
            Assert.AreEqual(string.Format("0004 UID FETCH 1 (BODY[HEADER]<0.{0}>)\r\n", fetchBlockSize),
                            server.DequeueRequest());

          for (;;) {
            var fetchLength = Math.Min(fetchBlockSize, messageHeader.Length - fetchOffset);

            // FETCH
            server.EnqueueResponse("* 1 FETCH (" +
                                   string.Format("BODY[HEADER]<{0}> {{{1}}}\r\n{2})\r\n",
                                               fetchOffset,
                                               fetchLength,
                                               messageHeader.Substring(fetchOffset, fetchLength)) +
                                   string.Format("{0:x4} OK FETCH completed\r\n", fetchCommandCount + 4));

            fetchOffset += fetchLength;
            fetchCommandCount++;

            if (fetchOffset == messageHeader.Length) {
              if (fetchLength == fetchBlockSize) {
                server.EnqueueResponse("* 1 FETCH (" +
                                       string.Format("BODY[HEADER]<{0}> {{0}}\r\n)\r\n", fetchOffset) +
                                       string.Format("{0:x4} OK FETCH completed\r\n", fetchCommandCount + 4));
                fetchCommandCount++;
              }

              break;
            }
          }

          var reader = new StreamReader(stream, NetworkTransferEncoding.Transfer7Bit);

          Assert.AreEqual(messageHeader,
                          reader.ReadToEnd());

          for (var i = 1; i < fetchCommandCount; i++) {
            if (peek)
              Assert.AreEqual(string.Format("{0:x4} UID FETCH 1 (BODY.PEEK[HEADER]<{1}.{2}>)\r\n",
                                            4 + i,
                                            i * fetchBlockSize,
                                            fetchBlockSize),
                              server.DequeueRequest());
            else
              Assert.AreEqual(string.Format("{0:x4} UID FETCH 1 (BODY[HEADER]<{1}.{2}>)\r\n",
                                            4 + i,
                                            i * fetchBlockSize,
                                            fetchBlockSize),
                              server.DequeueRequest());
          }
        }

        return fetchCommandCount;
      });
    }

    [Test]
    public void TestPrepareAndRead_FetchHeader_HeaderFieldsSpecified()
    {
      PrepareAndRead_FetchHeader_HeaderFieldsSpecified(false);
    }

    [Test]
    public void TestPrepareAndRead_FetchHeader_HeaderFieldsSpecified_Peek()
    {
      PrepareAndRead_FetchHeader_HeaderFieldsSpecified(true);
    }

    private void PrepareAndRead_FetchHeader_HeaderFieldsSpecified(bool peek)
    {
      var messageHeader = @"From: root@example.jp (Cron Daemon)
To: root@example.jp
Subject: Cron <root@example.jp> test.sh
Date: Fri,  7 Mar 2014 17:00:26 +0900 (JST)

";
      var headerFields = new[] {"From", "To", "Subject", "Date"};

      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, peek, ImapSequenceSet.CreateUidSet(1L), headerFields, 1024)) {
          Assert.AreEqual(0L, stream.Length, "Length before prepare");

          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (" +
                                 string.Format("BODY[HEADER.FIELDS (FROM TO SUBJECT DATE)]<0> {{{0}}}\r\n{1})\r\n",
                                               NetworkTransferEncoding.Transfer7Bit.GetByteCount(messageHeader),
                                               messageHeader) +
                                 "0004 OK FETCH completed\r\n");

          IImapMessageAttribute messageAttr;

          var ret = stream.Prepare(null, out messageAttr);

          Assert.IsTrue(ret.Succeeded, "Succeeded");
          Assert.AreEqual(0L, stream.Length, "Length after prepare");

          if (peek)
            Assert.AreEqual("0004 UID FETCH 1 (BODY.PEEK[HEADER.FIELDS (From To Subject Date)]<0.1024>)\r\n",
                            server.DequeueRequest());
          else
            Assert.AreEqual("0004 UID FETCH 1 (BODY[HEADER.FIELDS (From To Subject Date)]<0.1024>)\r\n",
                            server.DequeueRequest());

          var reader = new StreamReader(stream, NetworkTransferEncoding.Transfer7Bit);

          Assert.AreEqual(messageHeader,
                          reader.ReadToEnd());
        }

        return 1;
      });
    }

    [Test]
    public void TestReadToEnd_Stream_FromBegin()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");
          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {16}\r\n0123456789abcdef)\r\n" +
                                 "0005 OK FETCH completed\r\n");
          // FETCH #3
          server.EnqueueResponse("* 1 FETCH (BODY[]<32> {0}\r\n)\r\n" +
                                 "0006 OK FETCH completed\r\n");

          using (var destination = new MemoryStream()) {
            stream.ReadToEnd(destination);

            Assert.AreEqual(32L, stream.Position, "Position");
            Assert.AreEqual(32L, destination.Length, "destination.Length");
            Assert.AreEqual(32L, destination.Position, "destination.Position");

            destination.Close();

            Assert.AreEqual("0123456789abcdef0123456789abcdef",
                            NetworkTransferEncoding.Transfer7Bit.GetString(destination.ToArray()));

            Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());
            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                          server.DequeueRequest());
            Assert.AreEqual("0006 UID FETCH 1 (BODY[]<32.16>)\r\n",
                          server.DequeueRequest());
          }
        }

        return 3;
      });
    }

    [Test]
    public void TestReadToEnd_Stream_BufferedDataExists()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read() #1");
          Assert.AreEqual(8L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {16}\r\n0123456789abcdef)\r\n" +
                                 "0005 OK FETCH completed\r\n");
          // FETCH #3
          server.EnqueueResponse("* 1 FETCH (BODY[]<32> {0}\r\n)\r\n" +
                                 "0006 OK FETCH completed\r\n");

          using (var destination = new MemoryStream()) {
            stream.ReadToEnd(destination);

            Assert.AreEqual(32L, stream.Position, "Position #2");
            Assert.AreEqual(24L, destination.Length, "destination.Length");
            Assert.AreEqual(24L, destination.Position, "destination.Position");

            destination.Close();

            Assert.AreEqual("89abcdef0123456789abcdef",
                            NetworkTransferEncoding.Transfer7Bit.GetString(destination.ToArray()));

            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                            server.DequeueRequest());
            Assert.AreEqual("0006 UID FETCH 1 (BODY[]<32.16>)\r\n",
                            server.DequeueRequest());
          }
        }

        return 3;
      });
    }

    [Test]
    public void TestReadToEnd_BinaryWriter_FromBegin()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read() #1");
          Assert.AreEqual(8L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {16}\r\n0123456789abcdef)\r\n" +
                                 "0005 OK FETCH completed\r\n");
          // FETCH #3
          server.EnqueueResponse("* 1 FETCH (BODY[]<32> {0}\r\n)\r\n" +
                                 "0006 OK FETCH completed\r\n");

          using (var destination = new MemoryStream()) {
            var writer = new BinaryWriter(destination);

            stream.ReadToEnd(writer);

            Assert.AreEqual(32L, stream.Position, "Position #2");

            writer.Flush();

            Assert.AreEqual(24L, destination.Length, "destination.Length");
            Assert.AreEqual(24L, destination.Position, "destination.Position");

            writer.Close();

            Assert.AreEqual("89abcdef0123456789abcdef",
                            NetworkTransferEncoding.Transfer7Bit.GetString(destination.ToArray()));

            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                            server.DequeueRequest());
            Assert.AreEqual("0006 UID FETCH 1 (BODY[]<32.16>)\r\n",
                            server.DequeueRequest());
          }
        }

        return 3;
      });
    }

    [Test]
    public void TestReadToEnd_BinaryWriter_BufferedDataExists()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        using (var stream = new FetchStream(session, false, ImapSequenceSet.CreateUidSet(1L), 16)) {
          // FETCH #1
          server.EnqueueResponse("* 1 FETCH (BODY[]<0> {16}\r\n0123456789abcdef)\r\n" +
                                 "0004 OK FETCH completed\r\n");

          var buffer = new byte[8];

          Assert.AreEqual(8, stream.Read(buffer, 0, 8), "Read() #1");
          Assert.AreEqual(8L, stream.Position, "Position #1");

          Assert.AreEqual("0004 UID FETCH 1 (BODY[]<0.16>)\r\n",
                          server.DequeueRequest());

          // FETCH #2
          server.EnqueueResponse("* 1 FETCH (BODY[]<16> {16}\r\n0123456789abcdef)\r\n" +
                                 "0005 OK FETCH completed\r\n");
          // FETCH #3
          server.EnqueueResponse("* 1 FETCH (BODY[]<32> {0}\r\n)\r\n" +
                                 "0006 OK FETCH completed\r\n");

          using (var destination = new MemoryStream()) {
            var writer = new BinaryWriter(destination);

            stream.ReadToEnd(writer);

            Assert.AreEqual(32L, stream.Position, "Position #2");

            writer.Flush();

            Assert.AreEqual(24L, destination.Length, "destination.Length");
            Assert.AreEqual(24L, destination.Position, "destination.Position");

            writer.Close();

            Assert.AreEqual("89abcdef0123456789abcdef",
                            NetworkTransferEncoding.Transfer7Bit.GetString(destination.ToArray()));

            Assert.AreEqual("0005 UID FETCH 1 (BODY[]<16.16>)\r\n",
                            server.DequeueRequest());
            Assert.AreEqual("0006 UID FETCH 1 (BODY[]<32.16>)\r\n",
                            server.DequeueRequest());
          }
        }

        return 3;
      });
    }
  }
}