using System;
using System.IO;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading;
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionTests : ImapSessionTestsBase {
    [Test]
    public void TestReadOnlyProperty_ChangeCollection_ServerCapabilities()
    {
      Connect2(delegate(ImapSession session) {
        Assert.Throws<NotSupportedException>(() => session.ServerCapabilities.Add(ImapCapability.Imap4Rev1));
      });
    }

    [Test]
    public void TestReadOnlyProperty_ChangeCollection_HierarchyDelimiters()
    {
      Connect2(delegate(ImapSession session) {
#if NET_4_5
        Assert.Pass("read only interface");
#else
        Assert.Throws<NotSupportedException>(() => session.HierarchyDelimiters.Add("#refname", "."));
#endif
      });
    }

    [Test]
    public void TestReadOnlyProperty_ChangeCollection_ServerID()
    {
      Connect2(delegate(ImapSession session) {
#if NET_4_5
        Assert.Pass("read only interface");
#else
        Assert.Throws<NotSupportedException>(() => session.ServerID.Add("key", "value"));
#endif
      });
    }

    [Test]
    public void TestReceiveTimeout_SetZero()
    {
      Connect2(delegate(ImapSession session) {
        session.ReceiveTimeout = 0;

        Assert.AreEqual(0, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestReceiveTimeout_SetInfinite()
    {
      Connect2(delegate(ImapSession session) {
        session.ReceiveTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.ReceiveTimeout);
      });
    }

    [Test]
    public void TestSendTimeout_SetZero()
    {
      Connect2(delegate(ImapSession session) {
        session.SendTimeout = 0;

        Assert.AreEqual(0, session.SendTimeout);
      });
    }

    [Test]
    public void TestSendTimeout_SetInfinite()
    {
      Connect2(delegate(ImapSession session) {
        session.SendTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.SendTimeout);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetZero()
    {
      Connect2(delegate(ImapSession session) {
        session.TransactionTimeout = 0;

        Assert.AreEqual(0, session.TransactionTimeout);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetInfinite()
    {
      Connect2(delegate(ImapSession session) {
        session.TransactionTimeout = Timeout.Infinite;

        Assert.AreEqual(Timeout.Infinite, session.TransactionTimeout);
      });
    }

    private void Idle(Action<ImapSession> preprocessSession, Action<ImapSession> processSession)
    {
      Authenticate2(new[] {"IDLE"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        preprocessSession(session);

        server.EnqueueResponse("IDLE\r\n",
                               "+ idling\r\n");

        var asyncResult = session.BeginIdle();

        Assert.IsTrue(session.IsIdling);
        Assert.IsTrue(session.IsTransactionProceeding);

        processSession(session);

        server.EnqueueTaggedResponse("$tag OK done\r\n");

        Assert.IsTrue((bool)session.EndIdle(asyncResult));

        Assert.IsFalse(session.IsIdling);
        Assert.IsFalse(session.IsTransactionProceeding);
      });
    }

    [Test]
    public void TestTransactionTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.TransactionTimeout = 1000;
        Assert.AreEqual(1000, session.TransactionTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.TransactionTimeout = 1000);
      });
    }

    [Test]
    public void TestSendTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.SendTimeout = 1000;
        Assert.AreEqual(1000, session.SendTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.SendTimeout = 1000);
      });
    }

    [Test]
    public void TestReceiveTimeout_SetWhileTransactionProceeding()
    {
      Idle(delegate(ImapSession session) {
        session.ReceiveTimeout = 1000;
        Assert.AreEqual(1000, session.ReceiveTimeout);
      },
      delegate(ImapSession session) {
        Assert.Throws<InvalidOperationException>(() => session.ReceiveTimeout = 1000);
      });
    }

    [Test]
    public void TestBeginNewTransaction_WhileTransactionProceeding_1()
    {
      Idle(delegate(ImapSession session) {
        Assert.IsFalse(session.IsTransactionProceeding);
      },
      delegate(ImapSession session) {
        Assert.IsTrue(session.IsTransactionProceeding);
        Assert.Throws<InvalidOperationException>(() => session.NoOp());
      });
    }

    [Test]
    public void TestBeginNewTransaction_WhileTransactionProceeding_2()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        using (var transactionStartedEvent = new ManualResetEvent(false)) {
          server.EnqueueTaggedResponse("NOOP\r\n",
                                       "$tag ", // incomplete response
                                       null,
                                       () => transactionStartedEvent.Set());

          Exception unexpectedException = null;

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            try {
              (state as ImapSession).NoOp();
            }
            catch (Exception ex) {
              unexpectedException = ex;
            }
          }, session);

          transactionStartedEvent.WaitOne();

          Assert.IsTrue(session.IsTransactionProceeding);

          Assert.Throws<InvalidOperationException>(() => session.NoOp());

          server.EnqueueResponse("OK \r\n");

          if (unexpectedException != null)
            Assert.Fail("unexpected exception {0}", unexpectedException);
        }
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        session.HandlesIncapableAsException = false;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Binary));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapString[] {
          new ImapLiteralString("literal1", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
          new ImapLiteralString("literal2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
          new ImapLiteralString("literal3", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
          new ImapLiteralString("literal4", Encoding.ASCII, ImapLiteralOptions.Literal),
          new ImapLiteralString("literal5", Encoding.ASCII, ImapLiteralOptions.Literal8),
          new ImapLiteralString("literal6", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
          new ImapLiteralString("literal7", Encoding.ASCII, ImapLiteralOptions.Synchronizing | ImapLiteralOptions.Literal8IfCapable),
          new ImapLiteralString("literal8", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable | ImapLiteralOptions.Literal8),
          new ImapParenthesizedString(new ImapLiteralString("nested1", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
                                      new ImapLiteralString("nested2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
                                      new ImapLiteralString("nested3", Encoding.ASCII, ImapLiteralOptions.Literal8),
                                      new ImapLiteralString("nested4", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable)),
        };

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST " +
                        "{8}\r\nliteral1 " +
                        "{8+}\r\nliteral2 " +
                        "{8}\r\nliteral3 " +
                        "{8}\r\nliteral4 " +
                        "~{8}\r\nliteral5 " +
                        "{8}\r\nliteral6 " +
                        "{8}\r\nliteral7 " +
                        "~{8}\r\nliteral8 " +
                        "({7+}\r\nnested1 " +
                        "{7}\r\nnested2 " +
                        "~{7}\r\nnested3 " +
                        "{7}\r\nnested4)\r\n",
                        server.DequeueRequest());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_LiteralNonSyncCapable()
    {
      Connect2(new[] {"LITERAL+"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        session.HandlesIncapableAsException = true;

        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapParenthesizedString(new ImapString[] {
          new ImapLiteralString("literal1", Encoding.ASCII, ImapLiteralOptions.Synchronizing),
          new ImapLiteralString("literal2", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing),
          new ImapLiteralString("literal3", Encoding.ASCII, ImapLiteralOptions.NonSynchronizingIfCapable),
        });

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST (" +
                        "{8}\r\nliteral1 " +
                        "{8+}\r\nliteral2 " +
                        "{8+}\r\nliteral3)\r\n",
                        server.DequeueAll());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_LiteralNonSyncIncapable()
    {
      Connect2(delegate(ImapSession session) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.LiteralNonSync));

        Assert.Throws<ImapIncapableException>(delegate {
          session.GenericCommand("X-TEST", new ImapLiteralString("literal", Encoding.ASCII, ImapLiteralOptions.NonSynchronizing));
        });
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_Literal8Capable()
    {
      Connect2(new[] {"BINARY"}, delegate(ImapSession session, ImapPseudoServer2 server) {
        session.HandlesIncapableAsException = true;

        Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Binary));

        server.EnqueueTaggedResponse("X-TEST ",
                                     "$tag OK done\r\n");

        var arguments = new ImapParenthesizedString(new ImapString[] {
          new ImapLiteralString("literal1", Encoding.ASCII, ImapLiteralOptions.Literal),
          new ImapLiteralString("literal2", Encoding.ASCII, ImapLiteralOptions.Literal8),
          new ImapLiteralString("literal3", Encoding.ASCII, ImapLiteralOptions.Literal8IfCapable),
        });

        session.GenericCommand("X-TEST", arguments);

        Assert.AreEqual("0000 X-TEST (" +
                        "{8}\r\nliteral1 " +
                        "~{8}\r\nliteral2 " +
                        "~{8}\r\nliteral3)\r\n",
                        server.DequeueAll());
      });
    }

    [Test]
    public void TestPreProcessTransaction_SetLiteralOptions_Literal8Incapable()
    {
      Connect2(delegate(ImapSession session) {
        session.HandlesIncapableAsException = true;

        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Binary));

        Assert.Throws<ImapIncapableException>(delegate {
          session.GenericCommand("X-TEST", new ImapLiteralString("literal", Encoding.ASCII, ImapLiteralOptions.Literal8));
        });
      });
    }

    [Test]
    public void TestConnect()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        Assert.IsFalse(session.IsIdling);
        Assert.IsNull(session.SelectedMailbox);
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
        Assert.AreEqual(new Uri(string.Format("imap://{0}:{1}/", server.Host, server.Port)), session.Authority);
        Assert.AreEqual(0, session.ServerCapabilities.Count);
      });
    }

    [Test]
    public void TestConnect_Bye()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* BYE not available.\r\n",
                                   null, null, true);

        try {
          using (var session = new ImapSession(server.Host, server.Port)) {
            Assert.Fail("ImapConnectionException not thrown");
          }
        }
        catch (ImapConnectionException ex) {
          Assert.IsNull(ex.InnerException);
          StringAssert.Contains("not available.", ex.Message);
        }
      }
    }

    [Test]
    public void TestConnect_InvalidResponse()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("+OK pseudo POP server ready.\r\n");

        try {
          using (var session = new ImapSession(server.Host, server.Port)) {
            Assert.Fail("ImapConnectionException not thrown");
          }
        }
        catch (ImapConnectionException ex) {
          Assert.IsNull(ex.InnerException);
          StringAssert.Contains("unexpected data response", ex.Message);
        }
      }
    }

    [Test]
    public void TestConnect_TransactionTimeout()
    {
      using (var server = ImapPseudoServer2.StartNotAccepting()) {
        try {
          using (var session = new ImapSession(server.Host, server.Port, 500)) {
            Assert.Fail("ImapConnectionException not thrown");
          }
        }
        catch (ImapConnectionException ex) {
          var timeoutException = ex.InnerException as TimeoutException;

          Assert.IsNotNull(timeoutException);
        }
      }
    }

    [Test]
    public void TestConnect_ConnectionTimeout()
    {
      try {
        using (var session = new ImapSession("imap.localhost", 10143, 1000)) {
          Assert.Fail("ImapConnectionException not thrown");
        }
      }
      catch (ImapConnectionException ex) {
        var timeoutException = ex.InnerException as TimeoutException;

        Assert.IsNotNull(timeoutException);
      }
    }

    [Test]
    public void TestConnect_SocketError()
    {
      try {
        using (var session = new ImapSession("imap.invalid", 10143)) {
          Assert.Fail("ImapConnectionException not thrown");
        }
      }
      catch (ImapConnectionException ex) {
        var socketException = ex.InnerException as System.Net.Sockets.SocketException;

        Assert.IsNotNull(socketException);
      }
    }

    [Test]
    public void TestConnect_Ssl()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var streamUpgraded = false;

        using (var session = new ImapSession(server.Host, server.Port, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SslStream
        })) {
          Assert.IsFalse(session.IsIdling);
          Assert.IsNull(session.SelectedMailbox);
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(new Uri(string.Format("imaps://{0}:{1}/", server.Host, server.Port)), session.Authority);
          Assert.AreEqual(0, session.ServerCapabilities.Count);
          Assert.IsTrue(streamUpgraded, "stream upgraded");
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestConnect_Ssl_ExceptionWhileUpgrading()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        try {
          using (var session = new ImapSession(server.Host, server.Port, delegate(ConnectionBase connection, Stream baseStream) {
            throw new System.Security.Authentication.AuthenticationException();
          })) {
            Assert.Fail("connected");
          }
        }
        catch (ImapSecureConnectionException ex) {
          var upgradeException = ex.InnerException as ImapUpgradeConnectionException;

          Assert.IsNotNull(upgradeException);
          Assert.IsNotNull(upgradeException.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(upgradeException.InnerException);
        }
      }
    }

    [Test]
    public void TestDisconnect()
    {
      Connect2(delegate(ImapSession session) {
        session.Disconnect(false);

        Assert.IsTrue(session.IsDisposed);

        try {
          session.NoOp();
          Assert.Fail("ObjectDisposedException not thrown");
        }
        catch (ObjectDisposedException) {
        }
      });
    }

    [Test]
    public void TestImap4Rev1Incapable()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY NONE\r\n" +
                                     "$tag OK ImapCapability completed\r\n");

        ImapCapabilitySet caps;

        Assert.IsTrue((bool)session.Capability(out caps));
        Assert.IsNotNull(caps);
        Assert.IsFalse(caps.Contains(ImapCapability.Imap4Rev1));

        try {
          session.HandlesIncapableAsException = true;
          session.NoOp();
          Assert.Fail("exception not occured");
        }
        catch (ImapIncapableException) {
        }
      });
    }

    [Test]
    public void TestGreeting_WithCapability()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 AUTH=CRAM-MD5 AUTH=PLAIN CHILDREN THREAD=REFERENCES X-EXTENSION1 X-EXTENSION2] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(7, session.ServerCapabilities.Count);

          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Children));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.CRAMMD5));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

          Assert.IsTrue(session.ServerCapabilities.Contains("THREAD=REFERENCES"));
          Assert.IsTrue(session.ServerCapabilities.Contains("X-EXTENSION1"));
          Assert.IsTrue(session.ServerCapabilities.Contains("X-EXTENSION2"));

          try {
            session.ServerCapabilities.Add(ImapCapability.Imap4Rev1);
            Assert.Fail("NotSupportedException not thrown");
          }
          catch (NotSupportedException) {
          }
        }
      }
    }

    [Test]
    public void TestGreeting_WithNoCapability()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* OK ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);
          Assert.AreEqual(0, session.ServerCapabilities.Count);

          Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

          try {
            session.ServerCapabilities.Add(ImapCapability.Imap4Rev1);
            Assert.Fail("NotSupportedException not thrown");
          }
          catch (NotSupportedException) {
          }
        }
      }
    }

    [Test]
    public void TestGreeting_PreAuth_WithCapability()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* PREAUTH [CAPABILITY IMAP4rev1 AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(2, session.ServerCapabilities.Count);
          Assert.AreEqual(new Uri(string.Format("imap://{0}:{1}/", server.Host, server.Port)), session.Authority);

          Assert.IsTrue(session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));
          Assert.IsTrue(session.ServerCapabilities.IsCapable(ImapAuthenticationMechanism.Plain));

          try {
            session.ServerCapabilities.Add(ImapCapability.Imap4Rev1);
            Assert.Fail("NotSupportedException not thrown");
          }
          catch (NotSupportedException) {
          }
        }
      }
    }

    [Test]
    public void TestGreeting_WithNoImapCapability()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* OK [CAPABILITY AUTH=PLAIN] ImapSimulatedServer ready\r\n");

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.IsTrue(!session.ServerCapabilities.Contains(ImapCapability.Imap4Rev1));

          session.HandlesIncapableAsException = true;

          Assert.Throws<ImapIncapableException>(() => session.Login(new NetworkCredential("user", "pass")));
        }
      }
    }

    [Test]
    public void TestGreeting_HomeServerReferral_AsException()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n",
                                   null, null, true);

        try {
          using (var session = new ImapSession(server.Host, server.Port, true)) {
            Assert.Fail("connected without exception");
          }
        }
        catch (ImapLoginReferralException ex) {
          Assert.IsTrue(ex.Message.Contains("Server not accepting connections.  Try SERVER2"));
          Assert.AreEqual(new Uri("IMAP://user;AUTH=*@SERVER2/"), ex.ReferToUri);
        }
      }
    }

    [Test]
    public void TestGreeting_HomeServerReferral_AsError()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* BYE [REFERRAL IMAP://user;AUTH=*@SERVER2/] Server not accepting connections.  Try SERVER2\r\n",
                                   null, null, true);

        Assert.Throws<ImapConnectionException>(delegate {
          using (var session = new ImapSession(server.Host, server.Port)) {
          }
        });
      }
    }

    [Test]
    public void TestTransactionException_TransactionTimeout()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        session.TransactionTimeout = 250;
        session.SendTimeout = System.Threading.Timeout.Infinite;
        session.ReceiveTimeout = System.Threading.Timeout.Infinite;

        server.EnqueueTaggedResponse("NOOP\r\n",
                                     "$tag OK"); // not respond NOOP response termination(CRLF)

        try {
          session.NoOp();
          Assert.Fail("TimeoutException not thrown");
        }
        catch (TimeoutException) {
          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        }

        try {
          session.NoOp();
          Assert.Fail("ImapProtocolViolationException not thrown");
        }
        catch (ImapProtocolViolationException) {
        }
      });
    }

    [Test]
    public void TestTransactionException_SocketTimeout()
    {
      Connect2(delegate(ImapSession session) {
        session.TransactionTimeout = System.Threading.Timeout.Infinite;
        session.SendTimeout = 250;
        session.ReceiveTimeout = 250;

        // not respond NOOP
        //server.EnqueueResponse("0000 OK\r\n");

        try {
          session.NoOp();
          Assert.Fail("TimeoutException not thrown");
        }
        catch (TimeoutException) {
          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        }

        try {
          session.NoOp();
          Assert.Fail("ImapProtocolViolationException not thrown");
        }
        catch (ImapProtocolViolationException) {
        }
      });
    }

    [Test]
    public void TestTransactionException_InternalError()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "*\r\n" + // invalid response
                                     "$tag CAPABILITY Ok\r\n");

        try {
          session.Capability();
          Assert.Fail("ImapException not thrown");
        }
        catch (ImapException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        }

        try {
          session.NoOp();
          Assert.Fail("ImapProtocolViolationException not thrown");
        }
        catch (ImapProtocolViolationException) {
        }
      });
    }

    [Test]
    public void TestDiconnectedFromServer_WithByeResponse()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        Assert.AreEqual(ImapSessionState.NotAuthenticated, session.State);

        server.EnqueueUntaggedByeResponse("* BYE Disconnected for inactivity.\r\n");

        Assert.IsTrue(session.NoOp().Code == ImapCommandResultCode.Bye);

        Assert.AreEqual(ImapSessionState.NotConnected, session.State);
        Assert.AreEqual(null, session.Authority);

        try {
          session.NoOp();
          Assert.Fail("ImapProtocolViolationException not thrown");
        }
        catch (ImapProtocolViolationException) {
        }
      });
    }

    [Test]
    public void TestDiconnectedFromServer_WithoutByeResponse()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* OK [ALERT] will soon shutdown",
                                   null, null, true);

        using (var session = new ImapSession(server.Host, server.Port)) {
          Assert.Throws<ImapConnectionException>(() => session.NoOp());

          Assert.AreEqual(ImapSessionState.NotConnected, session.State);
          Assert.AreEqual(null, session.Authority);

          Assert.Throws<ImapProtocolViolationException>(() => session.NoOp());
        }
      }
    }

    [Test]
    public void TestReceivedCommandTagUnmatched()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueResponse("LOGIN",
                               "0001 OK authenticated\r\n");

        try {
          session.Login(new NetworkCredential("user", "pass"));
          Assert.Fail("ImapException not thrown");
        }
        catch (ImapException ex) {
          Assert.IsNotNull(ex.InnerException);
          Assert.IsInstanceOf<ImapMalformedResponseException>(ex.InnerException);
        }
      });
    }

    [Test]
    public void TestListStatusSelect()
    {
      Connect2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK authenticated\r\n");

        Assert.IsTrue((bool)session.Login(new NetworkCredential("user", "pass")));

        var expectedAuthority = new Uri(string.Format("imap://user@{0}/", server.HostPort));

        Assert.AreEqual(expectedAuthority, session.Authority);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK CAPABILITY completed\r\n");

        session.Capability();

        // LIST transaction
        server.EnqueueTaggedResponse("LIST \"\" *\r\n",
                                     "* LIST (\\HasChildren) \".\" \"INBOX\"\r\n" +
                                     "$tag OK LIST completed\r\n");

        ImapMailbox[] mailboxes;

        Assert.IsTrue((bool)session.List("*", out mailboxes));

        Assert.AreEqual(1, mailboxes.Length);
        Assert.AreEqual("INBOX", mailboxes[0].Name);
        Assert.AreEqual(".", mailboxes[0].HierarchyDelimiter);
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.HasChildren));
        Assert.AreEqual(new Uri(expectedAuthority, "./INBOX"), mailboxes[0].Url);

        // LSUB transaction
        server.EnqueueTaggedResponse("LSUB \"\" *\r\n",
                                     "* LSUB (\\HasChildren) \"/\" \"Trash\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \"/\" \"Trash/mail\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \".\" \"INBOX.&ZeVnLIqe-\"\r\n" +
                                     "* LSUB (\\HasNoChildren) \".\" \"INBOX.&U,BTFw-\"\r\n" +
                                     "$tag OK LSUB completed\r\n");

        Assert.IsTrue((bool)session.Lsub("*", out mailboxes));

        Assert.AreEqual(4, mailboxes.Length);
        Assert.AreEqual("Trash", mailboxes[0].Name);
        Assert.IsTrue(mailboxes[0].Flags.Contains(ImapMailboxFlag.HasChildren));

        Assert.AreEqual("Trash/mail", mailboxes[1].Name);
        Assert.IsTrue(mailboxes[1].Flags.Contains(ImapMailboxFlag.HasNoChildren));
        Assert.AreEqual(new Uri(expectedAuthority, "./Trash/mail"), mailboxes[1].Url);

        Assert.AreEqual("INBOX.日本語", mailboxes[2].Name);
        Assert.AreEqual("INBOX.台北", mailboxes[3].Name);

        // STATUS transaction
        server.EnqueueTaggedResponse("STATUS Trash (MESSAGES RECENT UIDNEXT UIDVALIDITY UNSEEN)\r\n",
                                     "* STATUS \"Trash\" (MESSAGES 3 RECENT 1 UIDVALIDITY 123456 UNSEEN 2 UIDNEXT 4)\r\n" +
                                     "$tag OK STATUS completed\r\n");

        Assert.IsTrue((bool)session.Status(mailboxes[0], ImapStatusDataItem.StandardAll));

        Assert.AreEqual(3, mailboxes[0].ExistsMessage);
        Assert.AreEqual(1, mailboxes[0].RecentMessage);
        Assert.AreEqual(123456, mailboxes[0].UidValidity);
        Assert.AreEqual(2, mailboxes[0].UnseenMessage);
        Assert.AreEqual(4, mailboxes[0].UidNext);

        // SELECT transaction
        Assert.IsNull(session.SelectedMailbox);

        server.EnqueueTaggedResponse("SELECT INBOX.&ZeVnLIqe-\r\n",
                                     "* FLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 custom2)\r\n" +
                                     "* OK [PERMANENTFLAGS (\\Answered \\Flagged \\Deleted \\Seen \\Draft custom1 \\*)] Flags permitted\r\n" +
                                     "* 15 EXISTS\r\n" +
                                     "* 2 RECENT\r\n" +
                                     "* OK [UIDVALIDITY 1202674433] UIDs valid\r\n" +
                                     "* OK [UIDNEXT 16]\r\n" +
                                     "* OK [UNSEEN 13]\r\n" +
                                     "$tag OK [READ-WRITE] SELECT completed\r\n");

        var mailbox = mailboxes[2];

        Assert.IsTrue((bool)session.Select(mailbox));

        Assert.AreEqual(ImapSessionState.Selected, session.State);
        Assert.AreSame(mailbox, session.SelectedMailbox);

        Assert.AreEqual("INBOX.日本語", mailbox.Name);
        Assert.AreEqual(15, mailbox.ExistsMessage);
        Assert.AreEqual(2, mailbox.RecentMessage);
        Assert.AreEqual(0, mailbox.UnseenMessage);
        Assert.AreEqual(13, mailbox.FirstUnseen);
        Assert.AreEqual(1202674433, mailbox.UidValidity);
        Assert.AreEqual(16, mailbox.UidNext);
        Assert.IsFalse(mailbox.ReadOnly);
        Assert.IsTrue(mailbox.UidPersistent);
        Assert.AreEqual(new Uri(expectedAuthority, "./INBOX.日本語;UIDVALIDITY=1202674433"), mailbox.Url);

        /*
        Assert.AreEqual(selectedMailbox.Name, mailbox.Name);
        Assert.AreEqual(selectedMailbox.ExistsMessage, mailbox.ExistsMessage);
        Assert.AreEqual(selectedMailbox.RecentMessage, mailbox.RecentMessage);
        Assert.AreEqual(selectedMailbox.UnseenMessage, mailbox.UnseenMessage);
        Assert.AreEqual(selectedMailbox.FirstUnseen, mailbox.FirstUnseen);
        Assert.AreEqual(selectedMailbox.UidValidity, mailbox.UidValidity);
        Assert.AreEqual(selectedMailbox.UidNext, mailbox.UidNext);
        Assert.AreEqual(selectedMailbox.ReadOnly, mailbox.ReadOnly);
        Assert.AreEqual(selectedMailbox.ApplicableFlags.Count, mailbox.ApplicableFlags.Count);
        Assert.AreEqual(selectedMailbox.PermanentFlags.Count, mailbox.PermanentFlags.Count);
        */

        Assert.AreEqual(7, mailbox.ApplicableFlags.Count);
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("custom1"));
        Assert.IsTrue(mailbox.ApplicableFlags.Contains("custom2"));

        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.AllowedCreateKeywords));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Answered));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(mailbox.PermanentFlags.Contains(ImapMessageFlag.Draft));
        Assert.IsTrue(mailbox.PermanentFlags.Contains("custom1"));

        // CLOSE transaction
        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK CLOSE completed\r\n");

        Assert.IsTrue((bool)session.Close());

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      });
    }

    [Test]
    public void TestSerializeBinary_CommandResultCodeOk()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        server.EnqueueResponse("CAPABILITY\r\n",
                               "* CAPABILITY IMAP4rev1\r\n" +
                               "0002 OK CAPABILITY completed\r\n");

        var result = session.Capability();

        Smdn.Net.TestUtils.SerializeBinary(result, delegate(ImapCommandResult deserialized) {
          Assert.AreEqual(ImapCommandResultCode.Ok, deserialized.Code);
          Assert.IsNotNull(deserialized.TaggedStatusResponse);
          Assert.AreEqual("0002", deserialized.TaggedStatusResponse.Tag);
          Assert.AreEqual(ImapResponseCondition.Ok, deserialized.TaggedStatusResponse.Condition);
          Assert.IsNull(deserialized.TaggedStatusResponse.ResponseText.Code);
          Assert.AreEqual("CAPABILITY completed", deserialized.ResponseText);
          Assert.IsNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(2, responses.Length);
          Assert.IsInstanceOf<ImapDataResponse>(responses[0]);
          Assert.IsInstanceOf<ImapTaggedStatusResponse>(responses[1]);
        });
      });
    }

    [Test]
    public void TestSerializeBinary_CommandResultCodeRequestDone()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        var result = session.Login(new NetworkCredential("user", "pass")); // already authenticated

        Smdn.Net.TestUtils.SerializeBinary(result, delegate(ImapCommandResult deserialized) {
          Assert.AreEqual(ImapCommandResultCode.RequestDone, deserialized.Code);
          Assert.IsNull(deserialized.TaggedStatusResponse);
          Assert.IsNull(deserialized.ResponseText);
          Assert.IsNotNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(0, responses.Length);
        });
      });
    }

    class NullCredential : ICredentialsByHost {
      public NetworkCredential GetCredential(string host, int port, string authType)
      {
        return null;
      }
    }

    [Test]
    public void TestSerializeBinary_CommandResultCodeRequestError()
    {
      Connect2(delegate(ImapSession session) {
        var result = session.Login(new NullCredential());

        Smdn.Net.TestUtils.SerializeBinary(result, delegate(ImapCommandResult deserialized) {
          Assert.AreEqual(ImapCommandResultCode.RequestError, deserialized.Code);
          Assert.IsNull(deserialized.TaggedStatusResponse);
          Assert.IsNull(deserialized.ResponseText);
          Assert.IsNotNull(deserialized.Description);

          var responses = deserialized.ReceivedResponses.ToArray();

          Assert.AreEqual(0, responses.Length);
        });
      });
    }

    [Test]
    public void TestGenericCommand_WithoutArguments()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        // generic command transaction
        server.EnqueueTaggedResponse("X-EXT-COMMAND\r\n",
                                     "* X-EXT-COMMAND data\r\n" +
                                     "$tag OK X-EXT-COMMAND completed\r\n");

        ImapDataResponse[] dataResps;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps));

        Assert.AreEqual(1, dataResps.Length);
        Assert.AreEqual(2, dataResps[0].Data.Length);
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[0].Format);
        Assert.AreEqual("X-EXT-COMMAND", dataResps[0].Data[0].GetTextAsString());
        Assert.AreEqual(ImapDataFormat.Text, dataResps[0].Data[1].Format);
        Assert.AreEqual("data", dataResps[0].Data[1].GetTextAsString());
      });
    }

    [Test]
    public void TestGenericCommand_WithArguments()
    {
      Authenticate2(delegate(ImapSession session, ImapPseudoServer2 server) {
        // generic command transaction
        server.EnqueueTaggedResponse("X-EXT-COMMAND ",
                                     "$tag OK X-EXT-COMMAND completed\r\n");

        ImapDataResponse[] dataResps;

        Assert.IsTrue((bool)session.GenericCommand("X-EXT-COMMAND", out dataResps,
                                           new ImapParenthesizedString("arg1",
                                                                       "arg2",
                                                                       new ImapParenthesizedString("arg3-1", "arg3-2"),
                                                                       new ImapLiteralString("ascii text", NetworkTransferEncoding.Transfer8Bit)),
                                           new ImapParenthesizedString(new ImapQuotedString("quoted"),
                                                                       "nonquoted")
                                           ));

        Assert.AreEqual("0002 X-EXT-COMMAND (arg1 arg2 (arg3-1 arg3-2) {10}\r\n" +
                        "ascii text) (\"quoted\" nonquoted)\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(0, dataResps.Length);
      });
    }
  }
}
