using System;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionCommandsSelectedStateTests : ImapSessionTestsBase {
    [Test]
    public void TestClose_InNonAuthenticatedState()
    {
      Connect(delegate(ImapSession session) {
        Assert.Throws<ImapProtocolViolationException>(() => session.Close());
      });
    }

    [Test]
    public void TestClose_InAuthenticatedState()
    {
      Authenticate(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);

        var result = session.Close();

        Assert.IsTrue((bool)result);
        Assert.AreEqual(result.Code, ImapCommandResultCode.RequestDone);

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);
      });
    }

    [Test]
    public void TestCheck_StatusUpdate()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // CHECK transaction
        server.EnqueueResponse("* 23 EXISTS\r\n" +
                               "* 1 RECENT\r\n" +
                               "0004 OK CHECK completed\r\n");

        Assert.IsTrue((bool)session.Check());

        Assert.AreEqual("0004 CHECK\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(23L, session.SelectedMailbox.ExistsMessage, "selected mailbox exist message count");
        Assert.AreEqual(1L, session.SelectedMailbox.RecentMessage, "selected mailbox recent message count");

        return 1;
      });
    }

    [Test]
    public void TestUnselect()
    {
      SelectMailbox(new[] {"UNSELECT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // EXPUNGE transaction
        server.EnqueueResponse("0004 OK UNSELECT completed\r\n");

        Assert.IsTrue((bool)session.Unselect());

        Assert.AreEqual("0004 UNSELECT\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(ImapSessionState.Authenticated, session.State);
        Assert.IsNull(session.SelectedMailbox);

        return -1; // no need to CLOSE
      });
    }

    [Test]
    public void TestSearch()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Unseen, out matched));

        Assert.AreEqual("0004 SEARCH UNSEEN\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);

        var messages = matched.ToArray();

        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(2,    messages[0]);
        Assert.AreEqual(84,   messages[1]);
        Assert.AreEqual(882,  messages[2]);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.AreEqual("2,84,882", d.ToString());
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_WithLiteral_DefaultCharset()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Subject("subject"), out matched));

        Assert.AreEqual("0004 SEARCH CHARSET us-ascii SUBJECT {7}\r\n" +
                        "subject\r\n",
                        server.DequeueAll());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("2,84,882", d.ToString());
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_WithLiteral_SpecifiedCharset()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Subject("件名"), Encoding.GetEncoding(932), out matched));

        Assert.AreEqual("0004 SEARCH CHARSET shift_jis SUBJECT {4}\r\n" +
                        "\x008c\x008f\x0096\x00bc\r\n",
                        server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("2,84,882", d.ToString());
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_NothingMatched()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SEARCH\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.Text("string not in mailbox"), Encoding.UTF8, out matched));

        Assert.AreEqual("0004 SEARCH CHARSET utf-8 TEXT {21}\r\n" +
                        "string not in mailbox\r\n",
                        server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsTrue(d.IsEmpty);
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_WithCriteriaFromUri()
    {
      SelectMailbox(new[] {"LITERAL+"}, delegate(ImapSession session, ImapPseudoServer server) {
        var b = new ImapUriBuilder(session.SelectedMailbox.Url);

        b.SearchCriteria = ImapSearchCriteria.From("差出人") & (ImapSearchCriteria.Seen | ImapSearchCriteria.Subject("件名"));
        b.Charset = Encoding.UTF8;

        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 84 882\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.FromUri(b.Uri),
                                           out matched));

        Assert.AreEqual("0004 SEARCH CHARSET utf-8 FROM {9+}\r\n" +
                                   "\x00E5\x00B7\x00AE\x00E5\x0087\x00BA\x00E4\x00BA\x00BA OR (SEEN) (SUBJECT {6+}\r\n" +
                                   "\x00E4\x00BB\x00B6\x00E5\x0090\x008D)\r\n",
                                   server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_BadCharset()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("0004 NO [BADCHARSET (ONDUL EUC-JP)] Unknown charset\r\n");

        ImapMatchedSequenceSet matched;

        var result = session.Search(ImapSearchCriteria.Body("ほげほげ"), Encoding.GetEncoding("Shift_JIS"), out matched);

        Assert.IsFalse((bool)result);
        Assert.AreEqual(ImapResponseCode.BadCharset, result.TaggedStatusResponse.ResponseText.Code);

        var supportedCharsets = ImapResponseTextConverter.FromBadCharset(result.TaggedStatusResponse.ResponseText);

        Assert.AreEqual(2, supportedCharsets.Length);
        Assert.AreEqual("ONDUL", supportedCharsets[0]);
        Assert.AreEqual("EUC-JP", supportedCharsets[1]);

        Assert.AreEqual("0004 SEARCH CHARSET shift_jis BODY {8}\r\n",
                        server.DequeueRequest());
        server.DequeueRequest();

        return 1;
      });
    }

    [Test]
    public void TestSearch_CondStore_ModSeq()
    {
      SelectMailbox(new[] {"CONDSTORE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 5 6 (MODSEQ 917162500)\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.ModSeq(620162338UL), out matched));

        Assert.AreEqual("0004 SEARCH MODSEQ 620162338\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        var messages = matched.ToArray();

        Assert.AreEqual(3, messages.Length);
        Assert.AreEqual(2, messages[0]);
        Assert.AreEqual(5, messages[1]);
        Assert.AreEqual(6, messages[2]);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.AreEqual("2,5,6", d.ToString());
          Assert.AreEqual(917162500UL, d.HighestModSeq);
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_CondStore_ModSeq_WithEntry()
    {
      SelectMailbox(new[] {"CONDSTORE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 2 5 6 7 11 12 18 19 20 23 (MODSEQ 917162500)\r\n" +
                               "0004 OK SEARCH completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Search(ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft), out matched));

        Assert.AreEqual("0004 SEARCH MODSEQ \"/flags/\\\\Draft\" all 620162338\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.AreEqual(917162500UL, matched.HighestModSeq);

        var messages = matched.ToArray();

        Assert.AreEqual(10, messages.Length);
        Assert.AreEqual(2, messages[0]);
        Assert.AreEqual(5, messages[1]);
        Assert.AreEqual(6, messages[2]);
        Assert.AreEqual(23, messages[9]);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.AreEqual("2,5,6,7,11,12,18,19,20,23", d.ToString());
          Assert.AreEqual(917162500UL, d.HighestModSeq);
        });

        return 1;
      });
    }

    [Test]
    public void TestSearch_CondStore_Incapable()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        ImapMatchedSequenceSet matched;

        Assert.Throws<ImapIncapableException>(delegate {
          session.Search(ImapSearchCriteria.ModSeqAllEntry(620162338UL, ImapMessageFlag.Draft), out matched);
        });

        return -1;
      });
    }

    [Test]
    public void TestESearch()
    {
      SelectMailbox(new[] {"ESEARCH"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* ESEARCH (TAG \"0004\") MIN 2 COUNT 3\r\n" +
                               "0004 OK SEARCH completed\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Min + ImapSearchResultOptions.Count, out matched));

        Assert.AreEqual("0004 SEARCH RETURN (MIN COUNT) CHARSET us-ascii FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);
        Assert.IsNotNull(matched.Tag);
        Assert.AreEqual("0004", matched.Tag);
        Assert.IsNotNull(matched.Min);
        Assert.AreEqual(2, matched.Min);
        Assert.IsNotNull(matched.Count);
        Assert.AreEqual(3, matched.Count);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsTrue(d.IsEmpty);
          Assert.IsNotNull(d.Tag);
          Assert.AreEqual("0004", d.Tag);
          Assert.IsNotNull(d.Min);
          Assert.AreEqual(2, d.Min);
          Assert.IsNotNull(d.Count);
          Assert.AreEqual(3, d.Count);
        });

        return 1;
      });
    }

    [Test]
    public void TestESearch_Incapable()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        ImapMatchedSequenceSet matched;

        Assert.Throws<ImapIncapableException>(delegate {
          session.ESearch(ImapSearchCriteria.All, ImapSearchResultOptions.Min, out matched);
        });

        return -1;
      });
    }

    [Test]
    public void TestESearch_Searchres()
    {
      SelectMailbox(new[] {"ESEARCH", "SEARCHRES"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("0004 OK SEARCH completed, result saved\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));

        Assert.AreEqual("0004 SEARCH RETURN (SAVE) CHARSET us-ascii FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.IsTrue(d.IsSavedResult);
        });

        // FETCH transaction
        server.EnqueueResponse("* 2 FETCH (UID 14)\r\n" +
                               "* 84 FETCH (UID 100)\r\n" +
                               "* 882 FETCH (UID 1115)\r\n" +
                               "0005 OK completed\r\n");

        ImapMessage[] fetched;

        Assert.IsTrue((bool)session.Fetch(matched, ImapFetchDataItem.Uid, out fetched));

        Assert.AreEqual("0005 FETCH $ (UID)\r\n",
                        server.DequeueRequest());

        return 2;
      });
    }

    [Test]
    public void TestESearch_Searchres_WithSavedResult()
    {
      SelectMailbox(new[] {"ESEARCH", "SEARCHRES"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("0004 OK SEARCH completed, result saved\r\n");

        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));

        Assert.AreEqual("0004 SEARCH RETURN (SAVE) CHARSET us-ascii FLAGGED SINCE 1-Feb-1994 NOT (FROM {5}\r\nSmith)\r\n",
                                   server.DequeueAll());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);
        Assert.IsTrue(matched.IsSavedResult);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.IsTrue(d.IsSavedResult);
        });

        // SEARCH transaction
        server.EnqueueResponse("* SEARCH 17 900 901\r\n" +
                               "0005 OK completed\r\n");

        ImapMatchedSequenceSet matched2;

        Assert.IsTrue((bool)session.UidSearch(ImapSearchCriteria.Uid(matched) & ImapSearchCriteria.Smaller(4096), out matched2));

        Assert.AreEqual("0005 UID SEARCH UID $ SMALLER 4096\r\n",
                        server.DequeueRequest());

        Assert.IsTrue(matched2.IsUidSet);
        Assert.IsFalse(matched2.IsEmpty);

        var arr = matched2.ToArray();

        Assert.AreEqual(3, arr.Length);
        Assert.AreEqual(17, arr[0]);
        Assert.AreEqual(900, arr[1]);
        Assert.AreEqual(901, arr[2]);

        Smdn.Net.TestUtils.SerializeBinary(matched2, delegate(ImapMatchedSequenceSet d) {
          Assert.IsTrue(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("17,900,901", d.ToString());
        });

        return 2;
      });
    }

    [Test]
    public void TestESearch_Searchres_Incapable()
    {
      SelectMailbox(new[] {"ESEARCH"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SEARCH transaction
        var criteria = ImapSearchCriteria.Flagged &
          ImapSearchCriteria.Since(new DateTime(1994, 2, 1)) &
          !(ImapSearchCriteria.From("Smith"));

        ImapMatchedSequenceSet matched;

        Assert.Throws<ImapIncapableException>(() => session.ESearch(criteria, ImapSearchResultOptions.Save, out matched));

        return -1;
      });
    }

    [Test]
    public void TestSort()
    {
      SelectMailbox(new[] {"SORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SORT transaction
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All, out matched));

        Assert.AreEqual("0004 SORT (REVERSE SIZE DATE) utf-8 ALL\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);

        var messages = matched.ToArray();

        Assert.AreEqual(5, messages.Length);
        Assert.AreEqual(5, messages[0]);
        Assert.AreEqual(3, messages[1]);
        Assert.AreEqual(4, messages[2]);
        Assert.AreEqual(1, messages[3]);
        Assert.AreEqual(2, messages[4]);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
          Assert.AreEqual("5,3,4,1,2", d.ToString());
        });

        return 1;
      });
    }

    [Test]
    public void TestSort_WithLiteral_DefaultCharset()
    {
      SelectMailbox(new[] {"SORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SORT transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date,
                                         ImapSearchCriteria.Subject("件名"),
                                         out matched));

        Assert.AreEqual("0004 SORT (REVERSE SIZE DATE) utf-8 SUBJECT {6}\r\n" +
                        "件名\r\n",
                        server.DequeueAll());

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
        });

        return 1;
      });
    }

    [Test]
    public void TestSort_WithLiteral_SpecifiedCharset()
    {
      SelectMailbox(new[] {"SORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SORT transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* SORT 5 3 4 1 2\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date,
                                         ImapSearchCriteria.Subject("件名"),
                                         Encoding.GetEncoding(932),
                                         out matched));

        Assert.AreEqual("0004 SORT (REVERSE SIZE DATE) shift_jis SUBJECT {4}\r\n" +
                        "\x008c\x008f\x0096\x00bc\r\n",
                        server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(matched);
        Assert.IsFalse(matched.IsUidSet);
        Assert.IsFalse(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsFalse(d.IsEmpty);
        });

        return 1;
      });
    }

    [Test]
    public void TestSort_NothingMatched()
    {
      SelectMailbox(new[] {"SORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SORT transaction
        server.EnqueueResponse("* SORT\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.Sort(ImapSortCriteria.Date, ImapSearchCriteria.All, out matched));

        Assert.AreEqual("0004 SORT (DATE) utf-8 ALL\r\n",
                        server.DequeueRequest());

        Assert.IsFalse(matched.IsUidSet);
        Assert.IsTrue(matched.IsEmpty);

        Smdn.Net.TestUtils.SerializeBinary(matched, delegate(ImapMatchedSequenceSet d) {
          Assert.IsFalse(d.IsUidSet);
          Assert.IsTrue(d.IsEmpty);
        });

        return 1;
      });
    }

    [Test]
    public void TestSort_Incapable()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        ImapMatchedSequenceSet matched;

        Assert.Throws<ImapIncapableException>(delegate {
          session.Sort(ImapSortCriteria.SizeReverse + ImapSortCriteria.Date, ImapSearchCriteria.All, out matched);
        });

        return -1;
      });
    }

    [Test]
    public void TestESort()
    {
      SelectMailbox(new[] {"SORT", "ESORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SORT transaction
        server.EnqueueResponse("* ESEARCH (TAG \"0004\") UID ALL 23765,23764,23763,23761\r\n" +
                               "0004 OK SORT completed\r\n");

        ImapMatchedSequenceSet matched;

        Assert.IsTrue((bool)session.UidESort(ImapSortCriteria.DateReverse,
                                       ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                                       ImapSearchResultOptions.All,
                                       out matched));

        Assert.AreEqual("0004 UID SORT RETURN (ALL) (REVERSE DATE) utf-8 UNDELETED UNKEYWORD $Junk\r\n",
                        server.DequeueRequest());

        Assert.IsTrue(matched.IsUidSet);
        Assert.AreEqual("0004", matched.Tag);

        var messages = matched.ToArray();

        Assert.AreEqual(4, messages.Length);
        Assert.AreEqual(23765, messages[0]);
        Assert.AreEqual(23764, messages[1]);
        Assert.AreEqual(23763, messages[2]);
        Assert.AreEqual(23761, messages[3]);

        return 1;
      });
    }

    [Test]
    public void TestESort_Incapable()
    {
      SelectMailbox(new[] {"SORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        // SORT transaction
        ImapMatchedSequenceSet matched;

        Assert.Throws<ImapIncapableException>(delegate {
          session.UidESort(ImapSortCriteria.DateReverse,
                           ImapSearchCriteria.Undeleted & ImapSearchCriteria.Unkeyword("$Junk"),
                           ImapSearchResultOptions.All,
                           out matched);
        });

        return -1;
      });
    }

    [Test]
    public void TestSort_DisplayIncapable()
    {
      SelectMailbox(new[] {"SORT"}, delegate(ImapSession session, ImapPseudoServer server) {
        ImapMatchedSequenceSet matched;

        try {
          session.Sort(ImapSortCriteria.Date + ImapSortCriteria.DisplayFrom, ImapSearchCriteria.All, out matched);
          Assert.Fail("ImapIncapableException not thrown");
        }
        catch (ImapIncapableException ex) {
          CollectionAssert.Contains(ex.RequiredCapabilities, ImapCapability.SortDisplay);
        }

        return -1;
      });
    }

    [Test]
    public void TestThread()
    {
      SelectMailbox(new[] {"THREAD=REFERENCES"}, delegate(ImapSession session, ImapPseudoServer server) {
        // THREAD transaction
        server.EnqueueResponse("* THREAD (166)(167)(168)(169)(172)(170)(171)(173)(174 (175)(176)(178)(181)(180))(179)(177 (183)(182)(188)(184)(185)(186)(187)(189))(190)(191)(192)(193)(194 195)(196 (197)(198))(199)(200 202)(201)(203)(204)(205)(206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References, ImapSearchCriteria.All, out threadList));

        Assert.AreEqual("0004 THREAD REFERENCES utf-8 ALL\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(25, threadList.Children.Count);
#if !NET_4_5
        Assert.IsTrue(threadList.Children.IsReadOnly);
        Assert.Throws<NotSupportedException>(() => threadList.Children.Clear());
#endif

        return 1;
      });
    }

    [Test]
    public void TestThread_NothingMatched()
    {
      SelectMailbox(new[] {"THREAD=REFERENCES"}, delegate(ImapSession session, ImapPseudoServer server) {
        // THREAD transaction
        server.EnqueueResponse("* THREAD \r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References, ImapSearchCriteria.All, out threadList));

        Assert.AreEqual("0004 THREAD REFERENCES utf-8 ALL\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(threadList);
        Assert.AreEqual(0, threadList.Children.Count);
#if !NET_4_5
        Assert.IsTrue(threadList.Children.IsReadOnly);
        Assert.Throws<NotSupportedException>(() => threadList.Children.Clear());
#endif

        return 1;
      });
    }

    [Test]
    public void TestThread_WithLiteral_DefaultCharset()
    {
      SelectMailbox(new[] {"THREAD=REFERENCES"}, delegate(ImapSession session, ImapPseudoServer server) {
        // THREAD transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* THREAD (206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References,
                                           ImapSearchCriteria.Subject("件名"),
                                           out threadList));

        Assert.AreEqual("0004 THREAD REFERENCES utf-8 SUBJECT {6}\r\n" +
                        "件名\r\n",
                        server.DequeueAll());

        Assert.IsNotNull(threadList);
        Assert.AreEqual(2, threadList.Children.Count);
#if !NET_4_5
        Assert.IsTrue(threadList.Children.IsReadOnly);
        Assert.Throws<NotSupportedException>(() => threadList.Children.Clear());
#endif

        return 1;
      });
    }

    [Test]
    public void TestThread_WithLiteral_SpecifiedCharset()
    {
      SelectMailbox(new[] {"THREAD=REFERENCES"}, delegate(ImapSession session, ImapPseudoServer server) {
        // THREAD transaction
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueResponse("* THREAD (206 207)(208)\r\n" +
                               "0004 OK THREAD completed\r\n");

        ImapThreadList threadList;

        Assert.IsTrue((bool)session.Thread(ImapThreadingAlgorithm.References,
                                           ImapSearchCriteria.Subject("件名"),
                                           Encoding.GetEncoding(932),
                                           out threadList));

        Assert.AreEqual("0004 THREAD REFERENCES shift_jis SUBJECT {4}\r\n" +
                        "\x008c\x008f\x0096\x00bc\r\n",
                        server.DequeueAll(NetworkTransferEncoding.Transfer8Bit));

        Assert.IsNotNull(threadList);

        return 1;
      });
    }

    [Test]
    public void TestThread_Incapable()
    {
      SelectMailbox(new[] {"THREAD=ORDEREDSUBJECT"}, delegate(ImapSession session, ImapPseudoServer server) {
        ImapThreadList threadList;

        Assert.Throws<ImapIncapableException>(delegate {
          session.Thread(ImapThreadingAlgorithm.References, ImapSearchCriteria.All, out threadList);
        });

        return -1;
      });
    }

    [Test]
    public void TestStore()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // STORE
        server.EnqueueResponse("* 2 FETCH (FLAGS (\\Deleted \\Seen))\r\n" +
                               "* 3 FETCH (FLAGS (\\Deleted))\r\n" +
                               "* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n" +
                               "0004 OK STORE completed\r\n");

        ImapMessageAttribute[] messagesAttrs;

        Assert.IsTrue((bool)session.Store(ImapSequenceSet.CreateRangeSet(2, 4),
                                    ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted),
                                    out messagesAttrs));

        Assert.AreEqual("0004 STORE 2:4 +FLAGS (\\Deleted)\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(3, messagesAttrs.Length);

        Assert.AreEqual(2, messagesAttrs[0].Sequence);
        Assert.AreEqual(0UL, messagesAttrs[0].ModSeq);
        Assert.AreEqual(2, messagesAttrs[0].Flags.Count);
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messagesAttrs[1].Sequence);
        Assert.AreEqual(0UL, messagesAttrs[1].ModSeq);
        Assert.AreEqual(1, messagesAttrs[1].Flags.Count);
        Assert.IsTrue(messagesAttrs[1].Flags.Contains(ImapMessageFlag.Deleted));

        Assert.AreEqual(4, messagesAttrs[2].Sequence);
        Assert.AreEqual(0UL, messagesAttrs[2].ModSeq);
        Assert.AreEqual(3, messagesAttrs[2].Flags.Count);
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Deleted));
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Flagged));
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Seen));

        return 1;
      });
    }

    [Test]
    public void TestStore_NotRetrieveResult()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // STORE
        server.EnqueueResponse("* 2 FETCH (FLAGS (\\Deleted \\Seen))\r\n" +
                               "* 3 FETCH (FLAGS (\\Deleted))\r\n" +
                               "* 4 FETCH (FLAGS (\\Deleted \\Flagged \\Seen))\r\n" +
                               "0004 OK STORE completed\r\n");

        Assert.IsTrue((bool)session.Store(ImapSequenceSet.CreateRangeSet(2, 4),
                                          ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted)));

        Assert.AreEqual("0004 STORE 2:4 +FLAGS (\\Deleted)\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestStore_EmptySequenceSet()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(delegate {
          session.Store(ImapSequenceSet.CreateSet(new long[] {}), ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted));
        });

        return -1;
      });
    }

    [Test]
    public void TestStore_EmptyUidSet()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(delegate {
          session.Store(ImapSequenceSet.CreateUidSet(new long[] {}), ImapStoreDataItem.AddFlags(ImapMessageFlag.Deleted));
        });

        return -1;
      });
    }

    [Test]
    public void TestStore_UnchangedSince()
    {
      SelectMailbox(new[] {"CONDSTORE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // STORE
        server.EnqueueResponse("* 7 FETCH (MODSEQ (320162342) FLAGS (\\Seen \\Deleted))\r\n" +
                               "* 5 FETCH (MODSEQ (320162350))\r\n" +
                               "* 9 FETCH (MODSEQ (320162349) FLAGS (\\Answered))\r\n" +
                               "0004 OK [MODIFIED 7,9] Conditional STORE failed\r\n");

        ImapMessageAttribute[] messagesAttrs;
        ImapSequenceSet failedMessageSet;

        Assert.IsTrue((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateSet(7, 5, 9),
                                                  ImapStoreDataItem.AddFlagsSilent(ImapMessageFlag.Deleted),
                                                  320162338UL,
                                                  out messagesAttrs,
                                                  out failedMessageSet));

        Assert.AreEqual("0004 STORE 7,5,9 (UNCHANGEDSINCE 320162338) +FLAGS.SILENT (\\Deleted)\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(3, messagesAttrs.Length);

        Assert.AreEqual(7, messagesAttrs[0].Sequence);
        Assert.AreEqual(320162342UL, messagesAttrs[0].ModSeq);
        Assert.AreEqual(2, messagesAttrs[0].Flags.Count);
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Seen));
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Deleted));

        Assert.AreEqual(5, messagesAttrs[1].Sequence);
        Assert.AreEqual(320162350UL, messagesAttrs[1].ModSeq);
        Assert.AreEqual(0, messagesAttrs[1].Flags.Count);

        Assert.AreEqual(9, messagesAttrs[2].Sequence);
        Assert.AreEqual(320162349UL, messagesAttrs[2].ModSeq);
        Assert.AreEqual(1, messagesAttrs[2].Flags.Count);
        Assert.IsTrue(messagesAttrs[2].Flags.Contains(ImapMessageFlag.Answered));

        var failed = failedMessageSet.ToArray();

        Assert.AreEqual(2, failed.Length);
        Assert.AreEqual(7, failed[0]);
        Assert.AreEqual(9, failed[1]);

        return 1;
      });
    }

    [Test]
    public void TestStore_UnchangedSince_NotRetrieveResult()
    {
      SelectMailbox(new[] {"CONDSTORE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // STORE
        server.EnqueueResponse("* 7 FETCH (MODSEQ (320162342) FLAGS (\\Seen \\Deleted))\r\n" +
                               "* 5 FETCH (MODSEQ (320162350))\r\n" +
                               "* 9 FETCH (MODSEQ (320162349) FLAGS (\\Answered))\r\n" +
                               "0004 OK [MODIFIED 7,9] Conditional STORE failed\r\n");

        Assert.IsTrue((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateSet(7, 5, 9),
                                                        ImapStoreDataItem.AddFlagsSilent(ImapMessageFlag.Deleted),
                                                        320162338UL));

        Assert.AreEqual("0004 STORE 7,5,9 (UNCHANGEDSINCE 320162338) +FLAGS.SILENT (\\Deleted)\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }


    [Test]
    public void TestStore_UnchangedSince_TaggedNoResponse()
    {
      SelectMailbox(new[] {"CONDSTORE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // STORE
        server.EnqueueResponse("* 1 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                               "* 3 FETCH (MODSEQ (320172342) FLAGS (\\SEEN))\r\n" +
                               "0004 NO [MODIFIED 2] Some of the messages no longer exist.\r\n");

        ImapMessageAttribute[] messagesAttrs;
        ImapSequenceSet failedMessageSet;

        Assert.IsFalse((bool)session.StoreUnchangedSince(ImapSequenceSet.CreateRangeSet(1, 7),
                                                   ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen),
                                                   320172338UL,
                                                   out messagesAttrs,
                                                   out failedMessageSet));

        Assert.AreEqual("0004 STORE 1:7 (UNCHANGEDSINCE 320172338) +FLAGS (\\Seen)\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(messagesAttrs);
        Assert.AreEqual(2, messagesAttrs.Length);

        Assert.AreEqual(1, messagesAttrs[0].Sequence);
        Assert.AreEqual(320172342UL, messagesAttrs[0].ModSeq);
        Assert.AreEqual(1, messagesAttrs[0].Flags.Count);
        Assert.IsTrue(messagesAttrs[0].Flags.Contains(ImapMessageFlag.Seen));

        Assert.AreEqual(3, messagesAttrs[1].Sequence);
        Assert.AreEqual(320172342UL, messagesAttrs[1].ModSeq);
        Assert.AreEqual(1, messagesAttrs[1].Flags.Count);
        Assert.IsTrue(messagesAttrs[1].Flags.Contains(ImapMessageFlag.Seen));

        var failed = failedMessageSet.ToArray();

        Assert.AreEqual(1, failed.Length);
        Assert.AreEqual(2, failed[0]);

        return 1;
      });
    }

    [Test]
    public void TestStore_UnchangedSince_Incapable()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // STORE
        ImapMessageAttribute[] messagesAttrs;
        ImapSequenceSet failedMessageSet;

        Assert.Throws<ImapIncapableException>(delegate {
          session.StoreUnchangedSince(ImapSequenceSet.CreateRangeSet(1, 7),
                                      ImapStoreDataItem.AddFlags(ImapMessageFlag.Seen),
                                      320172338UL,
                                      out messagesAttrs,
                                      out failedMessageSet);
        });

        return -1;
      });
    }

    [Test]
    public void TestExpunge()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 4 EXPUNGE\r\n" +
                               "* 7 EXPUNGE\r\n" +
                               "* 11 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");

        long[] expunged;

        Assert.IsTrue((bool)session.Expunge(out expunged));

        Assert.AreEqual("0004 EXPUNGE\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(4, expunged.Length);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(4,  expunged[1]);
        Assert.AreEqual(7,  expunged[2]);
        Assert.AreEqual(11, expunged[3]);

        return 1;
      });
    }

    [Test]
    public void TestExpunge_NotRetrieveResult()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 4 EXPUNGE\r\n" +
                               "* 7 EXPUNGE\r\n" +
                               "* 11 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");

        Assert.IsTrue((bool)session.Expunge());

        Assert.AreEqual("0004 EXPUNGE\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestUidExpunge()
    {
      SelectMailbox(new[] {"UIDPLUS"}, delegate(ImapSession session, ImapPseudoServer server) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");

        long[] expunged;

        Assert.IsTrue((bool)session.UidExpunge(ImapSequenceSet.CreateUidFromSet(1), out expunged));

        Assert.AreEqual("0004 UID EXPUNGE 1:*\r\n",
                        server.DequeueRequest());

        Assert.AreEqual(3, expunged.Length);
        Assert.AreEqual(3,  expunged[0]);
        Assert.AreEqual(3,  expunged[1]);
        Assert.AreEqual(3,  expunged[2]);

        return 1;
      });
    }

    [Test]
    public void TestUidExpunge_NotRetrieveResult()
    {
      SelectMailbox(new[] {"UIDPLUS"}, delegate(ImapSession session, ImapPseudoServer server) {
        // EXPUNGE transaction
        server.EnqueueResponse("* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "* 3 EXPUNGE\r\n" +
                               "0004 OK EXPUNGE completed\r\n");

        Assert.IsTrue((bool)session.UidExpunge(ImapSequenceSet.CreateUidFromSet(1)));

        Assert.AreEqual("0004 UID EXPUNGE 1:*\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestUidExpunge_SequenceSet()
    {
      SelectMailbox(new[] {"UIDPLUS"}, delegate(ImapSession session, ImapPseudoServer server) {
        long[] expunged;

        Assert.Throws<ArgumentException>(() => session.UidExpunge(ImapSequenceSet.CreateFromSet(1), out expunged));

        return -1;
      });
    }

    [Test]
    public void TestUidExpunge_EmptyUidSet()
    {
      SelectMailbox(new[] {"UIDPLUS"}, delegate(ImapSession session, ImapPseudoServer server) {
        long[] expunged;

        Assert.Throws<ArgumentException>(() => session.UidExpunge(ImapSequenceSet.CreateUidSet(new long[] {}), out expunged));

        return -1;
      });
    }

    [Test]
    public void TestUidExpunge_Incapable()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.UidPlus));

        long[] expunged;

        Assert.Throws<ImapIncapableException>(() => session.UidExpunge(ImapSequenceSet.CreateUidFromSet(1), out expunged));

        return -1;
      });
    }

    [Test]
    public void TestCopy()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // COPY transaction
        server.EnqueueResponse("0004 OK COPY completed\r\n");

        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX"));

        Assert.AreEqual("0004 COPY 1:10 INBOX\r\n",
                        server.DequeueRequest());

        return -1;
      });
    }

    [Test]
    public void TestCopy_EmptySequenceSet()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(() => session.Copy(ImapSequenceSet.CreateSet(new long[] {}), "INBOX"));

        return -1;
      });
    }

    [Test]
    public void TestCopy_EmptyUidSet()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(() => session.Copy(ImapSequenceSet.CreateUidSet(new long[] {}), "INBOX"));

        return -1;
      });
    }

    [Test]
    public void TestCopy_TryCreate()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server, Uri expectedAuthority) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
        server.EnqueueResponse("0005 OK CREATE completed\r\n");
        server.EnqueueResponse("0006 OK COPY completed\r\n");

        ImapMailbox created;

        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateUidRangeSet(1, 10), "INBOX.copyto", out created));

        Assert.AreEqual("INBOX.copyto", created.Name);
        Assert.AreEqual(new Uri(expectedAuthority, "./INBOX.copyto"),
                        created.Url);
        Assert.IsNotNull(created.Flags);
        Assert.IsNotNull(created.ApplicableFlags);
        Assert.IsNotNull(created.PermanentFlags);

        Assert.AreEqual("0004 UID COPY 1:10 INBOX.copyto\r\n",
                        server.DequeueRequest());
        Assert.AreEqual("0005 CREATE INBOX.copyto\r\n",
                        server.DequeueRequest());
        Assert.AreEqual("0006 UID COPY 1:10 INBOX.copyto\r\n",
                        server.DequeueRequest());

        return 3;
      });
    }

    [Test]
    public void TestCopy_TryCreate_NoReferralResponseCode()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [REFERRAL IMAP://user;AUTH=*@SERVER2/SHARED/STUFF] Unable to copy message(s) to SERVER2.\r\n");

        ImapMailbox created;

        Assert.IsFalse((bool)session.Copy(ImapSequenceSet.CreateUidRangeSet(1, 10), "SHARED/STUFF", out created));

        Assert.IsNull(created);

        Assert.AreEqual("0004 UID COPY 1:10 SHARED/STUFF\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestCopy_TryCreate_DontTryCreate()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // COPY transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");

        Assert.IsFalse((bool)session.Copy(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX.copyto"));

        Assert.AreEqual("0004 COPY 1:10 INBOX.copyto\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestCopy_WithAppendUidResponseCode()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        // COPY transaction
        server.EnqueueResponse("0004 OK [COPYUID 38505 304,319:320 3956:3958] Done\r\n");

        ImapCopiedUidSet copied;

        Assert.IsTrue((bool)session.Copy(ImapSequenceSet.CreateRangeSet(2, 4), "meeting", out copied));

        Assert.AreEqual("0004 COPY 2:4 meeting\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(copied);
        Assert.AreEqual(38505, copied.UidValidity);

        var copiedUids = copied.CopiedUidSet.ToArray();

        Assert.AreEqual(3, copiedUids.Length);
        Assert.AreEqual(304, copiedUids[0]);
        Assert.AreEqual(319, copiedUids[1]);
        Assert.AreEqual(320, copiedUids[2]);

        var assignedUids = copied.AssignedUidSet.ToArray();

        Assert.AreEqual(3, assignedUids.Length);
        Assert.AreEqual(3956, assignedUids[0]);
        Assert.AreEqual(3957, assignedUids[1]);
        Assert.AreEqual(3958, assignedUids[2]);

        Smdn.Net.TestUtils.SerializeBinary(copied, delegate(ImapCopiedUidSet deserialized) {
          Assert.AreEqual(38505L, deserialized.UidValidity);
          Assert.AreEqual("304,319:320", deserialized.CopiedUidSet.ToString());
          Assert.AreEqual("3956:3958", deserialized.AssignedUidSet.ToString());
        });

        return 1;
      });
    }

    [Test]
    public void TestMove()
    {
      SelectMailbox(new[] {"MOVE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // MOVE transaction
        server.EnqueueResponse("0004 OK MOVE completed\r\n");

        Assert.IsTrue((bool)session.Move(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX"));

        Assert.AreEqual("0004 MOVE 1:10 INBOX\r\n",
                        server.DequeueRequest());

        return -1;
      });
    }

    [Test]
    public void TestMove_EmptySequenceSet()
    {
      SelectMailbox(new[] {"MOVE"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(() => session.Move(ImapSequenceSet.CreateSet(new long[] {}), "INBOX"));

        return -1;
      });
    }

    [Test]
    public void TestMove_EmptyUidSet()
    {
      SelectMailbox(new[] {"MOVE"}, delegate(ImapSession session, ImapPseudoServer server) {
        Assert.Throws<ArgumentException>(() => session.Move(ImapSequenceSet.CreateUidSet(new long[] {}), "INBOX"));

        return -1;
      });
    }

    [Test]
    public void TestMove_TryCreate()
    {
      SelectMailbox(new[] {"MOVE"}, delegate(ImapSession session, ImapPseudoServer server, Uri expectedAuthority) {
        // MOVE transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");
        server.EnqueueResponse("0005 OK CREATE completed\r\n");
        server.EnqueueResponse("0006 OK COPY completed\r\n");

        ImapMailbox created;

        Assert.IsTrue((bool)session.Move(ImapSequenceSet.CreateUidRangeSet(1, 10), "INBOX.copyto", out created));

        Assert.AreEqual("INBOX.copyto", created.Name);
        Assert.AreEqual(new Uri(expectedAuthority, "./INBOX.copyto"),
                        created.Url);
        Assert.IsNotNull(created.Flags);
        Assert.IsNotNull(created.ApplicableFlags);
        Assert.IsNotNull(created.PermanentFlags);

        Assert.AreEqual("0004 UID MOVE 1:10 INBOX.copyto\r\n",
                        server.DequeueRequest());
        Assert.AreEqual("0005 CREATE INBOX.copyto\r\n",
                        server.DequeueRequest());
        Assert.AreEqual("0006 UID MOVE 1:10 INBOX.copyto\r\n",
                        server.DequeueRequest());

        return 3;
      });
    }

    [Test]
    public void TestMove_TryCreate_DontTryCreate()
    {
      SelectMailbox(new[] {"MOVE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // MOVE transaction
        server.EnqueueResponse("0004 NO [TRYCREATE] Mailbox doesn't exist: INBOX.copyto\r\n");

        Assert.IsFalse((bool)session.Move(ImapSequenceSet.CreateRangeSet(1, 10), "INBOX.copyto"));

        Assert.AreEqual("0004 MOVE 1:10 INBOX.copyto\r\n",
                        server.DequeueRequest());

        return 1;
      });
    }

    [Test]
    public void TestMove_WithCopyUidResponseCode()
    {
      SelectMailbox(new[] {"MOVE"}, delegate(ImapSession session, ImapPseudoServer server) {
        // MOVE transaction
        server.EnqueueResponse("0004 OK [COPYUID 432432 42:69 1202:1229] Done\r\n");

        ImapCopiedUidSet copied;

        Assert.IsTrue((bool)session.Move(ImapSequenceSet.CreateRangeSet(42, 69), "foo", out copied));

        Assert.AreEqual("0004 MOVE 42:69 foo\r\n",
                        server.DequeueRequest());

        Assert.IsNotNull(copied);
        Assert.AreEqual(432432L, copied.UidValidity);
        Assert.AreEqual("42:69", copied.CopiedUidSet.ToString());
        Assert.AreEqual("1202:1229", copied.AssignedUidSet.ToString());

        Smdn.Net.TestUtils.SerializeBinary(copied, delegate(ImapCopiedUidSet deserialized) {
          Assert.AreEqual(432432L, deserialized.UidValidity);
          Assert.AreEqual("42:69", deserialized.CopiedUidSet.ToString());
          Assert.AreEqual("1202:1229", deserialized.AssignedUidSet.ToString());
        });

        return 1;
      });
    }

    [Test]
    public void TestMove_Incapable()
    {
      SelectMailbox(delegate(ImapSession session, ImapPseudoServer server) {
        Assert.IsFalse(session.ServerCapabilities.Contains(ImapCapability.Move));

        Assert.Throws<ImapIncapableException>(() => session.Move(ImapSequenceSet.CreateUidFromSet(1), "foo"));

        return -1;
      });
    }
  }
}
