using System;
using System.IO;
using System.Net;
using System.Collections.Generic;
#if NET_4_0
using System.Threading;
using System.Threading.Tasks;
#endif
using NUnit.Framework;

using Smdn.Formats;
using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client.Session {
  [TestFixture]
  public class ImapSessionCreatorTests {
    private class Profile : IImapSessionProfile {
      public Uri Authority {
        get; set;
      }

      public int Timeout {
        get; set;
      }

      public int SendTimeout {
        get; set;
      }

      public int ReceiveTimeout {
        get; set;
      }

      public ICredentialsByHost Credentials {
        get; set;
      }

      public bool UseDeflateIfAvailable {
        get { return false; }
      }

      public string[] UsingSaslMechanisms {
        get; set;
      }

      public bool UseTlsIfAvailable {
        get; set;
      }

      public bool AllowInsecureLogin {
        get; set;
      }

      public Profile(NetworkCredential cred, string userinfo, string hostPort)
        : this(cred, ImapUri.UriSchemeImap, userinfo, hostPort)
      {
      }

      public Profile(NetworkCredential cred, string scheme, string userinfo, string hostPort)
      {
        this.UseTlsIfAvailable = true;
        this.Timeout = 1000;
        this.SendTimeout = -1;
        this.ReceiveTimeout = -1;
        this.Credentials = cred;
        this.AllowInsecureLogin = false;

        if (userinfo == null)
          this.Authority = new Uri(string.Format("{0}://{1}/", scheme, hostPort));
        else
          this.Authority = new Uri(string.Format("{0}://{1}@{2}/", scheme, userinfo, hostPort));
      }
    }

    [Test]
    public void TestCreateSession_CreateSslStreamCallback_Null()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "IMAPS", "user", "imap.invalid");

      Assert.Throws<ArgumentNullException>(() => ImapSessionCreator.CreateSession(prof, null, null));
    }

    [Test]
    public void TestCreateSession_ConnectionRefusedByTimeout()
    {
      using (var server = ImapPseudoServer2.StartNotAccepting()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.Timeout = 500;

        try {
          ImapSessionCreator.CreateSession(prof, null, null);

          Assert.Fail("ImapConnectionException not thrown");
        }
        catch (ImapConnectionException ex) {
          var timeoutException = ex.InnerException as TimeoutException;

          Assert.IsNotNull(timeoutException);
        }
      }
    }

    [Test]
    public void TestCreateSession_ConnectionRefusedByDnsError()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "user", "imap.invalid");

      prof.Timeout = -1;

      try {
        ImapSessionCreator.CreateSession(prof, null, null);

        Assert.Fail("ImapConnectionException not thrown");
      }
      catch (ImapConnectionException ex) {
        var socketException = ex.InnerException as System.Net.Sockets.SocketException;

        Assert.IsNotNull(socketException);
      }
    }

    [Test]
    public void TestCreateSession_Imap4Rev1Incapable()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4\r\n" +
                                     "$tag OK done\r\n");

        Assert.Throws<ImapIncapableException>(() => ImapSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestCreateSession_ConnectSecurePort()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imaps", "user", server.HostPort);

        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        var streamUpgraded = false;

        using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsTrue(streamUpgraded, "stream upgraded");

          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_ConnectSecurePort_Exception()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imaps", "user", server.HostPort);

        try {
          using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            throw new System.Security.Authentication.AuthenticationException();
          }));
          Assert.Fail("ImapSecureConnectionException not thrown");
        }
        catch (ImapSecureConnectionException ex) {
          Assert.IsNotNull(ex.InnerException);

          var upgradeException = ex.InnerException as ImapUpgradeConnectionException;

          Assert.IsNotNull(upgradeException);
          Assert.IsNotNull(upgradeException.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(upgradeException.InnerException);
        }
      }
    }

    [Test]
    public void TestCreateSession_StartTls_UseTlsIfAvailable()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        var streamUpgraded = false;

        using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsTrue(streamUpgraded, "stream upgraded");

          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_StartTls_DontUseTlsWhetherAvailable()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.UseTlsIfAvailable = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        var streamUpgraded = false;

        using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          streamUpgraded = true;
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.IsFalse(streamUpgraded, "stream upgraded");

          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsFalse(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_StartTls_Failure()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag BAD done\r\n");

        var streamUpgraded = false;

        try {
          using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            streamUpgraded = true;
            return baseStream; // TODO: return SSL stream
          }));
          Assert.Fail("ImapSecureConnectionException not thrown");
        }
        catch (ImapSecureConnectionException ex) {
          Assert.IsNull(ex.InnerException);
        }

        Assert.IsFalse(streamUpgraded, "stream upgraded");
      }
    }

    [Test]
    public void TestCreateSession_StartTls_Exception()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK done\r\n");

        try {
          using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
            throw new System.Security.Authentication.AuthenticationException();
          }));
          Assert.Fail("ImapSecureConnectionException not thrown");
        }
        catch (ImapSecureConnectionException ex) {
          Assert.IsNotNull(ex.InnerException);

          var upgradeException = ex.InnerException as ImapUpgradeConnectionException;

          Assert.IsNotNull(upgradeException);
          Assert.IsNotNull(upgradeException.InnerException);
          Assert.IsInstanceOf<System.Security.Authentication.AuthenticationException>(upgradeException.InnerException);
        }
      }
    }

    [Test]
    public void TestCreateSession_ProfileTimeoutMustBeApplied()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", server.HostPort);

        prof.AllowInsecureLogin = true;
        prof.Timeout        = 100 * 1000;
        prof.SendTimeout    = 300 * 1000;
        prof.ReceiveTimeout = 500 * 1000;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsFalse(session.IsSecureConnection);

          Assert.AreEqual(100 * 1000, session.TransactionTimeout);
          Assert.AreEqual(300 * 1000, session.SendTimeout);
          Assert.AreEqual(500 * 1000, session.ReceiveTimeout);
        }
      }
    }

    [Test]
    public void TestCreateSession_CapabilityAdvertised()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsFalse(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestCreateSession_CapabilityNotAdvertised()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsFalse(session.IsSecureConnection);
        }
      }
    }

#if NET_4_0
    [Test]
    public void TestCreateSession_CancellationRequested_BeforeCreateSession()
    {
      var prof = new Profile(new NetworkCredential("user", "pass"), "imap", "user", "imap.invalid");

      prof.AllowInsecureLogin = true;

      var cts = new CancellationTokenSource();

      cts.Cancel();

      var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

      Assert.Throws<AggregateException>(() => t.Wait());

      Assert.IsTrue(t.IsCanceled, "IsCancelled");
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeQueryingServerCapability()
    {
      // connect
      //   <--cancellation
      // CAPABILITY

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        var cts = new CancellationTokenSource();

        server.SetGreetingResponse("* OK ready\r\n",
                                   () => cts.Cancel(),
                                   null);

        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        Assert.IsNull(server.DequeueRequest()); // CAPABILITY not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeUpgradeStream()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // STARTTLS

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.AllowInsecureLogin = false;
        prof.UseTlsIfAvailable = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // STARTTLS not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_WithSpecificMechanism()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(SaslMechanism)

      using (var server = ImapPseudoServer2.StartNew()) {
        var credential = new NetworkCredential("imap-user", (string)null);
        var prof = new Profile(null, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n",
                                     () => cts.Cancel(),
                                     null);

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          authMechanism.Credential = credential;

          var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, authMechanism, null, cts.Token), cts.Token);

          Assert.Throws<AggregateException>(() => t.Wait());

          Assert.IsTrue(t.IsCanceled, "IsCancelled");
        }

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(AUTH=*)

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToNext()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(AUTH=*)
      //   <--cancellation
      // AUTHENTICATE(AUTH=*)

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE LOGIN\r\n",
                                     "$tag NO done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE LOGIN", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_SelectAppropriateMechanism_FallbackToLogin()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(AUTH=*)
      //   <--cancellation
      // LOGIN

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN"};
        prof.AllowInsecureLogin = true;

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE LOGIN\r\n",
                                     "$tag NO done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE LOGIN", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // LOGIN not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_MechanismSpecified()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(AUTH=XXX)

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=LOGIN", server.HostPort);

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_Anonymous()
    {
      // connect
      // CAPABILITY
      //   <--cancellation
      // AUTHENTICATE(AUTH=ANONYMOUS)

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, "user%40imap.example.net;auth=anonymous", server.HostPort);

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // AUTHENTICATE not requested
      }
    }

    [Test]
    public void TestCreateSession_CancellationRequested_BeforeAuth_Anonymous_FallbackToLogin()
    {
      // connect
      // CAPABILITY
      // AUTHENTICATE(AUTH=ANONYMOUS)
      //   <--cancellation
      // LOGIN

      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        var cts = new CancellationTokenSource();
        var t = Task.Factory.StartNew(() => ImapSessionCreator.CreateSession(prof, null, null, cts.Token), cts.Token);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE ANONYMOUS\r\n",
                                     "$tag NO done\r\n",
                                     () => cts.Cancel(),
                                     null);

        Assert.Throws<AggregateException>(() => t.Wait());

        Assert.IsTrue(t.IsCanceled, "IsCancelled");

        StringAssert.Contains(" CAPABILITY", server.DequeueRequest());
        StringAssert.Contains(" AUTHENTICATE ANONYMOUS", server.DequeueRequest());
        Assert.IsNull(server.DequeueRequest()); // LOGIN not requested
      }
    }
#endif

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAuthMechanismSpecified_AuthenticationFailure()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        try {
          Assert.IsNull(ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapAuthenticationException not thrown");
        }
        catch (ImapAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.ImapCommandResultCode.No, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK ready\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndAuthMechanismSpecified_AuthenticationFailure()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), ";auth=plain", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=PLAIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        try {
          Assert.IsNull(ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapAuthenticationException not thrown");
        }
        catch (ImapAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.ImapCommandResultCode.No, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified()
    {
      AuthenticateWithAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanism(string userinfo)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE LOGIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_ExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_ExceptionWhileAuthentication()
    {
      AuthenticateExceptionWhileAuthentication("user");
    }

    private void AuthenticateExceptionWhileAuthentication(string userinfo)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"DIGEST-MD5", "PLAIN"};

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                               "+ xxx-invalid-response-xxx\r\n");

        try {
          using (var session = ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapException not thrown");
        }
        catch (ImapException ex) {
          Assert.IsNotNull(ex.InnerException);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToNextAppropriateMechanism()
    {
      AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToNextAppropriateMechanism(string userinfo)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE LOGIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueResponse("AUTHENTICATE PLAIN\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToLogin()
    {
      AuthenticateWithAppropriateMechanismFallbackToLogin("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToLogin()
    {
      AuthenticateWithAppropriateMechanismFallbackToLogin("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToLogin(string userinfo)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_FallbackToLogin_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_FallbackToLogin_AuthenticationFailure()
    {
      AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure("user");
    }

    private void AuthenticateWithAppropriateMechanismFallbackToLoginAuthenticationFailure(string userinfo)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "$tag NO done\r\n");

        try {
          Assert.IsNull(ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapAuthenticationException not thrown");
        }
        catch (ImapAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.ImapCommandResultCode.No, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_UserAndAppropriateAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound("user;auth=*");
    }

    [Test]
    public void TestAuthenticate_UserAndNoAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound("user");
    }

    private void AuthenticateWithAppropriateMechanismAppropriateMechanismNotFound(string userinfo)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), userinfo, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");

        Assert.Throws<ImapNoAppropriateAuthMechanismException>(() => ImapSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_AnonymousMechanismSpecified()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, "user%40imap.example.net;auth=anonymous", server.HostPort);

        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE ANONYMOUS\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // AUTHENTICATE
        StringAssert.Contains("user@imap.example.net", Base64.GetDecodedString(server.DequeueRequest()));
      }
    }

    [Test]
    public void TestAuthenticate_AnonymousMechanismSpecified_AuthenticationFailure()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, "user%40imap.example.net;auth=anonymous", server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE ANONYMOUS\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        try {
          Assert.IsNull(ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapAuthenticationException not thrown");
        }
        catch (ImapAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.ImapCommandResultCode.No, ex.Result.Code);
        }

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // AUTHENTICATE ANONYMOUS
        StringAssert.Contains("user@imap.example.net", Base64.GetDecodedString(server.DequeueRequest()));
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE ANONYMOUS\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueTaggedResponse("LOGIN anonymous anonymous@\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }

        server.DequeueRequest(); // CAPABILITY
        server.DequeueRequest(); // AUTHENTICATE ANONYMOUS
        StringAssert.Contains("anonymous@", Base64.GetDecodedString(server.DequeueRequest()));
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified_AuthenticationFailure()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGIN anonymous anonymous@\r\n",
                                     "$tag NO done\r\n");

        try {
          Assert.IsNull(ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapAuthenticationException not thrown");
        }
        catch (ImapAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.ImapCommandResultCode.No, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_NoUserAndNoAuthMechanismSpecified_AppropriateMechanismNotFound()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(null, null, server.HostPort);

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");

        Assert.Throws<ImapNoAppropriateAuthMechanismException>(() => ImapSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "CRAM-MD5"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_SecureSession()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 STARTTLS LOGINDISABLED\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("STARTTLS\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 SASL-IR AUTH=LOGIN AUTH=PLAIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE LOGIN ",
                                     "$tag NO done\r\n");
        server.EnqueueTaggedResponse("AUTHENTICATE PLAIN ",
                                     "$tag NO done\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, delegate(ConnectionBase connection, Stream baseStream) {
          return baseStream; // TODO: return SSL stream
        })) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
          Assert.IsTrue(session.IsSecureConnection);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_FallbackToNextAppropriateMechanism()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                               "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0" +
                               "RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hh" +
                               "cnNldD11dGYtOA==\r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueResponse("AUTHENTICATE CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var session = ImapSessionCreator.CreateSession(prof, null, null)) {
          Assert.AreEqual(ImapSessionState.Authenticated, session.State);
          Assert.AreEqual(prof.Authority, session.Authority);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_FallbackToNextAppropriateMechanism_AuthenticationFailure()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE DIGEST-MD5\r\n",
                               "+ cmVhbG09ImVsd29vZC5pbm5vc29mdC5jb20iLG5vbmNlPSJPQTZNRzl0" +
                               "RVFHbTJoaCIscW9wPSJhdXRoIixhbGdvcml0aG09bWQ1LXNlc3MsY2hh" +
                               "cnNldD11dGYtOA==\r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");
        server.EnqueueResponse("AUTHENTICATE CRAM-MD5\r\n",
                               "+ PDQwMDEzNDQxMTIxNDM1OTQuMTI3MjQ5OTU1MEBsb2NhbGhvc3Q+\r\n");
        server.EnqueueTaggedResponse("$tag NO done\r\n");

        try {
          Assert.IsNull(ImapSessionCreator.CreateSession(prof, null, null));
          Assert.Fail("ImapAuthenticationException not thrown");
        }
        catch (ImapAuthenticationException ex) {
          Assert.IsNull(ex.InnerException);
          Assert.IsNotNull(ex.Result);
          Assert.AreEqual(Protocol.Client.ImapCommandResultCode.No, ex.Result.Code);
        }
      }
    }

    [Test]
    public void TestAuthenticate_InsecureLoginDisallowed_AppropriateMechanismNotFound()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(new NetworkCredential("user", "pass"), "user;auth=*", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"ANONYMOUS", "LOGIN", "PLAIN"};
        prof.AllowInsecureLogin = false;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1 AUTH=X-UNKNOWN AUTH=DIGEST-MD5 AUTH=CRAM-MD5 AUTH=NTLM AUTH=PLAIN AUTH=LOGIN AUTH=ANONYMOUS\r\n" +
                                     "$tag OK done\r\n");

        Assert.Throws<ImapNoAppropriateAuthMechanismException>(() => ImapSessionCreator.CreateSession(prof, null, null));
      }
    }

    [Test]
    public void TestAuthenticate_WithSpecificMechanism()
    {
      AuthenticateSpecificMechanism(new NetworkCredential("sasl-user", (string)null));
    }

    [Test]
    public void TestAuthenticate_WithSpecificMechanism_CredentialNotSet()
    {
      AuthenticateSpecificMechanism(null);
    }

    private void AuthenticateSpecificMechanism(NetworkCredential credential)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var prof = new Profile(credential, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          authMechanism.Credential = credential;

          using (var session = ImapSessionCreator.CreateSession(prof, authMechanism, null)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            if (credential == null)
              Assert.AreEqual(new Uri(string.Format("imap://AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort)),
                              session.Authority);
            else
              Assert.AreEqual(new Uri(string.Format("imap://sasl-user;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort)),
                              session.Authority);
          }

          Assert.AreSame(credential,
                         authMechanism.Credential,
                         "credential must be kept");

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }
      }
    }

    [Test]
    public void TestAuthenticate_WithSpecificMechanism_CapabilityNotAdvertised()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        var credential = new NetworkCredential("sasl-user", (string)null);
        var prof = new Profile(credential, "imap", server.HostPort);

        prof.UsingSaslMechanisms = new[] {"LOGIN", "PLAIN", "ANONYMOUS", "DIGEST-MD5", "CRAM-MD5"};
        prof.AllowInsecureLogin = true;

        server.SetGreetingResponse("* OK ready\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("$tag OK done\r\n");
        server.EnqueueTaggedResponse("CAPABILITY\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n");

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          authMechanism.Credential = credential;

          using (var session = ImapSessionCreator.CreateSession(prof, authMechanism, null)) {
            Assert.AreEqual(ImapSessionState.Authenticated, session.State);

            Assert.AreEqual(new Uri(string.Format("imap://sasl-user;AUTH=X-PSEUDO-MECHANISM@{0}/", server.HostPort)),
                            session.Authority);
          }

          Assert.AreSame(credential,
                         authMechanism.Credential,
                         "credential must be kept");

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }
      }
    }
  }
}
