using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Threading;
#if NET_4_0
using System.Threading.Tasks;
#endif
using NUnit.Framework;

using Smdn.Net.Imap4.Client.Session;
using Smdn.Net.Imap4.Protocol;
using Smdn.Net.Imap4.Protocol.Client;
using Smdn.Security.Authentication.Sasl;
using Smdn.Security.Authentication.Sasl.Client;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapClientTests {
    [Test]
    public void TestConstruct()
    {
      Assert.AreEqual(new Uri("imap://localhost/"),
                      DefaultPropertyAssertion(new ImapClient()).Profile.Authority,
                      "#1 authority");

      Assert.AreEqual(new Uri("imap://user;AUTH=LOGIN@imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient(new Uri("imap://user;AUTH=LOGIN@imap.example.net:10110/"))).Profile.Authority,
                      "#2 authority");

      Assert.AreEqual(new Uri("imap://imap.example.net/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net")).Profile.Authority,
                      "#3 authority");

      Assert.AreEqual(new Uri("imap://user@imap.example.net/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", "user")).Profile.Authority,
                      "#4 authority");

      Assert.AreEqual(new Uri("imaps://user@imap.example.net/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", true, "user")).Profile.Authority,
                      "#5 authority");

      Assert.AreEqual(new Uri("imap://imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110)).Profile.Authority,
                      "#6 authority");

      Assert.AreEqual(new Uri("imap://user@imap.example.net:10110/"), 
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, "user")).Profile.Authority,
                      "#7 authority");

      Assert.AreEqual(new Uri("imap://user@imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, "user")).Profile.Authority,
                      "#8 authority");

      Assert.AreEqual(new Uri("imaps://user@imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, true, "user")).Profile.Authority,
                      "#9 authority");

      Assert.AreEqual(new Uri("imap://user;AUTH=PLAIN@imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, "user", "PLAIN")).Profile.Authority,
                      "#10 authority");

      Assert.AreEqual(new Uri("imaps://user;AUTH=DIGEST-MD5@imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, true, "user", "DIGEST-MD5")).Profile.Authority,
                      "#11 authority");

      Assert.AreEqual(new Uri("imaps://user@imap.example.net:10110/"),
                      DefaultPropertyAssertion(new ImapClient("imap.example.net", 10110, true, "user", null, 1000), 1000).Profile.Authority,
                      "#12 authority");

      var profile = new ImapClientProfile(new Uri("imaps://user@imap.example.net/"));

      Assert.AreEqual(new Uri("imaps://user@imap.example.net/"),
                      DefaultPropertyAssertion(new ImapClient(profile)).Profile.Authority,
                      "#13 authority");
    }

    private ImapClient DefaultPropertyAssertion(ImapClient client)
    {
      return DefaultPropertyAssertion(client, System.Threading.Timeout.Infinite);
    }

    private ImapClient DefaultPropertyAssertion(ImapClient client, int timeoutMilliseconds)
    {
      Assert.IsNotNull(client.Profile);
      Assert.IsTrue(client.Profile.UseTlsIfAvailable);
      Assert.IsFalse(client.Profile.AllowInsecureLogin);
      Assert.IsFalse(client.IsConnected);
      Assert.AreEqual(timeoutMilliseconds, client.Timeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.ReceiveTimeout);
      Assert.AreEqual(System.Threading.Timeout.Infinite, client.SendTimeout);
      Assert.IsFalse(client.IsBusy);

      Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.OpenedMailbox));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerID));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerNamespace));

      Assert.DoesNotThrow(() => client.DisabledCapabilities.Clear());

      return client;
    }

    [Test]
    public void TestConnect_WithPassword()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnect_WithCredentials()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    [Test]
    public void TestConnect_WithSpecifiedSaslMechanism()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueResponse("AUTHENTICATE X-PSEUDO-MECHANISM\r\n",
                               "+ \r\n");
        server.EnqueueTaggedResponse("* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        using (var authMechanism = new SaslPseudoMechanism(1)) {
          client.Connect(authMechanism);

          Assert.AreEqual(Smdn.Security.Authentication.Sasl.SaslExchangeStatus.Succeeded,
                          authMechanism.ExchangeStatus);
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
      });
    }

    private class SaslXOAuth2 : XOAuth2Mechanism {
      public bool InitializeCalled = false;

      public SaslXOAuth2(string user, string accessToken)
        : base(user, accessToken)
      {
      }

      public override void Initialize()
      {
        base.Initialize();

        InitializeCalled = true;
      }
    }

    [Test]
    public void TestConnect_WithXOAuth2()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 SASL-IR] ImapPseudoServer ready\r\n"); // AUTH=XOAUTH2 not listed
        server.EnqueueTaggedResponse("AUTHENTICATE XOAUTH2 ",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        using (var oauth2 = new SaslXOAuth2("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
          client.Connect(oauth2);

          Assert.IsTrue(oauth2.InitializeCalled);
          Assert.AreEqual(SaslExchangeStatus.Continuing, oauth2.ExchangeStatus);
          Assert.IsNull(oauth2.LastErrorResults);
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
      });
    }

    [Test]
    public void TestConnect_WithXOAuth2_Error()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1 SASL-IR] ImapPseudoServer ready\r\n"); // AUTH=XOAUTH2 not listed
        server.EnqueueResponse("AUTHENTICATE XOAUTH2 ",
                               "+ eyJzdGF0dXMiOiI0MDEiLCJzY2hlbWVzIjoiYmVhcmVyIG1hYyIsInNjb\n3BlIjoiaHR0cHM6Ly9tYWlsLmdvb2dsZS5jb20vIn0K\r\n");
        server.EnqueueTaggedResponse("$tag NO SASL authentication failed\r\n");

        using (var oauth2 = new SaslXOAuth2("someuser@example.com", "vF9dft4qmTc2Nvb3RlckBhdHRhdmlzdGEuY29tCg==")) {
          Assert.Throws<ImapAuthenticationException>(() => client.Connect(oauth2));

          Assert.IsTrue(oauth2.InitializeCalled);
          Assert.AreEqual(SaslExchangeStatus.Failed, oauth2.ExchangeStatus);
          Assert.AreEqual("{\"status\":\"401\",\"schemes\":\"bearer mac\",\"scope\":\"https://mail.google.com/\"}\n", 
                          oauth2.LastErrorResults);
        }

        Assert.IsFalse(client.IsConnected);
      });
    }

    [Test]
    public void TestConnect_AlreadyConnected()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect(new NetworkCredential("user", "pass"));

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

        try {
          client.Connect("pass");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
        }
      });
    }

    [Test]
    public void TestConnect_IDCapable()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1 ID\r\n" + 
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("ID ",
                                     "* ID (\"name\" \"ImapPseudoServer\")\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.AreEqual("ImapPseudoServer", client.ServerID["name"]);
      });
    }

    [Test]
    public void TestConnect_NamespaceCapable()
    {
      Connect(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1 NAMESPACE\r\n" + 
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("NAMESPACE\r\n",
                                     "* NAMESPACE ((\"\" \"/\")) NIL NIL\r\n" + 
                                     "$tag OK done\r\n");

        client.Profile.AllowInsecureLogin = true;
        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
        Assert.AreEqual(1, client.ServerNamespace.PersonalNamespaces.Count);
        Assert.AreEqual("/", client.ServerNamespace.PersonalNamespaces[0].HierarchyDelimiter);
        Assert.AreEqual(0, client.ServerNamespace.SharedNamespaces.Count);
        Assert.AreEqual(0, client.ServerNamespace.OtherUsersNamespaces.Count);
      });
    }

    private void Connect(Action<ImapPseudoServer2, ImapClient> action)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          Assert.IsFalse(client.IsConnected);
          Assert.IsFalse(client.IsBusy);
          Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.OpenedMailbox));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerNamespace));
          Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerID));

          action(server, client);

          Assert.IsFalse(client.IsBusy);
        }
      }
    }

    [Test]
    public void TestBeginConnect_WithPassword()
    {
      BeginConnect(true);
    }

    [Test]
    public void TestBeginConnect_WithCredentials()
    {
      BeginConnect(false);
    }

    private void BeginConnect(bool usePassword)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 5000;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          var callbacked = false;
          var asyncStateObject = new object();

          var asyncCallback = (AsyncCallback)delegate(IAsyncResult ar) {
            Assert.IsNotNull(ar);
            Assert.AreSame(ar.AsyncState, asyncStateObject);
            callbacked = true;
          };

          client.Profile.AllowInsecureLogin = true;

          var asyncResult = usePassword
            ? (ImapClient.ConnectAsyncResult)client.BeginConnect("pass", asyncCallback, asyncStateObject)
            : (ImapClient.ConnectAsyncResult)client.BeginConnect(new NetworkCredential("user", "pass"), asyncCallback, asyncStateObject);

          Assert.IsFalse(client.IsConnected);
          Assert.IsNotNull(asyncResult);
          Assert.IsFalse(asyncResult.IsCompleted);
          Assert.IsFalse(asyncResult.EndConnectCalled);
          Assert.IsFalse(callbacked);

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          try {
            client.BeginConnect("pass");
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }

          Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

          var interval = TimeSpan.FromMilliseconds(250);

          Smdn.Net.TestUtils.Elapses(TimeSpan.FromMilliseconds(250), delegate {
            ThreadPool.QueueUserWorkItem(delegate(object state) {
              var s = state as ImapPseudoServer2;

              Thread.Sleep(interval);

              s.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                      "* CAPABILITY IMAP4rev1\r\n" +
                                      "$tag OK done\r\n");
            }, server);

            Assert.IsTrue(asyncResult.AsyncWaitHandle.WaitOne(3000, false));
          });

          Assert.IsFalse(client.IsConnected);

          Thread.Sleep(50); // wait for callback
          Assert.IsTrue(callbacked);
          Assert.IsFalse(asyncResult.EndConnectCalled);
          Assert.IsTrue(client.IsBusy, "IsBusy before EndConnect");

          client.EndConnect(asyncResult);

          Assert.IsFalse(client.IsBusy, "IsBusy after connected");

          Assert.IsTrue(asyncResult.EndConnectCalled);
          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsSecureSession);

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          try {
            client.EndConnect(asyncResult);
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (ArgumentException) {
          }
        }
      }
    }

    [Test]
    public void TestBeginConnect_AlreadyConnected()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        try {
          client.BeginConnect("pass");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
      });
    }

    [Test]
    public void TestBeginConnect_Exception()
    {
      using (var server = ImapPseudoServer2.StartNotAccepting()) {
        server.ThrowIfServerNotStopped = false;

        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 500;

          Assert.IsFalse(client.IsConnected);

          var asyncResult = client.BeginConnect("pass");

          /*
           * no response
           */

          Assert.Throws<ImapConnectionException>(() => client.EndConnect(asyncResult));

          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

#if NET_4_0
    [Test]
    public void TestConnectAsync_WithPassword()
    {
      ConnectAsync(true);
    }
    
    [Test]
    public void TestConnectAsync_WithCredentials()
    {
      ConnectAsync(false);
    }
    
    private void ConnectAsync(bool usePassword)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          var t = usePassword
            ? client.ConnectAsync("pass")
            : client.ConnectAsync(new NetworkCredential("user", "pass"));

          Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

          ThreadPool.QueueUserWorkItem(delegate(object state) {
            var s = state as ImapPseudoServer2;
            
            Thread.Sleep(250);

            s.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                    "* CAPABILITY IMAP4rev1\r\n" + 
                                    "$tag OK done\r\n");
          }, server);

          Assert.IsTrue(t.Wait(10000), "Task.Wait");
          Assert.IsTrue(client.IsConnected);
          Assert.IsFalse(client.IsBusy, "IsBusy after connected");
        }
      }
    }
    
    [Test]
    public void TestConnectAsync_AlreadyConnected()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);

        try {
          client.ConnectAsync("pass");
          Assert.Fail("InvalidOperationException not thrown");
        }
        catch (InvalidOperationException) {
        }

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
      });
    }
    
    [Test]
    public void TestConnectAsync_Exception()
    {
      using (var server = ImapPseudoServer2.StartNotAccepting()) {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 250;

          Assert.IsFalse(client.IsConnected);

          var t = client.ConnectAsync("pass");

          try {
            t.Wait(10000);
            Assert.Fail("AggregateException not thrown");
          }
          catch (AggregateException ex) {
            Assert.IsInstanceOf(typeof(ImapConnectionException), ex.InnerException);
            Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
          }

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

    [Test]
    public void TestConnectAsync_CancellationRequested()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.Timeout = 1000;
          client.Profile.AllowInsecureLogin = true;

          Assert.IsFalse(client.IsBusy, "IsBusy before connect");

          using (var e = new ManualResetEventSlim(false)) {
            var cts = new CancellationTokenSource();

            server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n",
                                       () => { e.Wait(); cts.Cancel(); },
                                       null);

            var t = client.ConnectAsync("pass", null, cts.Token);

            Assert.IsTrue(client.IsBusy, "IsBusy while connecting");

            e.Set();

            Assert.Throws<AggregateException>(() => t.Wait());

            Assert.IsTrue(t.IsCanceled, "IsCanceled");
          }

          Assert.IsFalse(client.IsBusy, "IsBusy after cancel");
          Assert.IsFalse(client.IsConnected);
          Assert.IsNull((client.Profile as IImapSessionProfile).Credentials);
        }
      }
    }
#endif

    [Test]
    public void TestLogout_AsyncConnectRunning()
    {
      DisconnectAsyncConnectRunning(true);
    }

    [Test]
    public void TestDisconnect_AsyncConnectRunning()
    {
      DisconnectAsyncConnectRunning(false);
    }

    private void DisconnectAsyncConnectRunning(bool logout)
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.ThrowIfServerNotStopped = false;

        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.BeginConnect("pass");

          try {
            if (logout)
              client.Logout();
            else
              client.Disconnect();

            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }
        }
      }
    }

    [Test]
    public void TestDispose_AsyncConnectRunning()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.Profile.AllowInsecureLogin = true;
          client.Profile.Timeout = 5000;

          Assert.IsFalse(client.IsConnected);

          server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");

          var asyncResult = (ImapClient.ConnectAsyncResult)client.BeginConnect("pass");

          server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                       "* CAPABILITY IMAP4rev1\r\n" + 
                                       "$tag OK done\r\n");

          Assert.IsFalse(asyncResult.EndConnectCalled);

          ((IDisposable)client).Dispose();

          Assert.IsFalse(client.IsConnected);

          client.EndConnect(asyncResult);

          Assert.IsTrue(asyncResult.EndConnectCalled);

          Assert.IsFalse(client.IsConnected);
        }
      }
    }

    [Test]
    public void TestEndConnect_InvalidAsyncResult()
    {
      using (ImapPseudoServer2
             server1 = ImapPseudoServer2.StartNew(),
             server2 = ImapPseudoServer2.StartNew()) {
        server1.ThrowIfServerNotStopped = false;
        server2.ThrowIfServerNotStopped = false;

        using (ImapClient
               client1 = new ImapClient(server1.Host, server1.Port, "user1"),
               client2 = new ImapClient(server2.Host, server2.Port, "user2")) {
          client1.Profile.Timeout = 250;
          client2.Profile.Timeout = 250;

          var asyncResult1 = client1.BeginConnect("pass1");
          var asyncResult2 = client2.BeginConnect(new NetworkCredential("user2", "pass2"));

          try {
            client1.EndConnect(asyncResult2);
            Assert.Fail("ArgumentException not thrown");
          }
          catch (ArgumentException) {
          }

          try {
            client2.EndConnect(asyncResult1);
            Assert.Fail("ArgumentException not thrown");
          }
          catch (ArgumentException) {
          }

          try {
            client1.EndConnect(asyncResult1);
          }
          catch (ImapConnectionException) {
            // expected exception
          }

          try {
            client2.EndConnect(asyncResult2);
          }
          catch (ImapConnectionException) {
            // expected exception
          }
        }
      }
    }

    [Test]
    public void TestConnect_BeginConnectRunning()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.ThrowIfServerNotStopped = false;

        using (var client = new ImapClient(server.Host, server.Port, "user")) {
          client.BeginConnect("pass");

          try {
            client.Connect("pass");
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }

          try {
            client.Connect(new NetworkCredential("user", "pass"));
            Assert.Fail("InvalidOperationException not thrown");
          }
          catch (InvalidOperationException) {
          }
        }
      }
    }

    [Test]
    public void TestDisconnect()
    {
      DisposeOrDisconnect(true);
    }

    [Test]
    public void TestDispose()
    {
      DisposeOrDisconnect(false);
    }

    private void DisposeOrDisconnect(bool disconnect)
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        if (disconnect)
          client.Disconnect();
        else
          ((IDisposable)client).Dispose();

        DisconnectedPropertyAssertion(client);
      });
    }

    private void DisconnectedPropertyAssertion(ImapClient client)
    {
      Assert.IsFalse(client.IsConnected);
      Assert.Throws<InvalidOperationException>(() => Assert.IsTrue(client.IsSecureSession));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.OpenedMailbox));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerCapabilities));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerNamespace));
      Assert.Throws<InvalidOperationException>(() => Assert.IsNotNull(client.ServerID));
    }

    [Test]
    public void TestCheckConnected()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        client.Disconnect();

        Assert.IsFalse(client.IsConnected);

        Assert.Throws<InvalidOperationException>(() => client.Refresh());
      });
    }

    [Test]
    public void TestOperationAfterDisconnected()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        client.Disconnect();

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestOperationAfterTimeout()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        client.Timeout = 100;

        try {
          client.Refresh();
        }
        catch (TimeoutException) {
        }

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestOperationAfterInternalError()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueResponse("-ERR\r\n");

        try {
          client.Refresh();
        }
        catch (ImapException) {
        }

        OperationAfterException(client);
      });
    }

    private void OperationAfterException(ImapClient client)
    {
      try {
        client.Refresh();
        Assert.Fail("InvalidOperationException not thrown");
      }
      catch (InvalidOperationException) {
      }

      DisconnectedPropertyAssertion(client);
    }

    [Test]
    public void TestDisconnectedFromServerWithByeResponse()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueUntaggedByeResponse("* BYE server shutdown\r\n");

        try {
          client.Refresh();
          Assert.Fail("ImapErrorResponseException not thrown");
        }
        catch (ImapConnectionException ex) {
          StringAssert.Contains("server shutdown", ex.Message);
        }

        OperationAfterException(client);
      });
    }

    [Test]
    public void TestDisconnectedFromServerWithoutByeResponse()
    {
      using (var server = ImapPseudoServer2.StartNew()) {
        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" +
                                     "$tag OK done\r\n",
                                     true);

        using (var client = new ImapClient(new Uri(string.Format("imap://user@{0}/", server.HostPort)))) {
          client.Timeout = 5000;
          client.Profile.AllowInsecureLogin = true;
          client.Profile.UsingSaslMechanisms = null;
          client.Profile.UseTlsIfAvailable = false;
          client.Connect("pass");

          try {
            client.Refresh();
            Assert.Fail("ImapErrorResponseException not thrown");
          }
          catch (ImapConnectionException) {
          }

          OperationAfterException(client);
        }
      }
    }

    [Test]
    public void TestLogout()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK done\r\n",
                                     true);

        client.Logout();

        DisconnectedPropertyAssertion(client);
      });
    }

    [Test]
    public void TestLogout_InSelectedState()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LIST \"\" INBOX\r\n",
                                     "* LIST () \"\" INBOX\r\n" +
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("SELECT INBOX\r\n",
                                     "$tag OK done\r\n");

        Assert.IsNotNull(client.OpenInbox());
        Assert.IsNotNull(client.OpenedMailbox);
        Assert.IsTrue(client.OpenedMailbox.IsInbox);

        server.EnqueueTaggedResponse("CLOSE\r\n",
                                     "$tag OK done\r\n");
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK done\r\n",
                                     true);

        client.Logout();

        DisconnectedPropertyAssertion(client);
      });
    }

    [Test]
    public void TestConnect_Reconnect()
    {
      TestUtils2.TestAuthenticated(delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("LOGOUT\r\n",
                                     "$tag OK done\r\n",
                                     true);

        client.Logout();

        Assert.IsFalse(client.IsConnected);

        // reconnect
        server.Restart();

        client.Profile.Host = server.Host;
        client.Profile.Port = server.Port;

        server.SetGreetingResponse("* OK [CAPABILITY IMAP4rev1] ImapPseudoServer ready\r\n");
        server.EnqueueTaggedResponse("LOGIN user pass\r\n",
                                     "* CAPABILITY IMAP4rev1\r\n" + 
                                     "$tag OK done\r\n");

        client.Connect("pass");

        Assert.IsTrue(client.IsConnected);
        Assert.IsFalse(client.IsSecureSession);
        Assert.IsNotNull(client.ServerCapabilities);
        Assert.IsNotNull(client.ServerNamespace);
        Assert.IsNotNull(client.ServerID);
      });
    }

    [Test]
    public void TestSerializeBinary_ErrorResponseException()
    {
      TestUtils2.TestAuthenticated(new[] {ImapCapability.Quota},
                                   delegate(ImapPseudoServer2 server, ImapClient client) {
        server.EnqueueTaggedResponse("GETQUOTA ",
                                     "$tag NO no such quota root\r\n");

        try {
          client.GetQuotaUsage("NON-EXISTENT", "STORAGE");
          Assert.Fail("ImapErrorResponseException not thrown");
        }
        catch (ImapErrorResponseException ex) {
          Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapErrorResponseException deserialized) {
            Assert.AreEqual(ex.Message, deserialized.Message);
            Assert.IsNotNull(deserialized.Result);
            Assert.IsNotNull(deserialized.Result.TaggedStatusResponse);
            Assert.AreEqual(ImapResponseCondition.No, deserialized.Result.TaggedStatusResponse.Condition);
            Assert.AreEqual("no such quota root", deserialized.Result.ResponseText);

            var responses = deserialized.Result.ReceivedResponses.ToArray();

            Assert.AreEqual(1, responses.Length);
            Assert.IsInstanceOf<ImapTaggedStatusResponse>(responses[0]);
          });
        }
      });
    }
  }
}
