using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using NUnit.Framework;

using Smdn.Net.Imap4.Protocol;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapMailboxInfoOperationsAppendTests : ImapMailboxInfoTestsBase {
    [Test]
    public void TestAppendMessage_IAppendMessage()
    {
      TestMailbox("INBOX",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        const string uploadingMessage = "MIME-Version: 1.0\r\n\r\ntest message";

        var message = new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage),
                                            new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                                            new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen));

        // APPEND
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");

        mailbox.AppendMessage(message);

        var requested = server.DequeueAll();

        Assert.That(requested,
                    Is.StringContaining(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n", uploadingMessage.Length)));
        Assert.That(requested,
                    Is.StringContaining(uploadingMessage));
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_MailboxNonExistent()
    {
      TestMailbox(null,
                  new[] {ImapMailboxFlag.NonExistent},
                  "INBOX",
      delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        using (var stream = new MemoryStream(new byte[] {0x40})) {
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessage(stream));
        }
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_MailboxUnselectable()
    {
      TestMailbox(null,
                  new[] {ImapMailboxFlag.NoSelect},
                  "INBOX",
      delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        using (var stream = new MemoryStream(new byte[] {0x40})) {
          Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessage(stream));
        }
      });
    }

    [Test]
    public void TestAppendMessage_IAppendMessage_NoTryCreate()
    {
      TestMailbox("NonExistent",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        // APPEND
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => mailbox.AppendMessage(new MemoryStream(new byte[] {0x40})));

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });

        Assert.That(server.DequeueRequest(), Is.StringEnding("APPEND NonExistent {1}\r\n"));
      });
    }

    [Test]
    public void TestAppendMessages()
    {
      TestMailbox("INBOX", AppendMessages);
    }

    [Test]
    public void TestAppendMessages_MultiappendDisabled()
    {
      TestMailbox(new[] {ImapCapability.MultiAppend},
                  null,
                  "INBOX",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.MultiAppend);

        AppendMessages(server, mailbox);
      });
    }

    private void AppendMessages(ImapPseudoServer server, ImapMailboxInfo mailbox)
    {
      const string uploadingMessage1 = "MIME-Version: 1.0\r\n\r\ntest message1";
      const string uploadingMessage2 = "MIME-Version: 1.0\r\n\r\ntest message2";
      const string uploadingMessage3 = "MIME-Version: 1.0\r\n\r\ntest message3";

      var messages = new[] {
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage1),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen)),
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage2),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 31, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Seen)),
        new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage3),
                              new DateTimeOffset(2011, 1, 24, 14, 52, 32, TimeSpan.FromHours(+9.0)),
                              new ImapMessageFlagSet(new[] {"$label3"}, ImapMessageFlag.Seen)),
      };

      // APPEND
      server.EnqueueResponse("+ OK continue\r\n");
      server.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");
      server.EnqueueResponse(string.Empty); // XXX
      server.EnqueueResponse("+ OK continue\r\n");
      server.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3956] APPEND completed\r\n");
      server.EnqueueResponse(string.Empty); // XXX
      server.EnqueueResponse("+ OK continue\r\n");
      server.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3957] APPEND completed\r\n");
      server.EnqueueResponse(string.Empty); // XXX

      mailbox.AppendMessages(messages);

      var requested = server.DequeueAll();

      Assert.That(requested,
                  Is.StringContaining(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n", uploadingMessage1.Length)));
      Assert.That(requested,
                  Is.StringContaining(uploadingMessage1));
      Assert.That(requested,
                  Is.StringContaining(string.Format("APPEND INBOX (\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n", uploadingMessage2.Length)));
      Assert.That(requested,
                  Is.StringContaining(uploadingMessage2));
      Assert.That(requested,
                  Is.StringContaining(string.Format("APPEND INBOX (\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n", uploadingMessage3.Length)));
      Assert.That(requested,
                  Is.StringContaining(uploadingMessage3));
    }

    [Test]
    public void TestAppendMessages_MultiappendCapable()
    {
      TestMailbox(new[] {ImapCapability.MultiAppend},
                  null,
                  "INBOX",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        const string uploadingMessage1 = "MIME-Version: 1.0\r\n\r\ntest message1";
        const string uploadingMessage2 = "MIME-Version: 1.0\r\n\r\ntest message2";
        const string uploadingMessage3 = "MIME-Version: 1.0\r\n\r\ntest message3";

        var messages = new[] {
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage1),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen)),
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage2),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 31, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label2"}, ImapMessageFlag.Seen)),
          new ImapAppendMessage(Encoding.ASCII.GetBytes(uploadingMessage3),
                                new DateTimeOffset(2011, 1, 24, 14, 52, 32, TimeSpan.FromHours(+9.0)),
                                new ImapMessageFlagSet(new[] {"$label3"}, ImapMessageFlag.Seen)),
        };

        // APPEND
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse(string.Empty); // XXX
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse(string.Empty); // XXX
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueResponse(string.Empty); // XXX
        server.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3955,3956,3957] APPEND completed\r\n");

        mailbox.AppendMessages(messages);

        var requested = server.DequeueAll();

        Assert.That(requested,
                    Is.StringContaining(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n", uploadingMessage1.Length)));
        Assert.That(requested,
                    Is.StringContaining(uploadingMessage1));
        Assert.That(requested,
                    Is.Not.StringContaining(string.Format("APPEND INBOX (\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n", uploadingMessage2.Length)));
        Assert.That(requested,
                    Is.StringContaining(string.Format("(\\Seen $label2) \"24-Jan-2011 14:52:31 +0900\" {{{0}}}\r\n", uploadingMessage2.Length)));
        Assert.That(requested,
                    Is.StringContaining(uploadingMessage2));
        Assert.That(requested,
                    Is.Not.StringContaining(string.Format("APPEND INBOX (\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n", uploadingMessage3.Length)));
        Assert.That(requested,
                    Is.StringContaining(string.Format("(\\Seen $label3) \"24-Jan-2011 14:52:32 +0900\" {{{0}}}\r\n", uploadingMessage3.Length)));
        Assert.That(requested,
                    Is.StringContaining(uploadingMessage3));
      });
    }

    [Test]
    public void TestAppendMessages_NonExistentMailbox()
    {
      TestMailbox(null,
                  new[] {ImapMailboxFlag.NonExistent},
                  "INBOX",
      delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessages(messages));
      });
    }

    [Test]
    public void TestAppendMessages_UnselectableMailbox()
    {
      TestMailbox(null,
                  new[] {ImapMailboxFlag.NoSelect},
                  "INBOX",
      delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        Assert.Throws<ImapProtocolViolationException>(() => mailbox.AppendMessages(messages));
      });
    }

    [Test]
    public void TestAppendMessages_NoTryCreate()
    {
      TestMailbox("NonExistent",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        // APPEND
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");

        var messages = new[] {
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
          new ImapAppendMessage(new MemoryStream(new byte[] {0x40})),
        };

        var ex = Assert.Throws<ImapMailboxNotFoundException>(() => mailbox.AppendMessages(messages));

        Assert.IsNotNull(ex.Mailbox);
        Assert.AreEqual("NonExistent", ex.Mailbox);

        Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
          Assert.IsNotNull(deserialized.Mailbox);
          Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
        });

        Assert.That(server.DequeueRequest(), Is.StringEnding("APPEND NonExistent {1}\r\n"));
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfoOfSameSession()
    {
      TestMailbox("INBOX", delegate(ImapPseudoServer server, ImapMailboxInfo sourceMailbox) {
        // LIST
        server.EnqueueTaggedResponse("* LIST () \"\" dest\r\n" +
                                     "$tag OK done\r\n");

        var destMailbox = sourceMailbox.Client.GetMailbox("dest");

        Assert.That(server.DequeueRequest(), Is.StringEnding("LIST \"\" dest\r\n"));

        // SELECT
        server.EnqueueTaggedResponse("* EXISTS 3\r\n" +
                                     "* OK [UIDVALIDITY 23]\r\n" +
                                     "$tag OK [READ-WRITE] done\r\n");

        using (var openedSourceMailbox = sourceMailbox.Open()) {
          Assert.That(server.DequeueRequest(), Is.StringEnding("SELECT INBOX\r\n"));

          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");

          var message = openedSourceMailbox.GetMessageBySequence(1L);

          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1 (UID)\r\n"));

          // UID COPY
          server.EnqueueTaggedResponse("$tag OK done\r\n");

          destMailbox.AppendMessage(message);

          Assert.That(server.DequeueRequest(), Is.StringEnding("UID COPY 3 dest\r\n"));

          // CLOSE
          server.EnqueueTaggedResponse("$tag OK done\r\n");
        }
      });
    }

    [Test]
    public void TestAppendMessage_MessageInfoOfDifferentSession()
    {
      TestMailbox("dest", delegate(ImapPseudoServer destServer, ImapMailboxInfo destMailbox) {
        var selectResp = "* EXISTS 3\r\n" +
          "* OK [UIDVALIDITY 23]\r\n" +
            "$tag OK [READ-WRITE] done\r\n";

        TestUtils.TestOpenedMailbox("INBOX",
                                    selectResp,
                                    delegate(ImapPseudoServer sourceServer, ImapOpenedMailboxInfo sourceMailbox) {
          // FETCH
          sourceServer.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                             "$tag OK done\r\n");

          var message = sourceMailbox.GetMessageBySequence(1L);

          Assert.That(sourceServer.DequeueRequest(), Is.StringEnding("FETCH 1 (UID)\r\n"));

          // FETCH (source)
          sourceServer.EnqueueTaggedResponse("* FETCH 1 (" +
                                             "FLAGS (\\Answered $label1) " +
                                             "INTERNALDATE \"25-Jan-2011 15:29:06 +0900\" " +
                                             "RFC822.SIZE 12 " +
                                             "BODY[] {12}\r\ntest message)\r\n" +
                                             "$tag OK done\r\n");
          // APPEND (dest)
          destServer.EnqueueResponse("+ OK continue\r\n");
          destServer.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");

          destMailbox.AppendMessage(message);

          Assert.That(sourceServer.DequeueRequest(), Is.StringEnding("UID FETCH 3 (RFC822.SIZE INTERNALDATE FLAGS BODY.PEEK[]<0.10240>)\r\n"));

          Assert.That(destServer.DequeueRequest(),
                      Is.StringEnding("APPEND dest (\\Answered $label1) \"25-Jan-2011 15:29:06 +0900\" {12}\r\n"));
          Assert.That(destServer.DequeueAll(),
                      Is.StringStarting("test message"));
        });
      });
    }

    [Test]
    public void TestAppendMessage_Action()
    {
      AppendMessageAction(false);
    }

    [Test]
    public void TestAppendMessage_Action_LengthSpecified()
    {
      AppendMessageAction(true);
    }

    private void AppendMessageAction(bool specifyLength)
    {
      TestMailbox("INBOX",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        const string uploadingMessage = "MIME-Version: 1.0\r\n\r\ntest message";

        var length = uploadingMessage.Length;
        var flags = new ImapMessageFlagSet(new[] {"$label1"}, ImapMessageFlag.Seen);
        var internalDate = new DateTimeOffset(2011, 1, 24, 14, 52, 30, TimeSpan.FromHours(+9.0));
        var writeUploadMessage = new Action<Stream>(delegate(Stream stream) {
          var writer = new StreamWriter(stream, Encoding.ASCII);

          writer.Write(uploadingMessage);
          writer.Flush();
        });

        // APPEND
        server.EnqueueResponse("+ OK continue\r\n");
        server.EnqueueTaggedResponse("$tag OK [APPENDUID 38505 3955] APPEND completed\r\n");

        if (specifyLength)
          mailbox.AppendMessage(length,
                                internalDate,
                                flags,
                                writeUploadMessage);
        else
          mailbox.AppendMessage(internalDate,
                                flags,
                                writeUploadMessage);

        Assert.That(server.DequeueRequest(),
                    Is.StringEnding(string.Format("APPEND INBOX (\\Seen $label1) \"24-Jan-2011 14:52:30 +0900\" {{{0}}}\r\n",
                                      length)));
        Assert.That(server.DequeueAll(),
                    Is.StringStarting(uploadingMessage));
      });
    }

    [Test]
    public void TestAppendMessage_Action_MailboxNonExistent()
    {
      TestMailbox(null,
                  new[] {ImapMailboxFlag.NonExistent},
      "INBOX",
      delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        Assert.IsFalse(mailbox.Exists);

        Assert.Throws<ImapProtocolViolationException>(delegate {
          mailbox.AppendMessage(delegate(Stream stream) {
          });
        });
      });
    }

    [Test]
    public void TestAppendMessage_Action_MailboxUnselectable()
    {
      TestMailbox(null,
                  new[] {ImapMailboxFlag.NoSelect},
      "INBOX",
      delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        Assert.IsTrue(mailbox.IsUnselectable);

        Assert.Throws<ImapProtocolViolationException>(delegate {
          mailbox.AppendMessage(delegate(Stream stream) {
          });
        });
      });
    }

    [Test]
    public void TestAppendMessage_Action_LengthSpecified_OutOfRange()
    {
      TestMailbox("INBOX",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {

        Assert.Throws<ArgumentOutOfRangeException>(delegate {
          mailbox.AppendMessage(-1L, delegate(Stream stream) {
          });
        });
      });
    }

    [Test]
    public void TestAppendMessage_Action_NoTryCreate()
    {
      TestMailbox("NonExistent",
                  delegate(ImapPseudoServer server, ImapMailboxInfo mailbox) {
        // APPEND
        server.EnqueueTaggedResponse("$tag NO [TRYCREATE] mailbox not exist\r\n");

        try {
          mailbox.AppendMessage(1, delegate(Stream stream) {
            stream.WriteByte(0x40);
          });
        }
        catch (ImapMailboxNotFoundException ex) {
          Assert.IsNotNull(ex.Mailbox);
          Assert.AreEqual("NonExistent", ex.Mailbox);

          Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMailboxNotFoundException deserialized) {
            Assert.IsNotNull(deserialized.Mailbox);
            Assert.AreEqual(ex.Mailbox, deserialized.Mailbox);
          });
        }

        Assert.That(server.DequeueRequest(), Is.StringEnding("APPEND NonExistent {1}\r\n"));
      });
    }
  }
}