using System;
using System.Collections.Generic;
using NUnit.Framework;

namespace Smdn.Net.Imap4.Client {
  [TestFixture]
  public class ImapOpenedMailboxInfoOperationsFetchingTests {
    [Test]
    public void TestGetMessages()
    {
      var selectResp =
        "* 5 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages();

        Assert.AreEqual(5L, mailbox.ExistMessageCount);

        var enumeratedMessages = new ImapMessageInfo[4];

        using (var enumerator = messages.GetEnumerator()) {
          // NOOP
          server.EnqueueTaggedResponse("* 5 EXPUNGE\r\n" +
                                       "* 4 EXISTS\r\n" + 
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "* FETCH 4 (UID 9)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("NOOP\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1:4 (UID)\r\n"));

          Assert.AreEqual(4L, mailbox.ExistMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          enumeratedMessages[0] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          enumeratedMessages[1] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");

          enumeratedMessages[2] = enumerator.Current;

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#3 non-null");
          Assert.AreEqual(9L, enumerator.Current.Uid, "#3 uid");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#3 seq");

          enumeratedMessages[3] = enumerator.Current;

          Assert.IsFalse(enumerator.MoveNext());
        }

        /*
         * re-enumerate
         */
        Assert.AreEqual(4L, mailbox.ExistMessageCount);

        using (var enumerator = messages.GetEnumerator()) {
          // NOOP
          server.EnqueueTaggedResponse("* 2 EXPUNGE\r\n" +
                                       "* 1 EXPUNGE\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 7)\r\n" +
                                       "* FETCH 2 (UID 9)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("NOOP\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1:2 (UID)\r\n"));

          Assert.AreEqual(2L, mailbox.ExistMessageCount);

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(7L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.AreSame(enumeratedMessages[2], enumerator.Current, "returns same instance #0");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(9L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

          Assert.AreSame(enumeratedMessages[3], enumerator.Current, "returns same instance #1");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_NoMessageExists()
    {
      var selectResp =
        "* 2 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // NOOP
        server.EnqueueTaggedResponse("* 1 EXPUNGE\r\n" +
                                     "* 1 EXPUNGE\r\n" +
                                     "$tag OK done\r\n");

        var messages = new List<ImapMessageInfo>(mailbox.GetMessages());

        Assert.AreEqual(0L, mailbox.ExistMessageCount);

        Assert.AreEqual(0, messages.Count);

        Assert.That(server.DequeueRequest(), Is.StringEnding("NOOP\r\n"));
      });
    }

    [Test]
    public void TestGetMessageByUid()
    {
      GetMessageByUidOrSequence(true);
    }

    [Test]
    public void TestGetMessageBySequence()
    {
      GetMessageByUidOrSequence(false);
    }

    private void GetMessageByUidOrSequence(bool uid)
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var message = uid
          ? mailbox.GetMessageByUid(3L)
          : mailbox.GetMessageBySequence(1L);

        Assert.IsNotNull(message);
        Assert.AreEqual(3L, message.Uid);
        Assert.AreEqual(1L, message.Sequence);
        Assert.AreEqual(23L, message.UidValidity);
        Assert.AreSame(mailbox, message.Mailbox);

        if (uid)
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 3 (UID)\r\n"));
        else
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1 (UID)\r\n"));

        // FETCH
        server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                     "$tag OK done\r\n");

        var messageSecond = uid
          ? mailbox.GetMessageByUid(3L)
          : mailbox.GetMessageBySequence(1L);

        Assert.AreSame(message, messageSecond, "returns same instance");

        if (uid)
          Assert.That(server.DequeueRequest(), Is.StringEnding("UID FETCH 3 (UID)\r\n"));
        else
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1 (UID)\r\n"));
      });
    }

    [Test]
    public void TestGetMessageByUid_NotFound()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse("$tag OK done\r\n");

        try {
          mailbox.GetMessageByUid(3L);
          Assert.Fail("ImapMessageNotFoundException not thrown");
        }
        catch (ImapMessageNotFoundException ex) {
          Assert.IsNotNull(ex.SequenceOrUidSet);
          Assert.AreEqual("3", ex.SequenceOrUidSet.ToString());

          Smdn.Net.TestUtils.SerializeBinary(ex, delegate(ImapMessageNotFoundException deserialized) {
            Assert.IsNotNull(deserialized.SequenceOrUidSet);
            Assert.AreEqual("3", deserialized.SequenceOrUidSet.ToString());
          });
        }
      });
    }

    [Test]
    public void TestGetMessageBySequence_GreaterThanExistCount()
    {
      var selectResp =
        "* 0 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ArgumentOutOfRangeException>(() => mailbox.GetMessageBySequence(1L));
      });
    }

    [Test]
    public void TestGetMessageBySequence_NoResponse()
    {
      var selectResp =
        "* 1 EXISTS\r\n" +
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // FETCH
        server.EnqueueTaggedResponse("$tag NO such message\r\n");

        Assert.Throws<ImapErrorResponseException>(() => mailbox.GetMessageBySequence(1L));
      });
    }

    [Test]
    public void TestGetMessages_ByUids()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(1L, 3L, 5L);

        using (var enumerator = messages.GetEnumerator()) {
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 1)\r\n" +
                                       "* FETCH 3 (UID 3)\r\n" +
                                       "* FETCH 5 (UID 5)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1,3,5 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(1L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#1 uid");
          Assert.AreEqual(3L, enumerator.Current.Sequence, "#1 seq");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(5L, enumerator.Current.Uid, "#2 uid");
          Assert.AreEqual(5L, enumerator.Current.Sequence, "#2 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
          "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, GetMessagesWithSearchCriteria);
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresDisabled()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        mailbox.Client.DisabledCapabilities.Add(ImapCapability.Searchres);

        GetMessagesWithSearchCriteria(server, mailbox);
      });
    }

    private void GetMessagesWithSearchCriteria(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox)
    {
      var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

      using (var enumerator = messages.GetEnumerator()) {
        // SEARCH
        server.EnqueueTaggedResponse("* SEARCH 1 2 3\r\n" +
                                     "$tag OK done\r\n");
        // FETCH
        server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                     "* FETCH 2 (UID 5)\r\n" +
                                     "* FETCH 3 (UID 7)\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(enumerator.MoveNext());

        Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH RECENT\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1:3 (UID)\r\n"));

        Assert.IsNotNull(enumerator.Current, "#0 non-null");
        Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
        Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

        Assert.IsTrue(enumerator.MoveNext());

        Assert.IsNotNull(enumerator.Current, "#1 non-null");
        Assert.AreEqual(5L, enumerator.Current.Uid, "#1 uid");
        Assert.AreEqual(2L, enumerator.Current.Sequence, "#1 seq");

        Assert.IsTrue(enumerator.MoveNext());

        Assert.IsNotNull(enumerator.Current, "#2 non-null");
        Assert.AreEqual(7L, enumerator.Current.Uid, "#2 uid");
        Assert.AreEqual(3L, enumerator.Current.Sequence, "#2 seq");

        Assert.IsFalse(enumerator.MoveNext());
      }

      /*
       * re-enumerate
       */
      using (var enumerator = messages.GetEnumerator()) {
        // SEARCH
        server.EnqueueTaggedResponse("* SEARCH 4\r\n" +
                                     "$tag OK done\r\n");
        // FETCH
        server.EnqueueTaggedResponse("* FETCH 4 (UID 8)\r\n" +
                                     "$tag OK done\r\n");

        Assert.IsTrue(enumerator.MoveNext());

        Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH RECENT\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 4 (UID)\r\n"));

        Assert.IsNotNull(enumerator.Current, "#0 non-null");
        Assert.AreEqual(8L, enumerator.Current.Uid, "#0 uid");
        Assert.AreEqual(4L, enumerator.Current.Sequence, "#0 seq");

        Assert.IsFalse(enumerator.MoveNext());
      }
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Subject("test"), System.Text.Encoding.UTF8);

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueResponse("+ continue\r\n");
          server.EnqueueTaggedResponse("$tag NO [BADCHARSET (US-ASCII Shift_JIS)] charset not supported\r\n");

          var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(() => enumerator.MoveNext());

          Assert.AreEqual("utf-8", ex.EncodingName);
          StringAssert.Contains("US-ASCII", ex.Message);
          StringAssert.Contains("Shift_JIS", ex.Message);

          Assert.That(server.DequeueAll(), Is.StringEnding("SEARCH CHARSET utf-8 SUBJECT {4}\r\ntest\r\n"));
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueTaggedResponse("$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH RETURN (SAVE) RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH $ (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable_NoNotSaved()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // ESEARCH
          server.EnqueueTaggedResponse("$tag NO [NOTSAVED] not saved\r\n");
          // SEARCH
          server.EnqueueTaggedResponse("* SEARCH 1\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH RETURN (SAVE) RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(3L, enumerator.Current.Uid, "#0 uid");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#0 seq");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_SearchresCapable_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ESearch, ImapCapability.Searchres},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Subject("test"),
                                           System.Text.Encoding.GetEncoding("Shift_JIS"));

        using (var enumerator = messages.GetEnumerator()) {
          // ESEARCH
          server.EnqueueResponse("+ continue\r\n");
          server.EnqueueTaggedResponse("$tag NO [BADCHARSET] charset not supported\r\n");

          var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(() => enumerator.MoveNext());

          Assert.AreEqual("shift_jis", ex.EncodingName);

          Assert.That(server.DequeueAll(), Is.StringEnding("SEARCH RETURN (SAVE) CHARSET shift_jis SUBJECT {4}\r\ntest\r\n"));
        }
      });
    }

    [Test]
    public void TestGetMessages_WithSearchCriteria_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessages(ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueTaggedResponse("* SEARCH\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsFalse(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH RECENT\r\n"));
        }
      });
    }

    [Test]
    public void TestGetSortedMessages()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Sort},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SORT
          server.EnqueueTaggedResponse("* SORT 2 4 1\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 2)\r\n" +
                                       "* FETCH 2 (UID 4)\r\n" +
                                       "* FETCH 4 (UID 8)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("SORT (ARRIVAL) utf-8 RECENT\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 2,4,1 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.AreEqual(2L, enumerator.Current.Sequence, "#0 seq");
          Assert.AreEqual(4L, enumerator.Current.Uid, "#0 uid");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.AreEqual(4L, enumerator.Current.Sequence, "#1 seq");
          Assert.AreEqual(8L, enumerator.Current.Uid, "#1 uid");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.AreEqual(1L, enumerator.Current.Sequence, "#2 seq");
          Assert.AreEqual(2L, enumerator.Current.Uid, "#2 uid");

          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetSortedMessages_SortIncapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetSortedMessages(ImapSortCriteria.Subject, ImapSearchCriteria.Recent));
      });
    }

    [Test]
    public void TestGetSortedMessages_SortCriteriaIncapable_SortDisplay()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Sort}, "INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetSortedMessages(ImapSortCriteria.DisplayTo, ImapSearchCriteria.Recent));
      });
    }

    [Test]
    public void TestGetSortedMessages_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Sort},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Arrival, ImapSearchCriteria.Recent);

        using (var enumerator = messages.GetEnumerator()) {
          // SORT
          server.EnqueueTaggedResponse("* SORT\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsFalse(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("SORT (ARRIVAL) utf-8 RECENT\r\n"));
        }
      });
    }

    [Test]
    public void TestGetThreadedMessages()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // THREAD
        server.EnqueueTaggedResponse("* THREAD (2)(3 6 (4 23)(44))\r\n" +
                                     "$tag OK done\r\n");
        // FETCH
        server.EnqueueTaggedResponse("* FETCH 2 (UID 2)\r\n" +
                                     "* FETCH 3 (UID 3)\r\n" +
                                     "* FETCH 6 (UID 6)\r\n" +
                                     "* FETCH 4 (UID 4)\r\n" +
                                     "* FETCH 23 (UID 23)\r\n" +
                                     "* FETCH 44 (UID 44)\r\n" +
                                     "$tag OK done\r\n");

        var thread = mailbox.GetThreadedMessages(ImapThreadingAlgorithm.References);

        Assert.That(server.DequeueRequest(), Is.StringEnding("THREAD REFERENCES utf-8 ALL\r\n"));
        Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 2,3,6,4,23,44 (UID)\r\n"));

        Assert.IsNotNull(thread);
        Assert.IsTrue(thread.IsRoot);
        Assert.IsNull(thread.Parent);
        Assert.AreEqual(0, thread.Depth);
        Assert.Throws<NotSupportedException>(() => Assert.IsNull(thread.Message));

        /*
          -- 2
          -- 3
             `-- 6
                 |-- 4
                 |   `-- 23
                 |
                 `-- 44
        */
        var expectedMessageSequences = new[] {
          2L, 3L, 6L, 4L, 23L, 44L,
        };
        var expectedDepths = new[] {
          1L, 1L, 2L, 3L, 4L, 3L,
        };
        var expectedChildrenCounts = new[] {
          0L, 1L, 2L, 1L, 0L, 0L,
        };
        var traversedMessageCount = 0;

        thread.Traverse(delegate(ImapMessageInfoThread t) {
          Assert.IsFalse(t.IsRoot, "IsRoot #{0}", traversedMessageCount);
          Assert.IsNotNull(t.Parent, "Parent #{0}", traversedMessageCount);
          Assert.AreEqual(expectedDepths[traversedMessageCount], t.Depth, "Depth #{0}", traversedMessageCount);
          Assert.IsNotNull(t.Children, "Children #{0}", traversedMessageCount);
          Assert.AreEqual(expectedChildrenCounts[traversedMessageCount], t.Children.Count, "Children.Count #{0}", traversedMessageCount);
          Assert.AreEqual(expectedMessageSequences[traversedMessageCount], t.Message.Sequence, "Message.Sequence #{0}", traversedMessageCount);

          traversedMessageCount++;
        });

        Assert.AreEqual(6, traversedMessageCount);
      });
    }

    [Test]
    public void TestGetThreadedMessages_ThreadingAlgorithmIncapable_References()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessages(ImapThreadingAlgorithm.References));
      });
    }

    [Test]
    public void TestGetThreadedMessages_ThreadingAlgorithmIncapable_OrderedSubject()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX", selectResp, delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        Assert.Throws<ImapIncapableException>(() => mailbox.GetThreadedMessages(ImapThreadingAlgorithm.OrderedSubject));
      });
    }

    [Test]
    public void TestGetThreadedMessages_NothingMatched()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        // THREAD
        server.EnqueueTaggedResponse("* THREAD\r\n" +
                                     "$tag OK done\r\n");

        var thread = mailbox.GetThreadedMessages(ImapThreadingAlgorithm.References);

        Assert.That(server.DequeueRequest(), Is.StringEnding("THREAD REFERENCES utf-8 ALL\r\n"));

        Assert.IsNotNull(thread);
        Assert.IsTrue(thread.IsRoot);
        Assert.IsNull(thread.Parent);
        Assert.AreEqual(0, thread.Depth);
        Assert.Throws<NotSupportedException>(() => Assert.IsNull(thread.Message));
        Assert.IsEmpty(thread.Children);

        thread.Traverse(delegate(ImapMessageInfoThread t) {
          Assert.Fail("children traversed");
        });
      });
    }

    [Test]
    public void TestGetThreadedMessages_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.ThreadReferences},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {

        // THREAD
        server.EnqueueResponse("+ continue\r\n");
        server.EnqueueTaggedResponse("$tag NO [BADCHARSET] charset not supported\r\n");

        var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(delegate {
          mailbox.GetThreadedMessages(ImapThreadingAlgorithm.References,
                                      ImapSearchCriteria.Subject("test"),
                                      System.Text.Encoding.GetEncoding("Shift_JIS"));
        });

        Assert.AreEqual("shift_jis", ex.EncodingName);

        Assert.That(server.DequeueAll(), Is.StringEnding("THREAD REFERENCES shift_jis SUBJECT {4}\r\ntest\r\n"));
      });
    }

    [Test]
    public void TestGetSortedMessages_NoBadCharset()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.Sort},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetSortedMessages(ImapSortCriteria.Date,
                                                 ImapSearchCriteria.Subject("test"),
                                                 System.Text.Encoding.GetEncoding("Shift_JIS"));

        using (var enumerator = messages.GetEnumerator()) {
          // SORT
          server.EnqueueResponse("+ continue\r\n");
          server.EnqueueTaggedResponse("$tag NO [BADCHARSET] charset not supported\r\n");

          var ex = Assert.Throws<Smdn.Formats.EncodingNotSupportedException>(() => enumerator.MoveNext());

          Assert.AreEqual("shift_jis", ex.EncodingName);

          Assert.That(server.DequeueAll(), Is.StringEnding("SORT (DATE) shift_jis SUBJECT {4}\r\ntest\r\n"));
        }
      });
    }

    [Test]
    public void TestGetMessagesGmailSearch()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
        "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox(new[] {ImapCapability.GimapGmExt1},
                                  "INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var messages = mailbox.GetMessagesGmailSearch("from:太郎 has:attachment");

        using (var enumerator = messages.GetEnumerator()) {
          // SEARCH
          server.EnqueueResponse("+ OK\r\n");
          server.EnqueueTaggedResponse("* SEARCH 1 2 3\r\n" +
                                       "$tag OK done\r\n");
          // FETCH
          server.EnqueueTaggedResponse("* FETCH 1 (UID 3)\r\n" +
                                       "* FETCH 2 (UID 5)\r\n" +
                                       "* FETCH 3 (UID 7)\r\n" +
                                       "$tag OK done\r\n");

          Assert.IsTrue(enumerator.MoveNext());

          Assert.That(server.DequeueRequest(), Is.StringEnding("SEARCH CHARSET utf-8 X-GM-RAW {26}\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("from:太郎 has:attachment\r\n"));
          Assert.That(server.DequeueRequest(), Is.StringEnding("FETCH 1:3 (UID)\r\n"));

          Assert.IsNotNull(enumerator.Current, "#0 non-null");
          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#1 non-null");
          Assert.IsTrue(enumerator.MoveNext());

          Assert.IsNotNull(enumerator.Current, "#2 non-null");
          Assert.IsFalse(enumerator.MoveNext());
        }
      });
    }

    [Test]
    public void TestGetMessagesGmailSearch_GimapGmExt1Incapable()
    {
      var selectResp =
        "* OK [UIDVALIDITY 23]\r\n" +
          "$tag OK [READ-WRITE] done\r\n";

      TestUtils.TestOpenedMailbox("INBOX",
                                  selectResp,
                                  delegate(ImapPseudoServer server, ImapOpenedMailboxInfo mailbox) {
        var ex = Assert.Throws<ImapIncapableException>(() => mailbox.GetMessagesGmailSearch("from:太郎 has:attachment"));

        CollectionAssert.Contains(ex.RequiredCapabilities, ImapCapability.GimapGmExt1);
      });
    }
  }
}

