using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Runtime.Remoting.Messaging;
using System.Threading;
using System.Text;
using System.Text.RegularExpressions;

using Smdn.Net;

namespace Smdn.Net.Imap4 {
  public class ImapPseudoServer2 : IDisposable {
    private class Response {
      public string ExpectedCommand {
        get { return expectedCommand; }
      }

      public string ResponseText {
        get { return responseText; }
      }

      public Action ActionBeforeRespond {
        get { return actionBeforeRespond; }
      }

      public Action ActionAfterRespond {
        get { return actionAfterRespond; }
      }

      public bool IsByeResponse {
        get; set;
      }

      public bool IsPushResponse {
        get; set;
      }

      public Response(string expectedCommand,
                      string responseText,
                      Action actionBeforeRespond,
                      Action actionAfterRespond)
      {
        this.expectedCommand = expectedCommand;
        this.responseText = responseText;
        this.actionBeforeRespond = actionBeforeRespond;
        this.actionAfterRespond = actionAfterRespond;

        IsByeResponse = false;
        IsPushResponse = false;
      }

      private readonly string expectedCommand;
      private readonly string responseText;
      private readonly Action actionBeforeRespond;
      private readonly Action actionAfterRespond;
    }

    public IPEndPoint ServerEndPoint {
      get { return serverEndPoint; }
    }

    public string Host {
      get { return ServerEndPoint.Address.ToString(); }
    }

    public int Port {
      get { return ServerEndPoint.Port; }
    }

    public string HostPort {
      get { return string.Format("{0}:{1}", Host, Port); }
    }

    public bool ThrowIfServerNotStopped {
      get; set;
    }

    public ImapPseudoServer2()
    {
      ThrowIfServerNotStopped = true;

      Initialize();
    }

    private void Initialize()
    {
      server = new TcpListener(new IPEndPoint(IPAddress.Loopback, 0));
      server.Start();

      serverEndPoint = server.LocalEndpoint as IPEndPoint;

      responseTag = 0;

      SetGreetingResponse("* OK ready\r\n");

#if !NET_4_0
      clientRequestWaitHandle = new ManualResetEvent(false);
#endif
    }

    void IDisposable.Dispose()
    {
      if (acceptClientAsyncResult != null) {
        var acceptClient = (Action)(acceptClientAsyncResult as AsyncResult).AsyncDelegate;

        if (acceptClientAsyncResult.AsyncWaitHandle.WaitOne(500))
          acceptClient.EndInvoke(acceptClientAsyncResult);
        else if (ThrowIfServerNotStopped)
          throw new ApplicationException("server not stopped");

        acceptClientAsyncResult = null;
      }

      if (server != null) {
        server.Stop();
        server = null;
      }

#if !NET_4_0
      if (clientRequestWaitHandle != null) {
        clientRequestWaitHandle.Close();
        clientRequestWaitHandle = null;
      }
#endif

      if (serverProcException != null) {
        Console.Error.WriteLine("unhandled server exception:{0}{1}", Environment.NewLine, serverProcException);

        throw new ApplicationException("exception not handled", serverProcException);
      }
    }

    public void Start()
    {
      if (acceptClientAsyncResult != null)
        throw new InvalidOperationException("already started");

      var acceptClient = new Action(AcceptClient);

      acceptClientAsyncResult = acceptClient.BeginInvoke(null, null);
    }

    public void Restart()
    {
      (this as IDisposable).Dispose();

      Initialize();

      Start();
    }

    public static ImapPseudoServer2 StartNew()
    {
      var server = new ImapPseudoServer2();

      server.Start();

      return server;
    }

    public static ImapPseudoServer2 StartNotAccepting()
    {
      var server = new ImapPseudoServer2();

      server.server.Start();

      return server;
    }

    public void SetGreetingResponse(string greeting)
    {
      SetGreetingResponse(greeting, null, null, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond)
    {
      SetGreetingResponse(greeting, actionBeforeRespond, actionAfterRespond, false);
    }

    public void SetGreetingResponse(string greeting,
                                    Action actionBeforeRespond,
                                    Action actionAfterRespond,
                                    bool isByeResponse)
    {
      greetingResponse = new Response(null, greeting, actionBeforeRespond, actionAfterRespond);

      greetingResponse.IsByeResponse = isByeResponse;
    }

    public void EnqueueTaggedResponse(string responseText)
    {
      EnqueueTaggedResponse(null, responseText, null, null);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText)
    {
      EnqueueTaggedResponse(expectedCommand,
                            responseText,
                            null,
                            null,
                            false);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      bool isByeResponse)
    {
      EnqueueTaggedResponse(expectedCommand,
                            responseText,
                            null,
                            null,
                            isByeResponse);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      Action actionBeforeRespond,
                                      Action actionAfterRespond)
    {
      EnqueueTaggedResponse(expectedCommand,
                            responseText,
                            actionBeforeRespond,
                            actionAfterRespond,
                            false);
    }

    public void EnqueueTaggedResponse(string expectedCommand,
                                      string responseText,
                                      Action actionBeforeRespond,
                                      Action actionAfterRespond,
                                      bool isByeResponse)
    {
#if DEBUG
      if (!responseText.Contains("$tag"))
        throw new ArgumentException("is not tagged response", "responseText");
#endif

      responseText = responseText.Replace("$tag", (responseTag++).ToString("x4"));

      var response = new Response(expectedCommand,
                                  responseText,
                                  actionBeforeRespond,
                                  actionAfterRespond);

      response.IsByeResponse = isByeResponse;

      responseQueue.Enqueue(response);
    }

    public void EnqueueUntaggedByeResponse(string responseText)
    {
#if DEBUG
      if (!(responseText.StartsWith("* BYE") || responseText.Contains("\r\n* BYE")))
        throw new ArgumentException("is not BYE response", "responseText");
#endif

      var response = new Response(null,
                                  responseText,
                                  null,
                                  null);

      response.IsByeResponse = true;

      responseQueue.Enqueue(response);
    }

    public void EnqueueResponse(string responseText)
    {
      EnqueueResponse(null,
                      responseText,
                      null,
                      null);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText)
    {
      EnqueueResponse(expectedCommand,
                      responseText,
                      null,
                      null);
    }

    public void EnqueueResponse(string expectedCommand,
                                string responseText,
                                Action actionBeforeRespond,
                                Action actionAfterRespond)
    {
      var response = new Response(expectedCommand,
                                  responseText,
                                  actionBeforeRespond,
                                  actionAfterRespond);

      responseQueue.Enqueue(response);
    }

    public void EnqueuePushResponse(string responseText)
    {
      var response = new Response(null,
                                  responseText,
                                  null,
                                  null);

      response.IsPushResponse = true;

      responseQueue.Enqueue(response);
    }

    // TODO: rename
    public string DequeueRequest()
    {
      lock (commandQueue) {
        if (0 < commandQueue.Count)
          return commandQueue.Dequeue();
        else
          return null;
      }
    }

    // TODO: rename
    public string DequeueAll()
    {
      var ret = new StringBuilder();

      lock (commandQueue) {
        for (;;) {
          if (commandQueue.Count == 0)
            break;

          ret.Append(commandQueue.Dequeue());
        }
      }

      return ret.ToString();
    }

    public void WaitForClientRequest()
    {
      WaitForClientRequest(TimeSpan.FromSeconds(5));
    }

    public void WaitForClientRequest(TimeSpan timeout)
    {
      clientRequestWaitHandle.Reset();

#if NET_4_0
      clientRequestWaitHandle.Wait(timeout);
#else
      clientRequestWaitHandle.WaitOne(timeout);
#endif
    }

    private void AcceptClient()
    {
      Console.WriteLine("listening...");

      var client = server.AcceptTcpClient();

      Console.WriteLine("client accepted");

      client.ReceiveTimeout = 100;

      try {
        Console.WriteLine("server process started");

        ServerProc(client);

        Console.WriteLine("server process successfully finished");
      }
      catch (Exception ex) {
        serverProcException = ex;

        Console.WriteLine("server process stopped by {0}", ex.GetType().FullName);
      }

      client.Close();

      Console.WriteLine("connection closed");
    }

    private void ServerProc(TcpClient client)
    {
      var commandStream = client.GetStream();

      Send(greetingResponse, commandStream);

      if (greetingResponse.IsByeResponse)
        return;

      var commandReceiveBuffer = new byte[0x1000];
      var commandBuffer = new StringBuilder(0x1000);
      long literalOctetsToBeSent = 0L;

      for (;;) {
        if (!IsConnected(client))
          return;

        if (0 < responseQueue.Count && responseQueue.Peek().IsPushResponse) {
          Send(responseQueue.Dequeue(), commandStream);
          continue;
        }

        /*
         * wait for command sent
         */
        try {
          var commandLength = commandStream.Read(commandReceiveBuffer, 0, commandReceiveBuffer.Length);

          if (commandLength == 0 && !IsConnected(client))
            return;

          var received = NetworkTransferEncoding.Transfer8Bit.GetString(commandReceiveBuffer, 0, commandLength);

          Console.Write("C:\t");
          Console.Write(received);

          commandBuffer.Append(received);

          literalOctetsToBeSent -= commandLength;

          if (0 < literalOctetsToBeSent)
            continue;
        }
        catch (IOException ex) {
          if (!client.Connected)
            return; // disconnected

          var exSocket = ex.InnerException as SocketException;

          if (exSocket != null &&
              (exSocket.SocketErrorCode == SocketError.WouldBlock || exSocket.SocketErrorCode == SocketError.TimedOut)) {
            continue;
          }

          throw;
        }

        var command = commandBuffer.ToString();

        var literalSizeMatches = literalSizeRegex.Matches(command);

        if (0 < literalSizeMatches.Count) {
          var literalSizeMatch = literalSizeMatches[literalSizeMatches.Count - 1]; // last match

          literalOctetsToBeSent = long.Parse(literalSizeMatch.Groups[2].Value);

          if (!"+".Equals(literalSizeMatch.Groups[1].Value, StringComparison.Ordinal)) {
            // sync literal
            var continuationResponse = new Response(null, "+ continue\r\n", null, null);

            Send(continuationResponse, commandStream);
          }

          continue;
        }

        if (!command.EndsWith("\r\n", StringComparison.Ordinal))
          // incomplete command
          continue;

        lock (commandQueue) {
          commandQueue.Enqueue(command);
        }

        /*
         * send response
         */
        while (responseQueue.Count == 0) {
          clientRequestWaitHandle.Set();

          if (!IsConnected(client))
            return;

          Thread.Sleep(10);
        }

        var response = responseQueue.Dequeue();

        if (string.IsNullOrEmpty(response.ExpectedCommand) ||
            1 /*SP*/ <= command.IndexOf(response.ExpectedCommand, StringComparison.OrdinalIgnoreCase))
          Send(response, commandStream);
        else
          throw new InvalidOperationException(string.Format("unexpected command{0}expected '{1}' but was '{2}'",
                                                            Environment.NewLine,
                                                            response.ExpectedCommand,
                                                            command.Replace("\r", "\\r").Replace("\n", "\\n")));

        commandBuffer.Length = 0;

        if (response.IsByeResponse)
          return;
      }
    }

    private void Send(Response response, Stream requestStream)
    {
      if (response.ActionBeforeRespond != null)
        response.ActionBeforeRespond();

      var bytes = NetworkTransferEncoding.Transfer8Bit.GetBytes(response.ResponseText);

      requestStream.Write(bytes, 0, bytes.Length);

      Console.Write("S:\t");
      Console.Write(response.ResponseText);

      if (response.ActionAfterRespond != null)
        response.ActionAfterRespond();
    }

    private static readonly byte[] peekClientBuffer = new byte[1];

    private static bool IsConnected(TcpClient client)
    {
      try {
        if (client.Client.Poll(0, SelectMode.SelectRead)) {
          if (client.Client.Receive(peekClientBuffer, SocketFlags.Peek) == 0) {
            Console.WriteLine("client disconnected");
            return false;
          }
        }
      }
      catch (SocketException ex) {
        switch (ex.SocketErrorCode) {
          case SocketError.ConnectionAborted:
          case SocketError.ConnectionReset:
            return false;

          default:
            Console.Error.WriteLine("unexpected socket error: {0}", ex.SocketErrorCode);
            throw;
        }
      }

      return true;
    }

    private TcpListener server;
    private IPEndPoint serverEndPoint;
    private IAsyncResult acceptClientAsyncResult;
    private readonly Queue<Response> responseQueue = new Queue<Response>();
    private readonly Queue<string> commandQueue = new Queue<string>();
#if NET_4_0
    private readonly ManualResetEventSlim clientRequestWaitHandle = new ManualResetEventSlim(false);
#else
    private ManualResetEvent clientRequestWaitHandle;
#endif
    private Response greetingResponse;
    private int responseTag = 0;
    private Exception serverProcException = null;

    private static readonly Regex literalSizeRegex = new Regex(@"\{(\+)?(\d+)\}\r\n$", RegexOptions.Singleline | RegexOptions.Compiled | RegexOptions.CultureInvariant);
  }
}

