using System;
using System.Text;
using NUnit.Framework;

using Smdn.Formats;

namespace Smdn.Net.Imap4 {
  [TestFixture]
  public class ImapEnvelopeTests {
    [Test]
    public void TestAddressListProperties_ReadOnly()
    {
#if !NET_4_5
      var addresses = new[] {
        new ImapAddress("test", "test", "test", "test"),
      };
      var envelope = new ImapEnvelope(null, null, addresses, addresses, addresses, addresses, addresses, addresses, null, null);

      Assert.IsTrue(envelope.From.IsReadOnly, "From");
      Assert.IsTrue(envelope.Sender.IsReadOnly, "Sender");
      Assert.IsTrue(envelope.ReplyTo.IsReadOnly, "ReplyTo");
      Assert.IsTrue(envelope.To.IsReadOnly, "To");
      Assert.IsTrue(envelope.Cc.IsReadOnly, "Cc");
      Assert.IsTrue(envelope.Bcc.IsReadOnly, "Bcc");

      Assert.Throws<NotSupportedException>(() => envelope.From.Clear(), "From");
      Assert.Throws<NotSupportedException>(() => envelope.Sender.Clear(), "Sender");
      Assert.Throws<NotSupportedException>(() => envelope.ReplyTo.Clear(), "ReplyTo");
      Assert.Throws<NotSupportedException>(() => envelope.To.Clear(), "To");
      Assert.Throws<NotSupportedException>(() => envelope.Cc.Clear(), "Cc");
      Assert.Throws<NotSupportedException>(() => envelope.Bcc.Clear(), "Bcc");
#endif
    }

    [Test]
    public void TestAddressListProperties_NonNull()
    {
      var envelope = new ImapEnvelope(null, null, null, null, null, null, null, null, null, null);

      CollectionAssert.IsEmpty(envelope.From, "From");
      CollectionAssert.IsEmpty(envelope.Sender, "Sender");
      CollectionAssert.IsEmpty(envelope.ReplyTo, "ReplyTo");
      CollectionAssert.IsEmpty(envelope.To, "To");
      CollectionAssert.IsEmpty(envelope.Cc, "Cc");
      CollectionAssert.IsEmpty(envelope.Bcc, "Bcc");
    }

    [Test]
    public void TestGetDecodedSubject()
    {
      var envelope1 = new ImapEnvelope(null, "not encoded", null, null, null, null, null, null, null, null);

      Assert.AreEqual("not encoded", envelope1.GetDecodedSubject());
      Assert.AreEqual("not encoded", envelope1.Subject);

      var envelope2 = new ImapEnvelope(null, "=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", null, null, null, null, null, null, null, null);

      Assert.AreEqual("=?ISO-2022-JP?B?GyRCJUYlOSVIGyhC?=", envelope2.Subject);
      Assert.AreEqual("テスト", envelope2.GetDecodedSubject());

      var envelope3 = new ImapEnvelope(null, null, null, null, null, null, null, null, null, null);

      Assert.IsNull(envelope3.Subject);
      Assert.IsNull(envelope3.GetDecodedSubject());
    }

    [Test]
    public void TestGetDecodedSubject_InvalidMimeEncoding()
    {
      var envelope = new ImapEnvelope(null, "=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", null, null, null, null, null, null, null, null);

      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.Subject);
      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject());
      Assert.AreEqual("=?ISO-2022-JP?X?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject(null, false));

      Assert.Throws<FormatException>(() => envelope.GetDecodedSubject(null, true));
    }

    [Test]
    public void TestGetDecodedSubject_InvalidCharset()
    {
      var envelope = new ImapEnvelope(null, "=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", null, null, null, null, null, null, null, null);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.Subject);
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject());
      Assert.AreEqual("=?X-UNKNOWN?B?GyRCJUYlOSVIGyhC?=", envelope.GetDecodedSubject(null ,false));

      Assert.Throws<EncodingNotSupportedException>(() => envelope.GetDecodedSubject(null, true));

      Assert.AreEqual("テスト", envelope.GetDecodedSubject(delegate(string encoding) {
        Assert.AreEqual("X-UNKNOWN", encoding);

        return Encoding.GetEncoding("iso-2022-jp");
      }, true));
    }

    [Test]
    public void TestToString()
    {
      var addresses = new[] {
        new ImapAddress("test1", "test1", "test1", "test1"),
        new ImapAddress("test2", "test2", "test2", "test2"),
      };
      var envelope = new ImapEnvelope("date", "subject", addresses, addresses, addresses, addresses, addresses, addresses, "in-reply-to", "message-id");

      Assert.DoesNotThrow(() => envelope.ToString());
    }
  }
}